#include <stdio.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "path_delay.h"
#include "path_delay2.h"
#include "net_delay.h"
#include "vpr_utils.h"
#include <assert.h>

 /*----------------------------------------------------------------------------
  *  Copyright 2002 By Kara K.W. Poon and Steve Wilton
  *
  * Permission to use, copy, and modify this software and its documentation is
  * hereby granted only under the following terms and conditions.  Both the
  * above copyright notice and this permission notice must appear in all copies
  * of the software, derivative works or modified versions, and any portions
  * thereof, and both notices must appear in supporting documentation.
  * Permission is granted ONLY for non-commercial use.  For
  * commercial use, please contact the authors.
  *
  * This software may be distributed (but not offered for sale or transferred
  * for compensation) to third parties, provided such third parties agree to
  * abide by the terms and conditions of this notice.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
  * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
  * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
  * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
  * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
  * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
  * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
  * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *----------------------------------------------------------------------------*/
/*
  VPR was first written by Vaugh Betz at the University of 
  Toronto. Descriptions on different versions of VPR is available
  in the descript.txt file.

  The power model is an extension of the original VPR. This 
  feature is added by Kara Poon at the University of British 
  Columbia as a master thesis project supervised by Dr. Steve
  Wilton.

 */


/****************** Timing graph Structure ************************************
 *                                                                            *
 * In the timing graph I create, input pads and constant generators have no   *
 * inputs; everything else has inputs.  Every input pad and output pad is     *
 * represented by two tnodes -- an input pin and an output pin.  For an input *
 * pad the input pin comes from off chip and has no fanin, while the output   *
 * pin drives outpads and/or CLBs.  For output pads, the input node is driven *
 * by a CLB or input pad, and the output node goes off chip and has no        *
 * fanout (out-edges).  I need two nodes to respresent things like pads       *
 * because I mark all delay on tedges, not on tnodes.                         *
 *                                                                            *
 * Every used (not OPEN) CLB pin becomes a timing node.  As well, every used  *
 * subblock pin within a CLB also becomes a timing node.  Unused (OPEN) pins  *
 * don't create any timing nodes. If a subblock is used in combinational mode *
 * (i.e. its clock pin is open), I just hook the subblock input tnodes to the *
 * subblock output tnode.  If the subblock is used in sequential mode, I      *
 * create two extra tnodes.  One is just the subblock clock pin, which is     *
 * connected to the subblock output.  This means that FFs don't generate      *
 * their output until their clock arrives.  For global clocks coming from an  *
 * input pad, the delay of the clock is 0, so the FFs generate their outputs  *
 * at T = 0, as usual.  For locally-generated or gated clocks, however, the   *
 * clock will arrive later, and the FF output will be generated later.  This  *
 * lets me properly model things like ripple counters and gated clocks.  The  *
 * other extra node is the FF storage node (i.e. a sink), which connects to   *
 * the subblock inputs and has no fanout.                                     *
 *                                                                            *
 * One other subblock that needs special attention is a constant generator.   *
 * This has no used inputs, but its output is used.  I create an extra tnode, *
 * a dummy input, in addition to the output pin tnode.  The dummy tnode has   *
 * no fanin.  Since constant generators really generate their outputs at T =  *
 * -infinity, I set the delay from the input tnode to the output to a large-  *
 * magnitude negative number.  This guarantees every block that needs the     *
 * output of a constant generator sees it available very early.               *
 *                                                                            *
 * For this routine to work properly, subblocks whose outputs are unused must *
 * be completely empty -- all their input pins and their clock pin must be    *
 * OPEN.  Check_netlist checks the input netlist to guarantee this -- don't   *
 * disable that check.                                                        *
 *                                                                            *
 * NB:  The discussion below is only relevant for circuits with multiple      *
 * clocks.  For circuits with a single clock, everything I do is exactly      *
 * correct.                                                                   *
 *                                                                            *
 * A note about how I handle FFs:  By hooking the clock pin up to the FF      *
 * output, I properly model the time at which the FF generates its output.    *
 * I don't do a completely rigorous job of modelling required arrival time at *
 * the FF input, however.  I assume every FF and outpad needs its input at    *
 * T = 0, which is when the earliest clock arrives.  This can be conservative *
 * -- a fuller analysis would be to do a fast path analysis of the clock      *
 * feeding each FF and subtract its earliest arrival time from the delay of   *
 * the D signal to the FF input.  This is too much work, so I'm not doing it. *
 * Alternatively, when one has N clocks, it might be better to just do N      *
 * separate timing analyses, with only signals from FFs clocked on clock i    *
 * being propagated forward on analysis i, and only FFs clocked on i being    *
 * considered as sinks.  This gives all the critical paths within clock       *
 * domains, but ignores interactions.  Instead, I assume all the clocks are   *
 * more-or-less synchronized (they might be gated or locally-generated, but   *
 * they all have the same frequency) and explore all interactions.  Tough to  *
 * say what's the better way.  Since multiple clocks aren't important for my  *
 * work, it's not worth bothering about much.                                 *
 *                                                                            *
 ******************************************************************************/



/***************** Variables local to this module ***************************/

/* Variables for "chunking" the tedge memory.  If the head pointer is NULL, *
 * no timing graph exists now.                                              */

static struct s_linked_vptr *tedge_ch_list_head = NULL;
static int tedge_ch_bytes_avail = 0;
static char *tedge_ch_next_avail = NULL;



/***************** Subroutines local to this module *************************/

static int alloc_and_load_pin_mappings (int ***block_pin_to_tnode_ptr,
            int ****sblk_pin_to_tnode_ptr,
            int ***sblk_cin_pin_to_tnode_ptr,
            int ***sblk_cout_pin_to_tnode_ptr,
            int **cluster_cin_pin_to_tnode_ptr,
            int **cluster_cout_pin_to_tnode_ptr,
            t_subblock_data subblock_data,
            int ***num_uses_of_sblk_opin); 

static void free_pin_mappings (int **block_pin_to_tnode, 
          int ***sblk_pin_to_tnode, 
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode,
          int *cluster_cin_pin_to_tnode,
          int *cluster_cout_pin_to_tnode,
          int *num_subblocks_per_block);

static void alloc_and_load_fanout_counts (int ***num_uses_of_clb_ipin_ptr,
          int ****num_uses_of_sblk_opin_ptr, t_subblock_data subblock_data); 

static void free_fanout_counts (int **num_uses_of_clb_ipin, int
             ***num_uses_of_sblk_opin,  t_subblock_data subblock_data); 

static float **alloc_net_slack (void); 

static void compute_net_slacks (float **net_slack);

static void alloc_and_load_tnodes_and_net_mapping (int **num_uses_of_clb_ipin, 
          int ***num_uses_of_sblk_opin, int **block_pin_to_tnode, int ***
          sblk_pin_to_tnode, 
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode,
          int *cluster_cin_pin_to_tnode,
          int *cluster_cout_pin_to_tnode,
          t_subblock_data subblock_data, t_timing_inf timing_inf);

static void build_clb_tnodes (int iblk, int *n_uses_of_clb_ipin, int
        **block_pin_to_tnode, int **sub_pin_to_tnode, int subblock_num_pin,
	int subblock_num_ipin, int subblock_num_opin,
        int num_subs, t_subblock *sub_inf, float T_clb_ipin_to_sblk_ipin,
        int *next_clb_ipin_edge, 
        int *cluster_cin_pin_to_tnode,
        int *cluster_cout_pin_to_tnode); 

static void build_subblock_tnodes (
          t_subblock_data subblock_data,
          int **n_uses_of_sblk_opin, 
          int *blk_pin_to_tnode, int **sub_pin_to_tnode, 
          int subblock_lut_size, int num_subs,
		  int subblock_num_pin,
		  int subblock_num_opin, int subblock_num_ipin,
		  float T_FX,
          float T_sblk_opin_to_sblk_ipin, float T_sblk_opin_to_clb_opin, 
          float T_sblk_cout_to_sblk_cin, float T_clb_cout_to_clb_cin,
          t_T_subblock *T_subblock, int **next_sblk_opin_edge, int iblk,
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode);

static void build_ipad_tnodes (int iblk, int **block_pin_to_tnode, float
            T_ipad, int *num_subblocks_per_block, t_subblock
            **subblock_inf);

static boolean is_global_clock (int iblk, int *num_subblocks_per_block,
     t_subblock **subblock_inf);

static void build_opad_tnodes (int *blk_pin_to_tnode, float T_opad, int iblk, float T_clk_opad);

static void build_block_output_tnode (int inode, int iblk, int ipin,
            int **block_pin_to_tnode); 


static void build_dsp_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *dsp_timing_inf);

static void build_memory_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *dsp_timing_inf);
		
static int get_opin(int pin_num);

static void build_emb_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *dsp_timing_inf) ;

/********************* Subroutine definitions *******************************/


void free_subblock_data (t_subblock_data *subblock_data_ptr) {

/* Frees all the subblock data structures.                                 */

 free_chunk_memory (subblock_data_ptr->chunk_head_ptr);
 free (subblock_data_ptr->num_subblocks_per_block);
 free (subblock_data_ptr->subblock_inf);

/* Mark as freed. */
 
 subblock_data_ptr->num_subblocks_per_block = NULL;
 subblock_data_ptr->subblock_inf = NULL;
 subblock_data_ptr->chunk_head_ptr = NULL;
}

void free_clock_inf(struct clock_details *clock_inf_ptr)
{
  free(clock_inf_ptr);
  clock_inf_ptr = NULL;
}

void free_temperature_record (struct temperature_record *temperatur_record_ptr)
{
  free(temperatur_record_ptr);
  temperatur_record_ptr = NULL;
}


float **alloc_and_load_timing_graph (t_timing_inf timing_inf, t_subblock_data
       subblock_data) {

/* This routine builds the graph used for timing analysis.  Every clb or    *
 * subblock pin is a timing node (tnode).  The connectivity between pins is *
 * represented by timing edges (tedges).  All delay is marked on edges, not *
 * on nodes.  This routine returns an array that will store slack values:   *
 * net_slack[0..num_nets-1][1..num_pins-1].                                 */

/* The two arrays below are valid only for CLBs, not pads.                  */

 int **num_uses_of_clb_ipin;  /* [0..num_blocks-1][0..pins_per_clb-1]       */
 //int **num_uses_of_sblk_opin; /* [0..num_blocks-1][0..num_subs_per[iblk]-1] */
 int ***num_uses_of_sblk_opin; /* [0..num_blocks-1][0..num_subs_per[iblk]-1][0..subblock_num_opin[iblk]-1] */

/* Array for mapping from a pin on a block to a tnode index. For pads, only *
 * the first two pin locations are used (input to pad is first, output of   *
 * pad is second).  For clbs, all OPEN pins on the clb have their mapping   *
 * set to OPEN so I won't use it by mistake.                                */

 int **block_pin_to_tnode;     /* [0..num_blocks-1][0..pins_per_clb-1+1]      */
 int *block_cin_pin_to_tnode;
 int *block_cout_pin_to_tnode;

/* Array for mapping from a pin on a subblock to a tnode index.  Unused     *
 * or nonexistent subblock pins have their mapping set to OPEN.             *
 * [0..num_blocks-1][0..num_subblocks_per_block-1][0..subblock_lut_size+1]  */

/* Note that for a DSP or memory block, there is really only one subblock.  *
   In that case, the final array contains nodes T1 and T2 within the        *
   memory block.  */

 int ***sblk_pin_to_tnode;
 int **sblk_cin_pin_to_tnode;
 int **sblk_cout_pin_to_tnode;
 int *cluster_cin_pin_to_tnode;
 int *cluster_cout_pin_to_tnode;

 int num_sinks;
 float **net_slack;    /* [0..num_nets-1][1..num_pins-1]. */

/************* End of variable declarations ********************************/

// printf("alloc_and_load_timing_graph\n");

 if (tedge_ch_list_head != NULL) {
    printf ("Error in alloc_and_load_timing_graph:\n"
            "\tAn old timing graph still exists.\n");
    exit (1);
 }

/* If either of the checks below ever fail, change the definition of        *
 * tnode_descript to use ints instead of shorts for isubblk or ipin.        */

 if (subblock_data.max_subblocks_per_block > MAX_SHORT) {
    printf ("Error in alloc_and_load_timing_graph: max_subblocks_per_block"
            "\tis %d -- will cause short overflow in tnode_descript.\n", 
            subblock_data.max_subblocks_per_block);
    exit (1);
 }

 if (pins_per_clb > MAX_SHORT) {
    printf ("Error in alloc_and_load_timing_graph: pins_per_clb is %d."
            "\tWill cause short overflow in tnode_descript.\n", pins_per_clb);
    exit (1);
 }
 if (pins_per_dsp > MAX_SHORT) {
    printf ("Error in alloc_and_load_timing_graph: pins_per_dsp is %d."
            "\tWill cause short overflow in tnode_descript.\n", pins_per_dsp);
    exit (1);
 }
 if (pins_per_memory > MAX_SHORT) {
    printf ("Error in alloc_and_load_timing_graph: pins_per_memory is %d."
            "\tWill cause short overflow in tnode_descript.\n", pins_per_memory);
    exit (1);
 }

// printf("alloc_and_load_fanout_counts\n");
 alloc_and_load_fanout_counts (&num_uses_of_clb_ipin, &num_uses_of_sblk_opin,
               subblock_data);
	       
 
 //printf("alloc_and_load_pin_mappings\n");
 num_tnodes = alloc_and_load_pin_mappings (&block_pin_to_tnode, 
               &sblk_pin_to_tnode,
               &sblk_cin_pin_to_tnode,
               &sblk_cout_pin_to_tnode,
               &cluster_cin_pin_to_tnode,
               &cluster_cout_pin_to_tnode,
               subblock_data, num_uses_of_sblk_opin);



//printf("alloc_and_load_tnodes_and_net_mappings\n");
 alloc_and_load_tnodes_and_net_mapping (num_uses_of_clb_ipin, 
               num_uses_of_sblk_opin, block_pin_to_tnode, sblk_pin_to_tnode, 
               sblk_cin_pin_to_tnode,
               sblk_cout_pin_to_tnode,
               cluster_cin_pin_to_tnode,
               cluster_cout_pin_to_tnode,       
               subblock_data, timing_inf);

//printf("dump timing_before\n");
//dump_timing_graph("timing_before.echo");

//printf("alloc_and_load_timing_graph_levels\n");

 num_sinks =  alloc_and_load_timing_graph_levels ();


printf("number of node %d , number of sinks %d\n", num_tnodes , num_sinks);

// printf("check_timing_graph, num_const_gen =%d\n", subblock_data.num_const_gen);
 check_timing_graph (subblock_data.num_const_gen, subblock_data.num_ff, num_sinks);

//printf("free_fanout_counts\n");
 free_fanout_counts (num_uses_of_clb_ipin, num_uses_of_sblk_opin, subblock_data);
// printf("free_pin_mappings\n");
 free_pin_mappings (block_pin_to_tnode, sblk_pin_to_tnode,
                   sblk_cin_pin_to_tnode,sblk_cout_pin_to_tnode,
                   cluster_cin_pin_to_tnode,cluster_cout_pin_to_tnode,
                   subblock_data.num_subblocks_per_block);


 net_slack = alloc_net_slack ();

 return (net_slack);
}


static float **alloc_net_slack (void) {

/* Allocates the net_slack structure.  Chunk allocated to save space.      */

 float **net_slack;   /* [0..num_nets-1][1..num_pins-1]  */
 float *tmp_ptr;
 int inet;

// printf("alloc_net_slack, num_nets %d \n", num_nets);
 net_slack = (float **) my_malloc (num_nets * sizeof (float *));

 for (inet=0;inet<num_nets;inet++) {
    tmp_ptr = (float *) my_chunk_malloc ((net[inet].num_pins - 1) * 
              sizeof (float), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    net_slack[inet] = tmp_ptr - 1;   /* [1..num_pins-1] */
 }
 
 return (net_slack);
}


static int alloc_and_load_pin_mappings (int ***block_pin_to_tnode_ptr,
            int ****sblk_pin_to_tnode_ptr,
            int ***sblk_cin_pin_to_tnode_ptr,
            int ***sblk_cout_pin_to_tnode_ptr,
            int **cluster_cin_pin_to_tnode_ptr,
            int **cluster_cout_pin_to_tnode_ptr,
            t_subblock_data subblock_data,
            int ***num_uses_of_sblk_opin) {

/* Allocates and loads the block_pin_to_tnode and sblk_pin_to_tnode         *
 * structures, and computes num_tnodes.                                     */

 int iblk, isub, ipin, opin,  num_subblocks; 
// int out_pin, clk_pin; 
 int clk_pin1, clk_pin2;
 int curr_tnode;
 int ***sblk_pin_to_tnode, **block_pin_to_tnode;
 int **sblk_cin_pin_to_tnode;
 int **sblk_cout_pin_to_tnode;
 int *cluster_cin_pin_to_tnode;
 int *cluster_cout_pin_to_tnode;
 int subblock_lut_size;
 int subblock_num_pin;
 int subblock_num_ipin;
 int subblock_num_opin;

 int *num_subblocks_per_block;
 t_subblock **subblock_inf;
 boolean has_inputs;
 int ppins;
 int has_input;
 int i;

 subblock_lut_size = subblock_data.subblock_lut_size;
 subblock_num_pin = subblock_data.subblock_num_pin;
 subblock_num_ipin = subblock_data.subblock_num_ipin;
 subblock_num_opin = subblock_data.subblock_num_opin;

 num_subblocks_per_block = subblock_data.num_subblocks_per_block;
 subblock_inf = subblock_data.subblock_inf;

 ppins = MAX(pins_per_clb,pins_per_dsp);
 ppins = MAX(ppins, pins_per_memory);

 for (i = 1; i < number_module; i++)
 {	
	 ppins = MAX(ppins, constraint[i].num_pins);
 }

 block_pin_to_tnode = (int **) alloc_matrix (0, num_blocks - 1, 0, 
                                             ppins, sizeof (int));

 sblk_pin_to_tnode = (int ***) my_malloc (num_blocks * sizeof (int **));

 sblk_cin_pin_to_tnode = (int **) alloc_matrix(0, num_blocks-1, 0,
                                    subblock_data.max_subblocks_per_block, sizeof (int));
 sblk_cout_pin_to_tnode = (int **) alloc_matrix(0, num_blocks-1, 0,
                                    subblock_data.max_subblocks_per_block, sizeof (int));
 cluster_cin_pin_to_tnode = (int *) my_malloc(num_blocks * sizeof(int));
 cluster_cout_pin_to_tnode = (int *) my_malloc(num_blocks * sizeof(int));

 curr_tnode = 0;

	for (iblk=0;iblk<num_blocks;iblk++) 
	{
    		if (block[iblk].type == CLB) 
		{

//			out_pin = subblock_lut_size;
//		    clk_pin = subblock_lut_size + 1;

       		/* First do the block mapping */

       		for (ipin=0;ipin<pins_per_clb;ipin++) 
			{
	          	if (block[iblk].nets[ipin] == OPEN) 
				{
		             block_pin_to_tnode[iblk][ipin] = OPEN;
		        }
          		else 
				{
             		block_pin_to_tnode[iblk][ipin] = curr_tnode;
	             	curr_tnode++;
			    }
       		}

	    	/* Now do the subblock mapping. */

	   	    num_subblocks = num_subblocks_per_block[iblk];
    	 	 //sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, num_subblocks - 1,
	   		//                      0, subblock_lut_size + 1, sizeof (int));
	
		//		printf("num subblock %d \n", num_subblocks);	
			sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, num_subblocks - 1, 0, subblock_num_pin , sizeof(int)); 	
			//pin order , input, output, clock

	       	for (isub=0;isub<num_subblocks;isub++) 
			{
		
        	 	/* Pin ordering:  inputs, output, clock.   */
    		      
				//has_inputs = FALSE;
				//Input first
				for (ipin=0; ipin<subblock_num_ipin; ipin ++)
				{
					if (subblock_inf[iblk][isub].inputs[ipin] != OPEN) 
					{
						//printf("alloc_and_load_pin_mapping, input, iblk :%d, isub : %d, ipin : %d, inode : %d , connect_to %d \n", iblk, isub, ipin, curr_tnode, subblock_inf[iblk][isub].inputs[ipin]);
    	            	//has_inputs = TRUE;
               		 	sblk_pin_to_tnode[iblk][isub][ipin] = curr_tnode;
				//		printf("~~ sblk_pin_to_tnode[%d][%d][%d] = %d \n", iblk, isub, ipin, curr_tnode);
	                	curr_tnode++;
    	         	}
        	     	else 
					{											
                		sblk_pin_to_tnode[iblk][isub][ipin] = OPEN;
	             	}	
				}

				//do for clock 
				if (subblock_inf[iblk][isub].clock != OPEN)
				{					
               			sblk_pin_to_tnode[iblk][isub][I_CLK] = curr_tnode;
						curr_tnode++;
				}
/*
          for (ipin=0;ipin<subblock_lut_size;ipin++) {
             if (subblock_inf[iblk][isub].inputs[ipin] != OPEN) {
                has_inputs = TRUE;
                sblk_pin_to_tnode[iblk][isub][ipin] = curr_tnode;
                curr_tnode++;
             }
             else {
                sblk_pin_to_tnode[iblk][isub][ipin] = OPEN;
             }
          }
*/
		//printf("******************************************************************************\n");
	        /* subblock output  */
				for (opin = 0; opin < subblock_num_opin; opin ++)
				{
					//printf("alloc_and_load_pin_mapping, output, iblk :%d, isub : %d, opin : %d, inode : %d , connect_to %d, subblock_num_ipin + opin = %d\n", iblk, isub, opin, curr_tnode, subblock_inf[iblk][isub].outputs[opin], subblock_num_ipin + opin);
					if (subblock_inf[iblk][isub].outputs[opin] != OPEN) 
					{
						
							sblk_pin_to_tnode[iblk][isub][subblock_num_ipin + opin] = curr_tnode;
							//printf("iblk %d, isub %d, opin %d, inode %d \n",iblk, isub, opin, sblk_pin_to_tnode[iblk][isub][subblock_num_ipin + opin]);
							curr_tnode++;

					}
					else 
					{
						sblk_pin_to_tnode[iblk][isub][subblock_num_ipin + opin] = OPEN;
					}
				}
	        	/* If the subblock opin is unused the subblock is empty and we    *
		         * shoudn't count it.                                             */
     		 }
	    } /* End if a clb. */

    else if (block[iblk].type == DSP) {

		clk_pin2 = pins_per_dsp - 3;
		//clk_pin2 = pins_per_dsp - 1;
		
	   //printf ("iblk %d (%s), pins per dsp = %d, block[iblk].nets[clk_pin2] = %d (%s)\n", iblk, block[iblk].name, pins_per_dsp, block[iblk].nets[clk_pin2], net[block[iblk].nets[clk_pin2]].name);

       for (ipin=0;ipin<pins_per_dsp;ipin++) {
          if (block[iblk].nets[ipin] == OPEN) 
		  {
             block_pin_to_tnode[iblk][ipin] = OPEN;
          }
          else 
		  {
             block_pin_to_tnode[iblk][ipin] = curr_tnode;			
             curr_tnode++;
          }
       }
       sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, 0,
                                  0, 1, sizeof (int));
       sblk_pin_to_tnode[iblk][0][0] = curr_tnode;
       curr_tnode++;
       sblk_pin_to_tnode[iblk][0][1] = curr_tnode;
       curr_tnode++;
       if (block[iblk].nets[clk_pin2] != OPEN) {
          curr_tnode+= 2;
       } 
	   /*
       if (block[iblk].nets[clk_pin2] != OPEN) {
          curr_tnode+= 2;
       } 
*/
		/*
       clk_pin1 = pins_per_dsp - 2;
       clk_pin2 = pins_per_dsp - 1;

	   printf ("pins per dsp = %d \n", pins_per_dsp);
       for (ipin=0;ipin<pins_per_dsp;ipin++) {
          if (block[iblk].nets[ipin] == OPEN) {
             block_pin_to_tnode[iblk][ipin] = OPEN;
          }
          else {
             block_pin_to_tnode[iblk][ipin] = curr_tnode;			

             curr_tnode++;
          }
       }
       sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, 0,
                                  0, 1, sizeof (int));
       sblk_pin_to_tnode[iblk][0][0] = curr_tnode;
       curr_tnode++;
       sblk_pin_to_tnode[iblk][0][1] = curr_tnode;
       curr_tnode++;
       if (block[iblk].nets[clk_pin1] != OPEN) {
          curr_tnode+= 2;
       } 
       if (block[iblk].nets[clk_pin2] != OPEN) {
          curr_tnode+= 2;
       } 
		*/
    }
    else if (block[iblk].type ==EMB) {

//		printf("pins per emb %d \n", constraint[block[iblk].module_index].num_pins);
		
       clk_pin1 = 0;       

       //for (ipin=0;ipin<pins_per_dsp;ipin++) {
	   for (ipin=0;ipin<constraint[block[iblk].module_index].num_pins;ipin++) 
	   {
          if (block[iblk].nets[ipin] == OPEN) {
             block_pin_to_tnode[iblk][ipin] = OPEN;
          }
          else {
             block_pin_to_tnode[iblk][ipin] = curr_tnode;
//			 printf("EMB %d (%s), inode %d \n", iblk, block[iblk].name, block_pin_to_tnode[iblk][ipin]);
             curr_tnode++;
          }
       }
	   //no subblock in EMB

       sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, 0,
                                  0, 1, sizeof (int));
       sblk_pin_to_tnode[iblk][0][0] = curr_tnode;
       curr_tnode++;
       sblk_pin_to_tnode[iblk][0][1] = curr_tnode;
       curr_tnode++;
/*
       if (block[iblk].nets[clk_pin1] != OPEN) {
          curr_tnode+= 2;
       } 
*/
	   

    }

    else if (block[iblk].type == MEMORY) {
       clk_pin1 = pins_per_memory - 2;
       clk_pin2 = pins_per_memory - 1;

       for (ipin=0;ipin<pins_per_memory;ipin++) {
          if (block[iblk].nets[ipin] == OPEN) {
             block_pin_to_tnode[iblk][ipin] = OPEN;
          }
          else {
             block_pin_to_tnode[iblk][ipin] = curr_tnode;
             curr_tnode++;
          }
       }
       sblk_pin_to_tnode[iblk] = (int **) alloc_matrix (0, 0,
                                  0, 1, sizeof (int));
       sblk_pin_to_tnode[iblk][0][0] = curr_tnode;
       curr_tnode++;
       sblk_pin_to_tnode[iblk][0][1] = curr_tnode;
       curr_tnode++;
       if (block[iblk].nets[clk_pin1] != OPEN) {
          curr_tnode+= 2;
       } 
       if (block[iblk].nets[clk_pin2] != OPEN) {
          curr_tnode+= 2;
       } 
    }
    else if (block[iblk].type == INPAD)
	{   /* INPAD */
		//for INPAD , 0 : pad input, 1: pad output
       block_pin_to_tnode[iblk][0] = curr_tnode;         /* Pad input  */
//	   printf("INPAD %d (%s), inode0 %d \n", iblk, block[iblk].name, block_pin_to_tnode[iblk][0]);
       block_pin_to_tnode[iblk][1] = curr_tnode + 1;     /* Pad output */
//	   printf("INPAD %d (%s), inode1 %d \n", iblk, block[iblk].name, block_pin_to_tnode[iblk][0]);
       curr_tnode += 2;

       for (ipin=2;ipin<pins_per_clb;ipin++) 
          block_pin_to_tnode[iblk][ipin] = OPEN;

       sblk_pin_to_tnode[iblk] = NULL;  /* No subblock pins */
	}else
	{
		has_input = 0; //set the output pad doesn't have input first
		/*OUTPAD */		
		//for OUTPAD, 0 to 6 : pad input, 7: pad output
		for (ipin = 0; ipin < 7; ipin ++)
		{
			//Check any connection to the input pin of the outpad
			if (block[iblk].nets[ipin] != OPEN)
			{
				has_input = 1; //indicate there is input to the outpad
				block_pin_to_tnode[iblk][ipin] = curr_tnode;         /* Pad input  */		
//				printf("outpad %d (%s), inode %d \n", iblk, block[iblk].name, block_pin_to_tnode[iblk][ipin]);

				curr_tnode ++;
			}
			else
			{
				//open input pin
				block_pin_to_tnode[iblk][ipin] = OPEN;		
			}
		}
		if (has_input == 1)
		{
			//there is input to outpad, create the output node
			block_pin_to_tnode[iblk][7] = curr_tnode ;     /* Pad output */		
			curr_tnode ++;
		}
		else
		{
			//no input , then no output node
			block_pin_to_tnode[iblk][7] = OPEN ;     /* Pad output */		
			printf("Error , no input to output pad \n");
			exit(1);
		}
       //curr_tnode += 8;

       for (ipin=8;ipin<pins_per_clb;ipin++) 
          block_pin_to_tnode[iblk][ipin] = OPEN;

       sblk_pin_to_tnode[iblk] = NULL;  /* No subblock pins */

    }
 }  /* End for all blocks */

 *sblk_pin_to_tnode_ptr = sblk_pin_to_tnode;
 *block_pin_to_tnode_ptr = block_pin_to_tnode;
 *sblk_cin_pin_to_tnode_ptr = sblk_cin_pin_to_tnode;
 *sblk_cout_pin_to_tnode_ptr = sblk_cout_pin_to_tnode;
 *cluster_cin_pin_to_tnode_ptr = cluster_cin_pin_to_tnode;
 *cluster_cout_pin_to_tnode_ptr = cluster_cout_pin_to_tnode;
 return (curr_tnode);
}

static void free_pin_mappings (int **block_pin_to_tnode, 
          int ***sblk_pin_to_tnode, 
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode,
          int *cluster_cin_pin_to_tnode,
          int *cluster_cout_pin_to_tnode,
          int *num_subblocks_per_block)
{

/* Frees the arrays that map from pins to tnode coordinates. */

 int iblk;

 free_matrix (block_pin_to_tnode, 0, num_blocks - 1, 0, sizeof (int));
 free_matrix (sblk_cin_pin_to_tnode, 0, num_blocks-1, 0, sizeof(int));
 free_matrix (sblk_cout_pin_to_tnode, 0, num_blocks-1, 0, sizeof(int));
 
 for (iblk=0;iblk<num_blocks;iblk++) {
    if (block[iblk].type == CLB) {
       free_matrix (sblk_pin_to_tnode[iblk], 0, num_subblocks_per_block[iblk] 
                     - 1, 0, sizeof (int));
    }
    if (block[iblk].type == DSP) {
       free_matrix (sblk_pin_to_tnode[iblk], 0, 0 , 0, sizeof (int));
    }

	if (block[iblk].type == EMB) {
       free_matrix (sblk_pin_to_tnode[iblk], 0, 0 , 0, sizeof (int));
    }

    if (block[iblk].type == MEMORY) {
       free_matrix (sblk_pin_to_tnode[iblk], 0, 0, 0, sizeof (int));
    }
 }
 free (cluster_cin_pin_to_tnode);
 free (cluster_cout_pin_to_tnode);
 free (sblk_pin_to_tnode);
}


static void alloc_and_load_fanout_counts (int ***num_uses_of_clb_ipin_ptr,
          int ****num_uses_of_sblk_opin_ptr, t_subblock_data subblock_data) {

/* Allocates and loads two arrays that say how many points each clb input   *
 * pin and each subblock output fan out to.                                 */

 int iblk, isub, opin;
 int **num_uses_of_clb_ipin, ***num_uses_of_sblk_opin;
 int *num_subblocks_per_block;
 t_subblock **subblock_inf;
 int subblock_lut_size; 
 int subblock_num_pin;
 int subblock_num_ipin;
 int subblock_num_opin;

 num_subblocks_per_block = subblock_data.num_subblocks_per_block;
 subblock_inf = subblock_data.subblock_inf;
 subblock_lut_size = subblock_data.subblock_lut_size;
 subblock_num_pin = subblock_data.subblock_num_pin;
 subblock_num_ipin = subblock_data.subblock_num_ipin;
 subblock_num_opin = subblock_data.subblock_num_opin;

 num_uses_of_clb_ipin = (int **) my_malloc (num_blocks * sizeof (int *));

 num_uses_of_sblk_opin = (int ***) my_malloc (num_blocks * sizeof (int **));

 for (iblk=0;iblk<num_blocks;iblk++) 
 {
    if (block[iblk].type != CLB) 
	{
       num_uses_of_clb_ipin[iblk] = NULL;
       num_uses_of_sblk_opin[iblk] = NULL;
    }

    else {   /* CLB */
       num_uses_of_clb_ipin[iblk] = (int *) my_calloc (pins_per_clb, 
                    sizeof (int));
	   num_uses_of_sblk_opin[iblk] = (int **) my_calloc (
                    num_subblocks_per_block[iblk], sizeof (int*));

	   for (isub = 0; isub < num_subblocks_per_block[iblk]; isub++)
	   {
			num_uses_of_sblk_opin[iblk][isub] = (int *) my_calloc(subblock_num_opin, sizeof (int));
			/*
			for (opin = 0; opin < subblock_num_opin; opin ++)
			{
				num_uses_of_sblk_opin[iblk][isub][opin] = 0;
			}
			*/
	   }

	 
       load_one_clb_fanout_count (subblock_lut_size, subblock_inf[iblk],
                    num_subblocks_per_block[iblk], num_uses_of_clb_ipin[iblk],
                    num_uses_of_sblk_opin[iblk], iblk, subblock_num_opin, subblock_num_ipin);

    }  /* End if CLB */
 }  /* End for all blocks */

 *num_uses_of_clb_ipin_ptr = num_uses_of_clb_ipin;
 *num_uses_of_sblk_opin_ptr = num_uses_of_sblk_opin;
}


static void free_fanout_counts (int **num_uses_of_clb_ipin, int 
             ***num_uses_of_sblk_opin, t_subblock_data subblock_data) {

/* Frees the fanout count arrays. */

 int iblk, isub;
  
 for (iblk=0;iblk<num_blocks;iblk++) {
    if (block[iblk].type == CLB) {
       free (num_uses_of_clb_ipin[iblk]);
	   //for (isub = 0; isub < max_subblocks_per_block; isub ++)
	   for (isub = 0; isub < subblock_data.num_subblocks_per_block[iblk] ; isub ++)
	   {
		   free(num_uses_of_sblk_opin[iblk][isub]);
	   }
       free (num_uses_of_sblk_opin[iblk]);
    }
 }

 free (num_uses_of_clb_ipin);
 free (num_uses_of_sblk_opin);
}




static void alloc_and_load_tnodes_and_net_mapping (int **num_uses_of_clb_ipin, 
          int ***num_uses_of_sblk_opin, int **block_pin_to_tnode, int ***
          sblk_pin_to_tnode, 
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode,
          int *cluster_cin_pin_to_tnode,
          int *cluster_cout_pin_to_tnode,
          t_subblock_data subblock_data, t_timing_inf timing_inf) {

/* Does the actual allocation and building of the timing graph. */

 int iblk, subblock_lut_size;
 int subblock_num_pin, subblock_num_ipin, subblock_num_opin;
 int *num_subblocks_per_block;
 int *next_clb_ipin_edge;
 int **next_sblk_opin_edge;
 t_subblock **subblock_inf;


 tnode = (t_tnode *) my_malloc (num_tnodes * sizeof (t_tnode));
 tnode_descript = (t_tnode_descript *) my_malloc (num_tnodes * 
                   sizeof (t_tnode_descript));

 net_to_driver_tnode = (int *) my_malloc (num_nets * sizeof (int));

 next_clb_ipin_edge = (int *) my_malloc (pins_per_clb * sizeof (int));
 
 //next_sblk_opin_edge = (int *) my_malloc (subblock_data.max_subblocks_per_block* sizeof (int));
 //next_sblk_opin_edge = (int *) my_malloc (subblock_data.subblock_num_opin* sizeof (int));
// next_sblk_opin_edge = (int **) alloc_matrix (0, num_blocks - 1, 0, 
  //                                           ppins, sizeof (int));

 subblock_inf = subblock_data.subblock_inf;
 subblock_lut_size = subblock_data.subblock_lut_size;
 subblock_num_pin = subblock_data.subblock_num_pin;
 subblock_num_ipin = subblock_data.subblock_num_ipin;
 subblock_num_opin = subblock_data.subblock_num_opin;
 num_subblocks_per_block = subblock_data.num_subblocks_per_block;

// printf("max subblocks per block %d , subblock num opin %d , num_tnode %d, num_nets %d, num_blocks %d\n",subblock_data.max_subblocks_per_block, subblock_num_opin, num_tnodes, num_nets, num_blocks);
 
 next_sblk_opin_edge = (int **) my_malloc (subblock_data.max_subblocks_per_block * sizeof (int *));
 
  for (iblk=0;iblk<subblock_data.max_subblocks_per_block;iblk++) {
	next_sblk_opin_edge[iblk] = (int *) my_malloc (subblock_num_opin * sizeof (int));		
  }
 

 for (iblk=0;iblk<num_blocks;iblk++) {
    switch (block[iblk].type) {

    case CLB:
//		printf("build_clb_tnodes \n");

    	build_clb_tnodes (iblk, num_uses_of_clb_ipin[iblk], block_pin_to_tnode, 
               sblk_pin_to_tnode[iblk], subblock_num_pin,
	       subblock_num_ipin, subblock_num_opin,
               num_subblocks_per_block[iblk], subblock_inf[iblk],
               timing_inf.T_clb_ipin_to_sblk_ipin, next_clb_ipin_edge,
               cluster_cin_pin_to_tnode,
               cluster_cout_pin_to_tnode);
			   
	//	printf("build_subblock_tnodes %d\n", iblk);
		
       build_subblock_tnodes (subblock_data,num_uses_of_sblk_opin[iblk], 
               block_pin_to_tnode[iblk], sblk_pin_to_tnode[iblk], 
               subblock_lut_size, num_subblocks_per_block[iblk], 
			   subblock_num_pin,
			   subblock_num_opin, subblock_num_ipin,
			   timing_inf.T_FX, 
               timing_inf.T_sblk_opin_to_sblk_ipin, 
               timing_inf.T_sblk_opin_to_clb_opin, 
               timing_inf.T_sblk_cout_to_sblk_cin, timing_inf.T_clb_cout_to_clb_cin,
               timing_inf.T_subblock,
               next_sblk_opin_edge, iblk,
               sblk_cin_pin_to_tnode,
               sblk_cout_pin_to_tnode);
//		printf("finish build_subblock_tnodes %d\n", iblk);

       break;

    case INPAD:
       build_ipad_tnodes (iblk, block_pin_to_tnode, timing_inf.T_ipad,
                   num_subblocks_per_block, subblock_inf);
       break;

    case OUTPAD:
       build_opad_tnodes (block_pin_to_tnode[iblk], timing_inf.T_opad, iblk, timing_inf.T_clk_opad);
       break;

    case MEMORY:
       build_memory_tnodes (iblk, block_pin_to_tnode, 
               sblk_pin_to_tnode[iblk], 
               &(timing_inf.T_memory));
       break;

    case DSP:
       build_dsp_tnodes (iblk, block_pin_to_tnode, 
               sblk_pin_to_tnode[iblk], 
               &(timing_inf.T_dsp));
       break;

	case EMB:
       build_emb_tnodes (iblk, block_pin_to_tnode, 
               sblk_pin_to_tnode[iblk], 
               &(timing_inf.T_dsp));
       break;
   
    default:
       printf ("Error in `:\n"
               "\tUnexpected block type (%d) for block %d (%s).\n", 
                block[iblk].type, iblk, block[iblk].name);
       exit (1);
    }
 }



 free (next_clb_ipin_edge);

  for (iblk=0;iblk<subblock_data.max_subblocks_per_block;iblk++) {
	free(next_sblk_opin_edge[iblk]);
  }


 free (next_sblk_opin_edge);

 
}


static void build_clb_tnodes (int iblk, int *n_uses_of_clb_ipin, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode, int subblock_num_pin,
	int subblock_num_ipin, int subblock_num_opin, 
        int num_subs, t_subblock *sub_inf, float T_clb_ipin_to_sblk_ipin,
        int *next_clb_ipin_edge,
        int *cluster_cin_pin_to_tnode,
        int *cluster_cout_pin_to_tnode) {

/* This routine builds the tnodes corresponding to the clb pins of this     *
 * block, and properly hooks them up to the rest of the graph. Note that    *
 * only the sblk_pin_to_tnode, etc. element for this block is passed in.    */

 int isub, ipin, iedge, from_pin, opin;
 int inode, to_node, num_edges;
 t_tedge *tedge;
// int clk_pin;


 //clk_pin = subblock_lut_size + 1;
// clk_pin = subblock_num_pin + 1;

/* Start by allocating the edge arrays, and for opins, loading them.    */

 for (ipin=0;ipin<pins_per_clb;ipin++) {
    inode = block_pin_to_tnode[iblk][ipin];

    if (inode != OPEN) {   /* Pin is used -> put in graph */
       if (is_opin (ipin)) {		  
          build_block_output_tnode (inode, iblk, ipin, block_pin_to_tnode);
		  //printf("CLB opin: inode = %d \n", inode);
          tnode_descript[inode].type = CLB_OPIN;
       }

       else {    /* CLB ipin */
          next_clb_ipin_edge[ipin] = 0;            /* Reset */
          num_edges = n_uses_of_clb_ipin[ipin];
          tnode[inode].num_edges = num_edges;	

		 //printf("CLB ipin: block_name: %s , iblk %d,, ipin = %d, inode = %d, num_edges = %d \n", block[iblk].name, iblk, ipin, inode, num_edges);
          tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

          tnode_descript[inode].type = CLB_IPIN;
       }

       tnode_descript[inode].ipin = ipin;
       tnode_descript[inode].isubblk = OPEN;
       tnode_descript[inode].iblk = iblk;
    }
 }

/* Now load the edge arrays for the CLB input pins. Do this by looking at   *
 * where the subblock input and clock pins are driven from.                 */

 for (isub=0;isub<num_subs;isub++) 
 {
    //for (ipin=0;ipin<subblock_lut_size;ipin++) {
    //for (ipin=0;ipin<subblock_num_ipin-1;ipin++) //don't check the clock pin
	for (ipin=0;ipin<subblock_num_ipin;ipin++) //don't check the clock pin
	{
		
			from_pin = sub_inf[isub].inputs[ipin];

			/* Not OPEN and comes from clb ipin? */

			if (from_pin != OPEN && from_pin < pins_per_clb) 
			{ 
				inode = block_pin_to_tnode[iblk][from_pin];
				to_node = sub_pin_to_tnode[isub][ipin];
				tedge = tnode[inode].out_edges;
				iedge = next_clb_ipin_edge[from_pin]++;
				tedge[iedge].to_node = to_node;
				tedge[iedge].Tdel = T_clb_ipin_to_sblk_ipin;
			}
		
    }

    from_pin = sub_inf[isub].clock;	
	
    if (from_pin != OPEN && from_pin < pins_per_clb) 
	{ 
		inode = block_pin_to_tnode[iblk][from_pin];
				
//		to_node = sub_pin_to_tnode[isub][subblock_num_ipin + 5];  /* Feeds seq. output to Q */	  
 		to_node = sub_pin_to_tnode[isub][I_CLK]; //connect to block clk pin 
		//printf("clock pin, from_pin %d, inode %d, to_node %d \n", from_pin, inode, to_node);
		if (to_node != OPEN)
		{
			tedge = tnode[inode].out_edges;
			iedge = next_clb_ipin_edge[from_pin]++;
			tedge[iedge].to_node = to_node;
			//printf("clock pin, from_pin %d, inode %d, to_node %d \n", from_pin, inode, to_node);

			/* For the earliest possible clock I want this delay to be zero, so it       *
			* arrives at flip flops at T = 0.  For later clocks or locally generated    *
			* clocks that may accumulate delay (like the clocks in a ripple counter),   *
			* I might want to make this delay nonzero.  Not worth bothering about now.  */

			tedge[iedge].Tdel = 0.;
		}
    }
 }
}

static void build_emb_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *dsp_timing_inf) {

/* This routine builds the tnodes corresponding to the dsp pins of this     *
 * block, and properly hooks them up to the rest of the graph.  */

 int ipin, iedge, cnt, inode, to_node, num_edges;
 int clk_pin;
 int num_output_nodes;
 int inet;

 clk_pin = 0;
 num_output_nodes = 0;

 /* Start by allocating the edge arrays, and for opins, loading them.    */

 /* EMB pins */

 //for (ipin=0;ipin<pins_per_dsp;ipin++) 
 for (ipin=0;ipin<constraint[block[iblk].module_index].num_pins;ipin++) 
 {
		inode = block_pin_to_tnode[iblk][ipin];

		if (inode != OPEN) 
		{   /* Pin is used -> put in graph */
			//if (is_opin_dsp (ipin)) 
			if (constraint[block[iblk].module_index].is_outpin[ipin])			
			{
					build_block_output_tnode (inode, iblk, ipin, block_pin_to_tnode);
					tnode_descript[inode].type = EMB_OPIN;
					num_output_nodes ++;
			}
			else 
			{    /* EMB ipin */
					num_edges = 1;
					tnode[inode].num_edges = num_edges;

					tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
						sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
						&tedge_ch_next_avail);

					tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0];
					//Seq in is not registered 					 
					//tnode[inode].out_edges[0].Tdel = constraint[block[iblk].module_index].T_emb.T_seq_in + constraint[block[iblk].module_index].T_emb.T_comb;

					//--------------------------
					inet = block[iblk].nets[ipin];
					if (!is_global[inet]) 
					{
						tnode[inode].out_edges[0].Tdel = constraint[block[iblk].module_index].T_emb.T_seq_in + constraint[block[iblk].module_index].T_emb.T_comb + constraint[block[iblk].module_index].pin_del[ipin];
						//printf("EMB internal delay %g\n", tnode[inode].out_edges[0].Tdel);
					}
					else
					{
						tnode[inode].out_edges[0].Tdel = 0;
					}
					//-------------------------
					tnode_descript[inode].type = EMB_IPIN;
			}
			tnode_descript[inode].ipin = ipin;
			tnode_descript[inode].isubblk = 0;
			tnode_descript[inode].iblk = iblk;
		}
 }

 /* EMB subblock in pin */

 inode = sub_pin_to_tnode[0][0];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = EMB_SUB_IPIN;


	if ((block[iblk].nets[clk_pin] == OPEN) ) 
	{
			tnode[inode].num_edges = 1;
			tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (1 * 
						sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
						&tedge_ch_next_avail);
	}
	else
	{
			tnode[inode].num_edges = 0;
			tnode[inode].out_edges = NULL;
	}





 /* EMB subblock out pin */

 inode = sub_pin_to_tnode[0][1];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = EMB_SUB_OPIN;

 tnode[inode].num_edges = num_output_nodes;

 //printf ("num_output_nodes %d \n", num_output_nodes);
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

 cnt = 0;
 //for (ipin=0;ipin<pins_per_dsp;ipin++) { 
	for (ipin=0;ipin<constraint[block[iblk].module_index].num_pins;ipin++) 
	{
		to_node = block_pin_to_tnode[iblk][ipin];
		if (to_node != OPEN) 
		{   /* Pin is used -> put in graph */
			//if (is_opin_dsp (ipin)) 
			if (constraint[block[iblk].module_index].is_outpin[ipin] == 1)			
			{
				tnode[inode].out_edges[cnt].to_node = to_node;
				//tnode[inode].out_edges[cnt].Tdel = dsp_timing_inf->T_seq_out;
				tnode[inode].out_edges[cnt].Tdel = constraint[block[iblk].module_index].T_emb.T_seq_out;
				cnt++;
			}
		}
	}

 /* Now do the stuff inside the EMB block */

	if ((block[iblk].nets[clk_pin] == OPEN) ) 
	{
			// If it is purely combinational 
		inode = sub_pin_to_tnode[0][0];
		to_node = sub_pin_to_tnode[0][1];
		tnode[inode].out_edges[0].to_node = to_node;
		//tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;
		//tnode[inode].out_edges[0].Tdel = constraint[block[iblk].module_index].T_emb.T_comb;
		tnode[inode].out_edges[0].Tdel = 0;
	}


/*
 if ((block[iblk].nets[clk_pin] == OPEN) ) {

    // If it is purely combinational 

    inode = sub_pin_to_tnode[0][0];
    to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;

 } else {

     //If there is an output flip-flop but not an input flip-flop 

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = EMB_FF2_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = EMB_FF2_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = block_pin_to_tnode[iblk][clk_pin];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

 }
*/
}




static void build_dsp_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *dsp_timing_inf) {

/* This routine builds the tnodes corresponding to the dsp pins of this     *
 * block, and properly hooks them up to the rest of the graph.  */

 int ipin, iedge, cnt, inode, to_node, num_edges;
 int clk_pin1, clk_pin2;
 int num_output_nodes;

 clk_pin2 = pins_per_dsp-3;
 //clk_pin1 = pins_per_dsp-2;
// clk_pin2 = pins_per_dsp-1;
 num_output_nodes = 0;

 /* Start by allocating the edge arrays, and for opins, loading them.    */

 /* DSP pins */

 for (ipin=0;ipin<pins_per_dsp;ipin++) {
    inode = block_pin_to_tnode[iblk][ipin];

    if (inode != OPEN) {   /* Pin is used -> put in graph */
       if (is_opin_dsp (ipin)) {
          build_block_output_tnode (inode, iblk, ipin, block_pin_to_tnode);
          tnode_descript[inode].type = DSP_OPIN;
          num_output_nodes ++;
       }

       else {    /* DSP ipin */
          num_edges = 1;
          tnode[inode].num_edges = num_edges;

          tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

          tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0];
          tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_seq_in;

          tnode_descript[inode].type = DSP_IPIN;
       }
       tnode_descript[inode].ipin = ipin;
       tnode_descript[inode].isubblk = 0;
       tnode_descript[inode].iblk = iblk;
    }
 }

 /* DSP subblock in pin */

 inode = sub_pin_to_tnode[0][0];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = DSP_SUB_IPIN;

 tnode[inode].num_edges = 1;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (1 * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

 /* DSP subblock out pin */

 inode = sub_pin_to_tnode[0][1];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = DSP_SUB_OPIN;

 tnode[inode].num_edges = num_output_nodes;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

 cnt = 0;
 for (ipin=0;ipin<pins_per_dsp;ipin++) {
    to_node = block_pin_to_tnode[iblk][ipin];
    if (to_node != OPEN) {   /* Pin is used -> put in graph */
       if (is_opin_dsp (ipin)) {
           tnode[inode].out_edges[cnt].to_node = to_node;
           tnode[inode].out_edges[cnt].Tdel = dsp_timing_inf->T_seq_out;
           cnt++;
       }
    }
 }

 /* Now do the stuff inside the DSP block */

// if ((block[iblk].nets[clk_pin1] == OPEN) &&
  //   (block[iblk].nets[clk_pin2] == OPEN)) {
if (block[iblk].nets[clk_pin2] == OPEN)
{
    /* If it is purely combinational */

    inode = sub_pin_to_tnode[0][0];
    to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;

	/*
 } else if ((block[iblk].nets[clk_pin1] != OPEN) &&
            (block[iblk].nets[clk_pin2] != OPEN)) {

    // If there is both an input and output flip-flop 

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = 0.0;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF1_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF1_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0] + 4;
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;

    inode = block_pin_to_tnode[iblk][clk_pin1];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = sub_pin_to_tnode[0][0] + 4;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF2_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 5;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF2_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = block_pin_to_tnode[iblk][clk_pin2];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+5;
    tnode[inode].out_edges[0].Tdel = 0.0;

 } else if ((block[iblk].nets[clk_pin1] != OPEN) &&
            (block[iblk].nets[clk_pin2] == OPEN)) {

    // If there is an input flip-flop but not an output flip-flop 

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = 0.0;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF1_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF1_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;

    inode = block_pin_to_tnode[iblk][clk_pin1];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;
*/

 } else {

    // If there is an output flip-flop but not an input flip-flop 

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = dsp_timing_inf->T_comb;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF2_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = DSP_FF2_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = block_pin_to_tnode[iblk][clk_pin2];		//CLK
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

	inode = block_pin_to_tnode[iblk][clk_pin2 + 1];	//CE
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

	inode = block_pin_to_tnode[iblk][clk_pin2 + 2];	//R
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

 }

}






static void build_memory_tnodes (int iblk, int 
        **block_pin_to_tnode, int **sub_pin_to_tnode,
        t_T_subblock *memory_timing_inf) {

/* This routine builds the tnodes corresponding to the memory pins of this     *
 * block, and properly hooks them up to the rest of the graph.  */

 int ipin, iedge, cnt, inode, to_node, num_edges;
 int clk_pin1, clk_pin2;
 int num_output_nodes;

 clk_pin1 = pins_per_memory-2;
 clk_pin2 = pins_per_memory-1;
 num_output_nodes = 0;

 /* Start by allocating the edge arrays, and for opins, loading them.    */

 /* MEMORY pins */

 for (ipin=0;ipin<pins_per_memory;ipin++) {
    inode = block_pin_to_tnode[iblk][ipin];

    if (inode != OPEN) {   /* Pin is used -> put in graph */
       if (is_opin_memory (ipin)) {
          build_block_output_tnode (inode, iblk, ipin, block_pin_to_tnode);
          tnode_descript[inode].type = MEMORY_OPIN;
          num_output_nodes ++;
       }

       else {    /* MEMORY ipin */
          num_edges = 1;
          tnode[inode].num_edges = num_edges;

          tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

          tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0];
          tnode[inode].out_edges[0].Tdel = memory_timing_inf->T_seq_in;

          tnode_descript[inode].type = MEMORY_IPIN;
       }
       tnode_descript[inode].ipin = ipin;
       tnode_descript[inode].isubblk = 0;
       tnode_descript[inode].iblk = iblk;
    }
 }

 /* MEMORY subblock in pin */

 inode = sub_pin_to_tnode[0][0];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = MEMORY_SUB_IPIN;

 tnode[inode].num_edges = 1;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (1 * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

 /* MEMORY subblock out pin */

 inode = sub_pin_to_tnode[0][1];
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = 0;
 tnode_descript[inode].iblk = iblk;
 tnode_descript[inode].type = MEMORY_SUB_OPIN;

 tnode[inode].num_edges = num_output_nodes;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);

 cnt = 0;
 for (ipin=0;ipin<pins_per_memory;ipin++) {
    to_node = block_pin_to_tnode[iblk][ipin];
    if (to_node != OPEN) {   /* Pin is used -> put in graph */
       if (is_opin_memory (ipin)) {
           tnode[inode].out_edges[cnt].to_node = to_node;
           tnode[inode].out_edges[cnt].Tdel = memory_timing_inf->T_seq_out;
           cnt++;
       }
    }
 }

 /* Now do the stuff inside the MEMORY block */

 if ((block[iblk].nets[clk_pin1] == OPEN) &&
     (block[iblk].nets[clk_pin2] == OPEN)) {

    /* If it is purely combinational */

    inode = sub_pin_to_tnode[0][0];
    to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = memory_timing_inf->T_comb;

 } else if ((block[iblk].nets[clk_pin1] != OPEN) &&
            (block[iblk].nets[clk_pin2] != OPEN)) {

    /* If there is both an input and output flip-flop */

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = 0.0;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF1_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF1_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0] + 4;
    tnode[inode].out_edges[0].Tdel = memory_timing_inf->T_comb;

    inode = block_pin_to_tnode[iblk][clk_pin1];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = sub_pin_to_tnode[0][0] + 4;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF2_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 5;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF2_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = block_pin_to_tnode[iblk][clk_pin2];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+5;
    tnode[inode].out_edges[0].Tdel = 0.0;

 } else if ((block[iblk].nets[clk_pin1] != OPEN) &&
            (block[iblk].nets[clk_pin2] == OPEN)) {

    /* If there is an input flip-flop but not an output flip-flop */

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = 0.0;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF1_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF1_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = memory_timing_inf->T_comb;

    inode = block_pin_to_tnode[iblk][clk_pin1];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

 } else {

    /* If there is an output flip-flop but not an input flip-flop */

    inode = sub_pin_to_tnode[0][0];
    to_node = inode + 2;
    tnode[inode].out_edges[0].to_node = to_node;
    tnode[inode].out_edges[0].Tdel = memory_timing_inf->T_comb;
 
    inode = sub_pin_to_tnode[0][0] + 2;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF2_SINK;
    tnode[inode].num_edges = 0;
    tnode[inode].out_edges = (t_tedge *) NULL;

    inode = sub_pin_to_tnode[0][0] + 3;
    tnode_descript[inode].ipin = 0;
    tnode_descript[inode].isubblk = 0;
    tnode_descript[inode].iblk = iblk;
    tnode_descript[inode].type = MEMORY_FF2_SOURCE;
    tnode[inode].num_edges = 1;
    tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_output_nodes * 
               sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
               &tedge_ch_next_avail);
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][1];
    tnode[inode].out_edges[0].Tdel = 0.0;

    inode = block_pin_to_tnode[iblk][clk_pin2];
    tnode[inode].out_edges[0].to_node = sub_pin_to_tnode[0][0]+3;
    tnode[inode].out_edges[0].Tdel = 0.0;

 }
}




static void build_block_output_tnode (int inode, int iblk, int ipin,
        int **block_pin_to_tnode) {

/* Sets the number of edges and the edge array for an output pin from a      *
 * block.  This pin must be hooked to something -- i.e. not OPEN.            */

 int iedge, to_blk, to_pin, to_node, num_edges, inet;
 int q_pin;
 t_tedge *tedge;
 
 inet = block[iblk].nets[ipin];          /* Won't be OPEN, as inode exists */
 
// printf("build_block_output_tnode , net name : %s\n", net[block[iblk].nets[ipin]].name);

 assert (inet != OPEN);                  /* Sanity check. */

 net_to_driver_tnode[inet] = inode;
 
 num_edges = net[inet].num_pins - 1;
// printf("inode : %d, num_edges : %d, ipin %d \n", inode, num_edges, ipin);
 tnode[inode].num_edges = num_edges;

 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
        sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail, 
        &tedge_ch_next_avail);

 tedge = tnode[inode].out_edges;

 //printf("build block output tnode: inode = %d, num_edges = %d \n", inode, num_edges);

 for (iedge=0;iedge<net[inet].num_pins-1;iedge++) {
    to_blk = net[inet].blocks[iedge+1];

    if ((block[to_blk].type == CLB) ||
        (block[to_blk].type == DSP) ||
		(block[to_blk].type == EMB) ||
        (block[to_blk].type == MEMORY)) {
       to_pin = net[inet].blk_pin[iedge+1];
    }
    else                  /* OUTPAD */
       to_pin = 0;

    to_node = block_pin_to_tnode[to_blk][to_pin];
    tedge[iedge].to_node = to_node;
	
   /* Set delay from net delays with a later call */
 }
}


static void build_subblock_tnodes (
          t_subblock_data subblock_data,
          int **n_uses_of_sblk_opin, 
          int *blk_pin_to_tnode, int **sub_pin_to_tnode, 
          int subblock_lut_size, int num_subs,
		  int subblock_num_pin,
		  int subblock_num_opin, int subblock_num_ipin,
		  float T_FX,
          float T_sblk_opin_to_sblk_ipin, float T_sblk_opin_to_clb_opin, 		  
          float T_sblk_cout_to_sblk_cin, float T_clb_cout_to_clb_cin,
          t_T_subblock *T_subblock, int **next_sblk_opin_edge, int iblk,
          int **sblk_cin_pin_to_tnode,
          int **sblk_cout_pin_to_tnode) {

/* This routine builds the tnodes of the subblock pins within one CLB. Note *
 * that only the block_pin_to_tnode, etc. data for *this* block are passed  *
 * in.                                                                      */

 int isub, ipin, inode, to_node, from_pin, to_pin, out_pin, clk_pin;
 int iedge, num_edges;
 int cin_exists, cout_exists;
 float ipin_to_sink_Tdel;
 t_tedge *tedge;
 boolean has_inputs;
 t_subblock *sub_inf;
 
 int opin;
 int allopen;
 int q_pin;

 int internal_pin;
 int from_subblock;

 sub_inf = subblock_data.subblock_inf[iblk];
 
 out_pin = subblock_lut_size;
 //clk_pin = subblock_lut_size + 1;
 clk_pin = subblock_num_ipin + subblock_num_opin;


/* Allocate memory for output pins first. */
//printf("create nodes for outpin of each subblock, subblock_num_opin %d\n", subblock_num_opin);
//for each subblock and each outpin
	for (isub=0;isub<num_subs;isub++) 
	{
		for (opin= 0; opin < subblock_num_opin; opin ++)
		{
			//get the inode number
			inode = sub_pin_to_tnode[isub][subblock_num_ipin + opin];
			//printf("create nodes for outpin of each subblock, block number: %d, inode : %d, opin : %d, to_pin : %d\n", isub, inode, opin, sub_inf[isub].outputs[opin] );
			  if (inode != OPEN) 
			  { 	    /* Output is used -> timing node exists. */
    	   		next_sblk_opin_edge[isub][opin] = 0;     /* Reset */
       			num_edges = n_uses_of_sblk_opin[isub][opin];	
       			tnode[inode].num_edges = num_edges;
			
	//			printf("isub %d, inode %d, num_edges %d, opin %d\n", isub, inode, num_edges, opin);
	
    	   		tnode[inode].out_edges = (t_tedge *) my_chunk_malloc (num_edges * 
        	    		sizeof (t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail,
            			&tedge_ch_next_avail);
    
				if (sub_inf[isub].carryout_cluster != 1)
				{
					//normal output pin
					tnode_descript[inode].type = SUBBLK_OPIN;
					tnode_descript[inode].ipin = opin;
				}
				else
				{
					//it is carry out 
					tnode_descript[inode].type = SUBBLK_COUT;
					tnode_descript[inode].ipin = -1;
				}
       		
	    		tnode_descript[inode].isubblk = isub;
	        	tnode_descript[inode].iblk = iblk;			
				//printf("\t\t num edge: %d, out_edge %d\n", tnode[inode].num_edges,tnode[inode].out_edges );
   			}			
		}	
 	}

// Load the output pins edge arrays. 
//for each input pin , check the internal connection
// printf("Load output pins edge array\n");
	for (isub=0;isub<num_subs;isub++) 
	{
 		for (ipin = 0; ipin < subblock_num_ipin; ipin ++ )
		{
			//printf("Load output pins edge array, load inputs\n");
			//from which input of the subblock
			from_pin = sub_inf[isub].inputs[ipin];
//			printf("block_name %s, iblk %d, isub %d, from_pin %d, ipin %d, subblock_num_ipin + from_pin - pins_per_clb = %d\n", block[iblk].name, iblk, isub, from_pin, ipin, subblock_num_ipin + from_pin - pins_per_clb);
	//			printf("block_name %s, iblk %d, isub %d, from_pin %d, ipin %d, pins_per_clb = %d\n", block[iblk].name, iblk, isub, from_pin, ipin, pins_per_clb);
	
			// Not OPEN and comes from local subblock output?
			if (from_pin >= pins_per_clb) 
			{
				from_subblock = (from_pin - pins_per_clb)/subblock_num_pin;
				internal_pin = (from_pin - pins_per_clb) % subblock_num_pin ;
			
				switch (ipin)
				{
					case I_CIN:  //CIN
							//get the output pin inode from the other subblock
							//internal connection of COUT to CIN

							//opin = get_opin(from_pin-pins_per_clb);
							opin = get_opin(internal_pin);
													
							//no need to handle isub-1 = -1, since first block doesn't have internal CIN
							//inode = sub_pin_to_tnode[isub - 1][subblock_num_ipin + opin]; // get the previous COUT node
							inode = sub_pin_to_tnode[from_subblock][subblock_num_ipin + opin]; // get the previous COUT node
												
							to_node = sub_pin_to_tnode[isub][ipin];						
							tedge = tnode[inode].out_edges;
							iedge = next_sblk_opin_edge[from_subblock][opin]++;
							tedge[iedge].to_node = to_node;

							tedge[iedge].Tdel = T_sblk_cout_to_sblk_cin;				
						break;
//					case I_SHIFTIN:
					default:

							//opin = get_opin(from_pin-pins_per_clb);
							opin = get_opin(internal_pin);

	//						printf("iblk %d, isub %d, opin  %d, subblock_num_ipin %d\n", iblk, isub, opin, subblock_num_ipin);
							//shift in is from the upper subblock

							inode = sub_pin_to_tnode[from_subblock][subblock_num_ipin + opin]; //subblocks in a slice for virtex2
						
							//printf("inode %d \n", inode);
							to_node = sub_pin_to_tnode[isub][ipin];
							//printf ("from block isub %d , inode %d, opin %d,  from_pin %d\n", 1-isub, inode, opin, from_pin);
							tedge = tnode[inode].out_edges;
							iedge = next_sblk_opin_edge[from_subblock][opin]++;
							tedge[iedge].to_node = to_node;
							tedge[iedge].Tdel = T_sblk_opin_to_sblk_ipin;
						

						break;
						/*
					case I_DY:	//the DX/DY pin
							//nothing connect to DX/DY pin
							//internal connection from X/Y to DX/DY
							//get the output pin inode from the same subblock
					//		opin = 4;	//opin for X/Y
					//		inode = sub_pin_to_tnode[isub][subblock_num_ipin + opin];
										
					//		to_node = sub_pin_to_tnode[isub][ipin];	//to this input pin DX/DY				
					//		printf ("output to input internal connection X/Y to DX/DY\n ");
					//		printf ("from block isub %d , inode %d, opin %d,  from_pin %d\n", isub, inode, opin, from_pin);
					//		printf ("to block isub %d, to_node %d, ipin %d \n", isub, to_node, ipin);
					//		tedge = tnode[inode].out_edges;
					//		iedge = next_sblk_opin_edge[isub][opin]++;
					//		tedge[iedge].to_node = to_node;
					//		tedge[iedge].Tdel = 0;	//no time delay for this connection
							break;				
					default:
						//internal connection
						//get the output pin inode from the other subblock
						
						break;
						*/
				}			
			}			
		}
	

	//	printf("Load output pins edge array, output to clb, clb name: %s, block num : %d, subblock num : %d , subblock_num_opin : %d, subblock_num_ipin %d\n", block[iblk].name, iblk, isub, subblock_num_opin, subblock_num_ipin);	
		for (opin = 0 ; opin < subblock_num_opin; opin ++)
		{
			to_pin = sub_inf[isub].outputs[opin];
			
			//printf("to_pin : %d, opin : %d , pins_per_clb : %d \n", to_pin, opin, pins_per_clb);
			//ignore the internal connection to another subblock input, added before
			if (to_pin < pins_per_clb)
			{	
				if (to_pin != OPEN) 
				{   //  sblk opin goes to clb opin? 
							
			
					// Check that CLB pin connects to something ->     [subblock_num_ipin + opin
					// not just a mandatory BLE to CLB opin connection 
					if (block[iblk].nets[to_pin] != OPEN) 
					{
						//printf("net name: %s \n", net[block[iblk].nets[to_pin]].name);
						to_node = blk_pin_to_tnode[to_pin];					
						inode = sub_pin_to_tnode[isub][subblock_num_ipin + opin];					
						tedge = tnode[inode].out_edges;					
						iedge = next_sblk_opin_edge[isub][opin]++;					
						tedge[iedge].to_node = to_node;

						if (opin == O_COUT)	//subblock cout to clb cout					
						{
							if (block[iblk].carry_head)
							{
								//printf("carry head , Tdel = %g \n", T_clb_cout_to_clb_cin);
								tedge[iedge].Tdel = T_clb_cout_to_clb_cin;
							}
							else
							{
								tedge[iedge].Tdel = T_sblk_opin_to_clb_opin;
							}
						}
						else
						{
							tedge[iedge].Tdel = T_sblk_opin_to_clb_opin;
						}
//						printf("to_pin : %d, opin : %d , pins_per_clb : %d, inode %d, iedge %d, to_node %d \n", to_pin, opin, pins_per_clb, inode, iedge, to_node);
					}
				}	
			}	
		}				

	 }

for (isub=0;isub<num_subs;isub++) 
{
//	printf("Build and hook up subblock inputs, isub %d\n", isub);
 	has_inputs = FALSE;	//check any subblock input
	

	//Hook the subblock input
	//for (ipin=0;ipin<subblock_num_ipin - 1;ipin++) //don't count the last clk pin
	for (ipin=0;ipin<subblock_num_ipin ;ipin++) //don't count the last clk pin
	{	
		//get the inode number of this input
		inode = sub_pin_to_tnode[isub][ipin];		
		if (inode != OPEN) 
		{    // tnode exists -> pin is used 
          
			has_inputs = TRUE;
			
			//get the node this ipin connect to 
			switch (ipin)
			{
				case I_SHIFTIN:
					//SHIFTIN
					opin = O_SHIFTOUT;	//connect to SHIFTOUT
					to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];
					ipin_to_sink_Tdel = 0;
					break;
				case I_CIN:
					//CIN
					opin = O_COUT;	//connect to COUT
					to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];						
					//ipin_to_sink_Tdel = T_clb_cout_to_clb_cin;	//time delay from others cout to this cin					
					ipin_to_sink_Tdel = T_sblk_opin_to_sblk_ipin;
					break;							
				case I_BY:
					//BY/BX
//					opin = 5;	//connect to XQ/YQ
					to_node = OPEN;
//					ipin_to_sink_Tdel = T_subblock[isub].T_seq_in;
					break;
				case I_DY:
					//DX/DY
					opin = O_YQ;	//connect to XQ/YQ
					to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];	
//					ipin_to_sink_Tdel = T_subblock[isub].T_seq_in;
					ipin_to_sink_Tdel = T_subblock[isub].T_seq_in + T_subblock[isub].T_seq_out;

					break;
				case I_CE:
					//CE
					opin = O_YQ;	//connect to XQ/YQ
					//to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];	
					to_node = OPEN;	
					ipin_to_sink_Tdel = 0;
					break;
				case I_SR:
					//SR
					opin = O_YQ;	//connect to XQ/YQ
					//to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];	
					to_node = OPEN;	
					ipin_to_sink_Tdel = 0;
					break;
				case I_CLK:
					//clk , not check here
					opin = O_YQ;	//connect to XQ/YQ
					to_node = OPEN;	
					ipin_to_sink_Tdel = 0;
				case I_FXINA:
				case I_FXINB:
					//handle FX
					opin = O_FX;	//connect to X/Y
					to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];	
					ipin_to_sink_Tdel = T_FX;
				default:
					if ((ipin>= I_LUT_IN )&& (ipin <  I_LUT_IN + subblock_lut_size))
					{				
						//LUT input					
						opin = O_Y;	//connect to X/Y
						to_node = sub_pin_to_tnode[isub][subblock_num_ipin + opin];	
						ipin_to_sink_Tdel = T_subblock[isub].T_comb;
					}
					break;	

					
			}
	
			//if the output pin
			if (to_node != OPEN)
			{
//				printf("From node %d, To node %d \n", inode , to_node);
				tnode[inode].num_edges = 1;
				tnode[inode].out_edges = (t_tedge *) my_chunk_malloc ( sizeof 
						(t_tedge), &tedge_ch_list_head, &tedge_ch_bytes_avail, 
						&tedge_ch_next_avail);
				tedge = tnode[inode].out_edges;
				tedge[0].to_node = to_node;
				tedge[0].Tdel = ipin_to_sink_Tdel;

				//printf("tedge[0].Tdel %f \n", tedge[0].Tdel);
				
				tnode_descript[inode].type = SUBBLK_IPIN;
				tnode_descript[inode].ipin = ipin;
				tnode_descript[inode].isubblk = isub;
				tnode_descript[inode].iblk = iblk;
			}
			else
			{
				tnode[inode].num_edges = 0;
				tnode[inode].out_edges = NULL;
				tedge = tnode[inode].out_edges;

				tnode_descript[inode].type = SUBBLK_IPIN;
				tnode_descript[inode].ipin = ipin;
				tnode_descript[inode].isubblk = isub;
				tnode_descript[inode].iblk = iblk;
			}
       }
	}
}	//end each subblock ipin


	#define T_CONSTANT_GENERATOR -1000   /* Essentially -ve infinity */
	//no constant generator
}


static void build_ipad_tnodes (int iblk, int **block_pin_to_tnode, float
            T_ipad, int *num_subblocks_per_block, t_subblock **subblock_inf) {

/* Builds the two tnodes corresponding to an input pad, and hooks them into *
 * the timing graph.                                                        */

 int to_node, inode;
 t_tedge *tedge;

/* First node: input node to the pad -> nothing comes into this.  Second    *
 * node is the output node of the pad, that has edges to CLB ipins.         */

 inode = block_pin_to_tnode[iblk][0];
 to_node = block_pin_to_tnode[iblk][1];

 tnode[inode].num_edges = 1;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc ( sizeof (t_tedge),
        &tedge_ch_list_head, &tedge_ch_bytes_avail, &tedge_ch_next_avail);
 tedge = tnode[inode].out_edges;
 tedge[0].to_node =  to_node; 

/* By definition, global clocks from pads arrive at T = 0.  The earliest any *
 * clock can arrive at any flip flop is T = 0, and the fastest global clock  *
 * from a pad should have zero delay on its edges so it does get to the      *
 * flip flop clock pin at T = 0.                                             */

 if (is_global_clock (iblk, num_subblocks_per_block, subblock_inf))   
 {
	 printf("inode %d is global\n", inode);
	tedge[0].Tdel = 0.;
 }    
 else 
 {
	// printf("IPAD iblk %d (%s) , node %d \n", iblk , block[iblk].name, inode);
    tedge[0].Tdel = T_ipad;
 }

 tnode_descript[inode].type = INPAD_SOURCE;
 tnode_descript[inode].ipin = OPEN;
 tnode_descript[inode].isubblk = OPEN;
 tnode_descript[inode].iblk = iblk;

/* Now do pad output. */

 build_block_output_tnode (to_node, iblk, 0, block_pin_to_tnode);

 tnode_descript[to_node].type = INPAD_OPIN;
 //tnode_descript[to_node].type = INPAD_OPIN;
 tnode_descript[to_node].ipin = 0;
 tnode_descript[to_node].isubblk = OPEN;
 tnode_descript[to_node].iblk = iblk;
}


static boolean is_global_clock (int iblk, int *num_subblocks_per_block,
     t_subblock **subblock_inf) {

/* Returns TRUE if the net driven by this block (which must be an INPAD) is  *
 * (1) a global signal, and (2) used as a clock input to at least one block. */

 int inet, ipin, to_blk, to_pin, isub;
 
 inet = block[iblk].nets[0];

 if (!is_global[inet]) 
    return (FALSE);

 for (ipin=1;ipin<net[inet].num_pins;ipin++) {
    to_blk = net[inet].blocks[ipin];
    to_pin = net[inet].blk_pin[ipin];

    for (isub=0;isub<num_subblocks_per_block[to_blk];isub++) {
       if (subblock_inf[to_blk][isub].clock == to_pin)
          return (TRUE);
    }
 }

 return (FALSE);
}


static void build_opad_tnodes (int *blk_pin_to_tnode, float T_opad, int iblk, float T_clk_opad) {

/* Builds the two tnodes in an output pad and connects them up.   */

 int to_node, inode;
 int outpad_incount;

 int clk0_inode, clk1_inode;
 t_tedge *tedge;
 

 to_node = blk_pin_to_tnode[7];  //output pin of the outpad
/* First node: input node to the pad -> will be driven some net.  Second    *
 * node is the output node of the pad, which connects to nothing.           */

 clk0_inode = blk_pin_to_tnode[3];
 clk1_inode = blk_pin_to_tnode[6];

if (to_node != OPEN)
{
	//the output node exist 
	for (outpad_incount = 0; outpad_incount < 7 ; outpad_incount ++)
	{ 	
		if (block[iblk].nets[outpad_incount] != OPEN)
		{
			//the input pin is connected to an edge, get the inode number 
			inode = blk_pin_to_tnode[outpad_incount];
			
			//set the num of edges to 1 (connect to 1 output only)
			tnode[inode].num_edges = 1;
			tnode[inode].out_edges = (t_tedge *) my_chunk_malloc ( sizeof (t_tedge),
					&tedge_ch_list_head, &tedge_ch_bytes_avail, &tedge_ch_next_avail);
			tedge = tnode[inode].out_edges;
			tedge[0].to_node =  to_node; 
			
			tedge[0].Tdel = T_opad;
			if ( (outpad_incount == 0) || (outpad_incount ==1))
			{
				//input pin of outpad
				if ( (clk0_inode != OPEN) || ((clk1_inode) != OPEN))
				{
					//clocked output
					tedge[0].Tdel = T_clk_opad;
				}
			}

			if ( outpad_incount >1)
			{
				//OCE, OCLK1, SR, REV, OCLK2								
				//no delay
				tedge[0].Tdel = 0;
				
			}
			

			//printf("output pad, inode %d(%s), to_node %d, num_edges %d \n", inode, block[iblk].name, to_node, tnode[inode].num_edges);
			
			tnode_descript[inode].type = OUTPAD_IPIN;
			tnode_descript[inode].ipin = 0;
			tnode_descript[inode].isubblk = OPEN;
			tnode_descript[inode].iblk = iblk;
					
		}
		else	
		{
			//the input pin is open, no node should be set		
		}	
	}

		//the destination node doesn't have edge connect to , set to 0
		tnode[to_node].num_edges = 0;
		tnode[to_node].out_edges = NULL;
			
		tnode_descript[to_node].type = OUTPAD_SINK;
		tnode_descript[to_node].ipin = OPEN;
		tnode_descript[to_node].isubblk = OPEN;
		tnode_descript[to_node].iblk = iblk;
}
  
 //inode = blk_pin_to_tnode[0];
 //to_node = blk_pin_to_tnode[1];
 
/*
 tnode[inode].num_edges = 1;
 tnode[inode].out_edges = (t_tedge *) my_chunk_malloc ( sizeof (t_tedge),
        &tedge_ch_list_head, &tedge_ch_bytes_avail, &tedge_ch_next_avail);
 tedge = tnode[inode].out_edges;
 tedge[0].to_node =  to_node; 
 tedge[0].Tdel = T_opad;

 tnode_descript[inode].type = OUTPAD_IPIN;
 tnode_descript[inode].ipin = 0;
 tnode_descript[inode].isubblk = OPEN;
 tnode_descript[inode].iblk = iblk;

 tnode[to_node].num_edges = 0;
 tnode[to_node].out_edges = NULL;

 tnode_descript[to_node].type = OUTPAD_SINK;
 tnode_descript[to_node].ipin = OPEN;
 tnode_descript[to_node].isubblk = OPEN;
 tnode_descript[to_node].iblk = iblk;
 */
}


void load_timing_graph_net_delays (float **net_delay) {

/* Sets the delays of the inter-CLB nets to the values specified by          *
 * net_delay[0..num_nets-1][1..num_pins-1].  These net delays should have    *
 * been allocated and loaded with the net_delay routines.  This routine      *
 * marks the corresponding edges in the timing graph with the proper delay.  */

 int inet, ipin, inode;
 t_tedge *tedge;

// printf("load_timing_graph_net_delays \n");
 for (inet=0;inet<num_nets;inet++) {
    inode = net_to_driver_tnode[inet];
    tedge = tnode[inode].out_edges;

/* Note that the edges of a tnode corresponding to a CLB or INPAD opin must  *
 * be in the same order as the pins of the net driven by the tnode.          */

    for (ipin=1;ipin<net[inet].num_pins;ipin++) {
		
       tedge[ipin-1].Tdel = net_delay[inet][ipin];
	   //printf("inode %d, tedge[%d].Tdel = %g, %g , inet=%d (%s)\n", inode, ipin-1, tedge[ipin-1].Tdel,  net_delay[inet][ipin], inet, net[inet].name);	   
    }
 }
}


void free_timing_graph (float **net_slack) {

/* Frees the timing graph data. */
//printf("free_timing_graph \n");
 if (tedge_ch_list_head == NULL) {
    printf ("Error in free_timing_graph: No timing graph to free.\n");
    exit (1);
 }

 free_chunk_memory (tedge_ch_list_head);
 free (tnode);
 free (tnode_descript);
 free (net_to_driver_tnode);
 free_ivec_vector (tnodes_at_level, 0, num_tnode_levels - 1);
 free (net_slack);
 
 tedge_ch_list_head = NULL;
 tedge_ch_bytes_avail = 0;
 tedge_ch_next_avail = NULL;

 tnode = NULL;
 tnode_descript = NULL;
 num_tnodes = 0;
 net_to_driver_tnode = NULL;
 tnodes_at_level = NULL;
 num_tnode_levels = 0;
}


void print_net_slack (char *fname, float **net_slack) {

/* Prints the net slacks into a file.                                     */

 int inet, ipin;
 FILE *fp;

 fp = my_fopen (fname, "w", 0);

 fprintf (fp, "Net #\tSlacks\n\n");

 for (inet=0;inet<num_nets;inet++) {
    fprintf (fp, "%5d", inet);
    for (ipin=1;ipin<net[inet].num_pins;ipin++) {
       fprintf (fp, "\t%g", net_slack[inet][ipin]);
    }
    fprintf (fp, "\n");
 }
}


void print_timing_graph (char *fname) {

/* Prints the timing graph into a file.           */

 FILE *fp;
 int inode, iedge, ilevel, i;
 t_tedge *tedge;
 t_tnode_type itype;
 static char *tnode_type_names[] = {"INPAD_SOURCE", "INPAD_OPIN", 
           "OUTPAD_IPIN", "OUTPAD_SINK", "CLB_IPIN", "CLB_OPIN", 
           "SUBBLK_IPIN", "SUBBLK_OPIN", "FF_SINK", "FF_SOURCE",
           "CONSTANT_GEN_SOURCE",
           "DSP_IPIN", "DSP_OPIN","DSP_SUB_IPIN", "DSP_SUB_OPIN", "DSP_FF1_SINK", "DSP_FF2_SINK", "DSP_FF1_SOURCE", "DSP_FF2_SOURCE",
           "MEMORY_IPIN", "MEMORY_OPIN","MEMORY_SUB_IPIN", "MEMORY_SUB_OPIN", "MEMORY_FF1_SINK", "MEMORY_FF2_SINK", "MEMORY_FF1_SOURCE", "MEMORY_FF2_SOURCE", "SUBBLK_COUT", "SUBBLK_CIN",
		   "EMB_IPIN", "EMB_OPIN","EMB_SUB_IPIN", "EMB_SUB_OPIN", "EMB_FF1_SINK", "EMB_FF2_SINK", "EMB_FF1_SOURCE", "EMB_FF2_SOURCE"
 };

 fp = my_fopen (fname, "w", 0);
	//printf ("num_tnodes: %d\n", num_tnodes);
 fprintf (fp, "num_tnodes: %d\n", num_tnodes);

 /*
 printf ("Node #\tType\t\tipin\tisubblk\tiblk\t# edges\t"
              "Edges (to_node, Tdel)\n\n");
			  */
 fprintf (fp, "Node #\tType\t\tipin\tisubblk\tiblk\t# edges\t"
              "Edges (to_node, Tdel)\n\n");

 for (inode=0;inode<num_tnodes;inode++) {
	 //printf ("%d\t", inode);
    fprintf (fp, "%d\t", inode);

    itype = tnode_descript[inode].type;
	//printf ("%-15.15s\t", tnode_type_names[itype]);
    fprintf (fp, "%-15.15s\t", tnode_type_names[itype]);

	/*
	printf ("%d\t%d\t%d\t", tnode_descript[inode].ipin, 
             tnode_descript[inode].isubblk, tnode_descript[inode].iblk);
*/
    fprintf (fp, "%d\t%d\t%d\t", tnode_descript[inode].ipin, 
             tnode_descript[inode].isubblk, tnode_descript[inode].iblk);

	//printf ("%d\t", tnode[inode].num_edges);

    fprintf (fp, "%d\t", tnode[inode].num_edges);
    tedge = tnode[inode].out_edges;
    for (iedge=0;iedge<tnode[inode].num_edges;iedge++) {
		//printf ("\t(%4d\t%7.3g)", tedge[iedge].to_node, tedge[iedge].Tdel);
       fprintf (fp, "\t(%4d\t%7.3g)", tedge[iedge].to_node, tedge[iedge].Tdel);
    }
	//printf ("\n");
    fprintf (fp, "\n");
 }

 //printf ("\n\nnum_tnode_levels: %d\n", num_tnode_levels);
 fprintf (fp, "\n\nnum_tnode_levels: %d\n", num_tnode_levels);

 for (ilevel=0;ilevel<num_tnode_levels;ilevel++) {
    fprintf (fp, "\n\nLevel: %d  Num_nodes: %d\nNodes:", ilevel, 
          tnodes_at_level[ilevel].nelem);
    for (i=0;i<tnodes_at_level[ilevel].nelem;i++) 
       fprintf (fp, "\t%d", tnodes_at_level[ilevel].list[i]);
 }
 
 fprintf (fp, "\n");
 fprintf (fp, "\n\nNet #\tNet_to_driver_tnode\n");
 
 for (i=0;i<num_nets;i++) 
    fprintf (fp, "%4d\t%6d\n", i, net_to_driver_tnode[i]);

 fprintf (fp, "\n\nNode #\t\tT_arr\t\tT_req\n\n");

 for (inode=0;inode<num_tnodes;inode++) 
    fprintf (fp, "%d\t%12g\t%12g\n", inode, tnode[inode].T_arr, 
            tnode[inode].T_req);
 
 fclose (fp);
}


float load_net_slack (float **net_slack, float target_cycle_time) {

/* Determines the slack of every source-sink pair of block pins in the      *
 * circuit.  The timing graph must have already been built.  target_cycle_  *
 * time is the target delay for the circuit -- if 0, the target_cycle_time  *
 * is set to the critical path found in the timing graph.  This routine     *
 * loads net_slack, and returns the current critical path delay.            */
 
 float T_crit, T_arr, Tdel, T_cycle, T_req;
 int inode, ilevel, num_at_level, i, num_edges, iedge, to_node;
 t_tedge *tedge;
 
/* Reset all arrival times to -ve infinity.  Can't just set to zero or the   *
 * constant propagation (constant generators work at -ve infinity) won't     *
 * work.      
 */

 //printf("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
//printf("load_net_slack\n");
 for (inode=0;inode<num_tnodes;inode++) 
    tnode[inode].T_arr = T_CONSTANT_GENERATOR;
	//tnode[inode].T_arr = 0;
 
/* Compute all arrival times with a breadth-first analysis from inputs to   *
 * outputs.  Also compute critical path (T_crit).                           */

 T_crit = 0.;

/* Primary inputs arrive at T = 0. */

 num_at_level = tnodes_at_level[0].nelem;
 for (i=0;i<num_at_level;i++) {
    inode = tnodes_at_level[0].list[i];
    tnode[inode].T_arr = 0.;
 }

 for (ilevel=0;ilevel<num_tnode_levels;ilevel++) 
 {
    num_at_level = tnodes_at_level[ilevel].nelem;
//printf("======================================================\n");
//printf("level = %d \n" ,ilevel);

    for (i=0;i<num_at_level;i++) {				
       inode = tnodes_at_level[ilevel].list[i];	   
		
		
       T_arr = tnode[inode].T_arr;
	   
       num_edges = tnode[inode].num_edges;
       tedge = tnode[inode].out_edges;
       T_crit = max (T_crit, T_arr);
		//printf(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
	   //printf("\n");
	  //printf("inode = %d,  T_arr = %g, T_crit = %g\n", inode, T_arr, T_crit);
       
       for (iedge=0;iedge<num_edges;iedge++) {
          to_node = tedge[iedge].to_node;
          Tdel = tedge[iedge].Tdel; 
          tnode[to_node].T_arr = max (tnode[to_node].T_arr, T_arr + Tdel);

		 // printf("\t edge = %d, to_node  = %d Tdel = %g , tnode[to_node].T_arr %g\n ", iedge, to_node, Tdel, tnode[to_node].T_arr);
		    
       }
    }
 }


 if (target_cycle_time > 0.)     /* User specified target cycle time */
    T_cycle = target_cycle_time;
 else                            /* Otherwise, target = critical path */
    T_cycle = T_crit;
 
	//printf("T_cycle = %g \n", T_cycle);

/* Compute the required arrival times with a backward breadth-first analysis *
 * from sinks (output pads, etc.) to primary inputs.                         */

 for (ilevel=num_tnode_levels-1;ilevel>=0;ilevel--) {
    num_at_level = tnodes_at_level[ilevel].nelem;
 
    for (i=0;i<num_at_level;i++) {
       inode = tnodes_at_level[ilevel].list[i];
       num_edges = tnode[inode].num_edges;
 
       if (num_edges == 0) {   /* sink */
          tnode[inode].T_req = T_cycle;
       }
       else {
          tedge = tnode[inode].out_edges;
          to_node = tedge[0].to_node;
          Tdel = tedge[0].Tdel;

		 // printf("inode = %d, to_node %d,  T_req = %g , T_req - Tdel = %g\n", inode, to_node,T_req, T_req - Tdel);

          T_req = tnode[to_node].T_req - Tdel;

		  
          
          for (iedge=1;iedge<num_edges;iedge++) {
             to_node = tedge[iedge].to_node;
             Tdel = tedge[iedge].Tdel;

			//printf("inode = %d, to_node %d,  T_req = %g , T_req - Tdel = %g\n", inode, to_node, T_req, T_req - Tdel);
             T_req = min (T_req, tnode[to_node].T_req - Tdel);
          }
          
          tnode[inode].T_req = T_req;
			

       }
    }
 }

 compute_net_slacks (net_slack);

 return (T_crit);
}


static void compute_net_slacks (float **net_slack) {

/* Puts the slack of each source-sink pair of block pins in net_slack.     */

 int inet, iedge, inode, to_node, num_edges;
 t_tedge *tedge;
 float T_arr, Tdel, T_req;

 //printf("compute_net_slacks \n");

 for (inet=0;inet<num_nets;inet++) {
    inode = net_to_driver_tnode[inet];
    T_arr = tnode[inode].T_arr;
    num_edges = tnode[inode].num_edges;
    tedge = tnode[inode].out_edges;

    for (iedge=0;iedge<num_edges;iedge++) {
       to_node = tedge[iedge].to_node;
       Tdel = tedge[iedge].Tdel;
       T_req = tnode[to_node].T_req;
       net_slack[inet][iedge + 1] = T_req - T_arr - Tdel;
    }
 }
}


void print_critical_path (char *fname) {

/* Prints out the critical path to a file.  */

 t_linked_int *critical_path_head, *critical_path_node;
 FILE *fp;
 int non_global_nets_on_crit_path, global_nets_on_crit_path;
 int tnodes_on_crit_path, inode, iblk, inet;
 t_tnode_type type;
 float total_net_delay, total_logic_delay, Tdel;
	
// printf("allocate_and_load_critical_path\n");
 critical_path_head = allocate_and_load_critical_path ();
 critical_path_node = critical_path_head;

 fp = my_fopen (fname, "w", 0);
 
 non_global_nets_on_crit_path = 0;
 global_nets_on_crit_path = 0;
 tnodes_on_crit_path = 0;
 total_net_delay = 0.;
 total_logic_delay = 0.;

 while (critical_path_node != NULL) {
	 
    Tdel = print_critical_path_node (fp, critical_path_node);
    inode = critical_path_node->data;
    type = tnode_descript[inode].type;
    tnodes_on_crit_path++;

    if (type == INPAD_OPIN || type == CLB_OPIN || type ==EMB_OPIN || type == DSP_OPIN || type == MEMORY_OPIN) 
	{		
       get_tnode_block_and_output_net (inode, &iblk, &inet);

       if (!is_global[inet]) 
          non_global_nets_on_crit_path++;
       else 
          global_nets_on_crit_path++;

       total_net_delay += Tdel;
    }
    else {
       total_logic_delay += Tdel;
    }

    critical_path_node = critical_path_node->next;	
 }

 fprintf (fp, "\nTnodes on crit. path: %d  Non-global nets on crit. path: %d."
          "\n", tnodes_on_crit_path, non_global_nets_on_crit_path);
 fprintf (fp, "Global nets on crit. path: %d.\n", global_nets_on_crit_path);
 fprintf (fp, "Total logic delay: %g (s)  Total net delay: %g (s)\n", 
          total_logic_delay, total_net_delay);

 printf ("Nets on crit. path: %d normal, %d global.\n", 
          non_global_nets_on_crit_path, global_nets_on_crit_path);

 printf ("Total logic delay: %g (s)  Total net delay: %g (s)\n", 
          total_logic_delay, total_net_delay);

 fclose (fp);
 free_int_list (&critical_path_head);
}


t_linked_int *allocate_and_load_critical_path (void) {

/* Finds the critical path and puts a list of the tnodes on the critical    *
 * path in a linked list, from the path SOURCE to the path SINK.            */

 t_linked_int *critical_path_head, *curr_crit_node, *prev_crit_node;
 int inode, iedge, to_node, num_at_level, i, crit_node, num_edges;
 float min_slack, slack;
 t_tedge *tedge;

 num_at_level = tnodes_at_level[0].nelem;
 min_slack = HUGE_FLOAT;
 crit_node = OPEN;   /* Stops compiler warnings. */

 for (i=0;i<num_at_level;i++) {   /* Path must start at SOURCE (no inputs) */
    inode = tnodes_at_level[0].list[i];
    slack = tnode[inode].T_req - tnode[inode].T_arr;

    if (slack < min_slack) {
       crit_node = inode;
       min_slack = slack;
    }
 }
 
 critical_path_head = (t_linked_int *) my_malloc (sizeof (t_linked_int));
 critical_path_head->data = crit_node;
 prev_crit_node = critical_path_head;
 num_edges = tnode[crit_node].num_edges;
 
 while (num_edges != 0) {   /* Path will end at SINK (no fanout) */
    curr_crit_node = (t_linked_int *) my_malloc (sizeof (t_linked_int));
    prev_crit_node->next = curr_crit_node;
    tedge = tnode[crit_node].out_edges;
    min_slack = HUGE_FLOAT;
       
	for (iedge=0;iedge<num_edges;iedge++) {
       to_node = tedge[iedge].to_node;
       slack = tnode[to_node].T_req - tnode[to_node].T_arr;
 
       if (slack < min_slack) {
          crit_node = to_node;
          min_slack = slack;
       } 
    }

    curr_crit_node->data = crit_node;
    prev_crit_node = curr_crit_node;
    num_edges = tnode[crit_node].num_edges;
 }
 
 prev_crit_node->next = NULL;
 return (critical_path_head);
}


void get_tnode_block_and_output_net (int inode, int *iblk_ptr, int *inet_ptr) {

/* Returns the index of the block that this tnode is part of.  If the tnode *
 * is a CLB_OPIN or INPAD_OPIN (i.e. if it drives a net), the net index is  *
 * returned via inet_ptr.  Otherwise inet_ptr points at OPEN.               */

 int inet, ipin, iblk;
 t_tnode_type tnode_type;

 iblk = tnode_descript[inode].iblk;
 tnode_type = tnode_descript[inode].type;
 
 if (tnode_type == CLB_OPIN || tnode_type == INPAD_OPIN || tnode_type == EMB_OPIN || tnode_type == DSP_OPIN || tnode_type == MEMORY_OPIN) {
    ipin = tnode_descript[inode].ipin;
    inet = block[iblk].nets[ipin];
 }
 else {
    inet = OPEN;
 }

 *iblk_ptr = iblk;
 *inet_ptr = inet;
}


void do_constant_net_delay_timing_analysis (t_timing_inf timing_inf,
        t_subblock_data subblock_data, float constant_net_delay_value) {

/* Does a timing analysis (simple) where it assumes that each net has a      *
 * constant delay value.  Used only when operation == TIMING_ANALYSIS_ONLY.  */

 struct s_linked_vptr *net_delay_chunk_list_head;
 float **net_delay, **net_slack;

 float T_crit;
 
 net_slack = alloc_and_load_timing_graph (timing_inf, subblock_data);
 net_delay = alloc_net_delay (&net_delay_chunk_list_head);

 load_constant_net_delay (net_delay, constant_net_delay_value);
 load_timing_graph_net_delays (net_delay);
 T_crit = load_net_slack (net_slack, 0);
 
 printf ("\n");
  printf("do_constant_net_timing_analysis \n");
 print_critical_path ("critical_path.echo");
 printf ("\nCritical Path: %g (s)\n", T_crit);

/* print_timing_graph ("timing_graph.echo");
 print_net_slack ("net_slack.echo", net_slack);
 print_net_delay (net_delay, "net_delay.echo"); */

 free_timing_graph (net_slack);
 free_net_delay (net_delay, &net_delay_chunk_list_head);
}

int get_opin(int pin_num)
{
	switch (pin_num)
	{
		case B_SHIFTOUT:
			return O_SHIFTOUT;
			break;
		case B_COUT:
			return O_COUT;
			break;
		case B_BYOUT:
			return O_BYOUT;
			break;
		case B_YB:
			return O_YB;		
			break;
		case B_Y:
			return O_Y;		
			break;
		case B_YQ:
			return O_YQ;		
			break;
		case B_FX:
			return O_FX;
			break;
		default: 
			return O_Y;
			break;						
	}
	return 0;
	
}


