/*#include <stdlib.h> */
#include <stdio.h>
#include <math.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "place.h"
#include "read_place.h"
#include "draw.h"
#include "place_and_route.h"
#include "net_delay.h"
#include "path_delay.h"
#include "timing_place_lookup.h"
#include "timing_place.h"


/*#define DEBUG 1*/

/************** Types and defines local to place.c ***************************/

#define SMALL_NET 4    /* Cut off for incremental bounding box updates. */
/* 4 is fastest -- I checked.                    */

/*#define MAX_CARRY_CHAIN_LENGTH 129*/
#define MAX_CARRY_CHAIN_LENGTH 1000

/* For comp_cost.  NORMAL means use the method that generates updateable  *
* bounding boxes for speed.  CHECK means compute all bounding boxes from *
* scratch using a very simple routine to allow checks of the other       *
* costs.                                                                 */

enum cost_methods {NORMAL, CHECK, POWER};

#define FROM 0      /* What block connected to a net has moved? */
#define TO 1
#define FROM_AND_TO 2

//#define ERROR_TOL .001
#define ERROR_TOL .008
#define MAX_MOVES_BEFORE_RECOMPUTE 1000000

#define EMPTY -1      

/********************** Variables local to place.c ***************************/

/* [0..num_nets-1]  0 if net never connects to the same block more than  *
*  once, otherwise it gives the number of duplicate connections.        */

static int *duplicate_pins;    

/* [0..num_nets-1][0..num_unique_blocks-1]  Contains a list of blocks with *
* no duplicated blocks for ONLY those nets that had duplicates.           */

static int **unique_pin_list;

/* Cost of a net, and a temporary cost of a net used during move assessment. */

static float *net_cost = NULL, *temp_net_cost = NULL;     /* [0..num_nets-1] */

/* [0..num_nets-1][1..num_pins-1]. What is the value of the timing   */
/* driven portion of the cost function. These arrays will be set to  */
/* (criticality * delay) for each point to point connection. */
static float **point_to_point_timing_cost = NULL;
static float **temp_point_to_point_timing_cost = NULL;



/* [0..num_nets-1][1..num_pins-1]. What is the value of the delay */
/* for each connection in the circuit */
static float **point_to_point_delay_cost = NULL;
static float **temp_point_to_point_delay_cost = NULL;


/* [0..num_blocks-1][0..pins_per_clb-1]. Indicates which pin on the net */
/* this block corresponds to, this is only required during timing-driven */
/* placement. It is used to allow us to update individual connections on */
/* each net */
static int **net_pin_index = NULL;


/* [0..num_nets-1].  Store the bounding box coordinates and the number of    *
* blocks on each of a net's bounding box (to allow efficient updates),      *
* respectively.                                                             */

static struct s_bb *bb_coords = NULL, *bb_num_on_edges = NULL; 

/* Stores the maximum and expected occupancies, plus the cost, of each   *
* region in the placement.  Used only by the NONLINEAR_CONG cost        *
* function.  [0..num_region-1][0..num_region-1].  Place_region_x and    *
* y give the situation for the x and y directed channels, respectively. */

static struct s_place_region **place_region_x, **place_region_y;

/* Used only with nonlinear congestion.  [0..num_regions].            */

static float *place_region_bounds_x, *place_region_bounds_y;

/* The arrays below are used to precompute the inverse of the average   *
* number of tracks per channel between [subhigh] and [sublow].  Access *
* them as chan?_place_cost_fac[subhigh][sublow].  They are used to     *
* speed up the computation of the cost function that takes the length  *
* of the net bounding box in each dimension, divided by the average    *
* number of tracks in that direction; for other cost functions they    *
* will never be used.                                                  */

static float **chanx_place_cost_fac, **chany_place_cost_fac;


/* Expected crossing counts for nets with different #'s of pins.  From *
* ICCAD 94 pp. 690 - 695 (with linear interpolation applied by me).   */

static const float cross_count[50] = {   /* [0..49] */
1.0,    1.0,    1.0,    1.0828, 1.1536, 1.2206, 1.2823, 1.3385, 1.3991, 1.4493,
1.4974, 1.5455, 1.5937, 1.6418, 1.6899, 1.7304, 1.7709, 1.8114, 1.8519, 1.8924,
1.9288, 1.9652, 2.0015, 2.0379, 2.0743, 2.1061, 2.1379, 2.1698, 2.2016, 2.2334,
2.2646, 2.2958, 2.3271, 2.3583, 2.3895, 2.4187, 2.4479, 2.4772, 2.5064, 2.5356,
2.5610, 2.5864, 2.6117, 2.6371, 2.6625, 2.6887, 2.7148, 2.7410, 2.7671, 2.7933};


/********************* Static subroutines local to place.c *******************/

static void alloc_and_load_unique_pin_list (void);

static void free_unique_pin_list (void); 

static void alloc_place_regions (int num_regions);

static void load_place_regions (int num_regions);

static void free_place_regions (int num_regions); 

static void alloc_and_load_placement_structs (int place_cost_type, 
											  int num_regions, float place_cost_exp, float ***old_region_occ_x, 
											  float ***old_region_occ_y, struct s_placer_opts placer_opts ); 

static void free_placement_structs (int place_cost_type, int num_regions,
									float **old_region_occ_x, float **old_region_occ_y, 
									struct s_placer_opts placer_opts); 

static void alloc_and_load_for_fast_cost_update (float place_cost_exp);

static void initial_placement (enum e_pad_loc_type pad_loc_type,
							   char *pad_loc_file);

static float comp_bb_cost (int method, int place_cost_type, int num_regions);

/* Added: float *pwr_cost, float power_tradeoff, float inverse_prev_pwr_cost, */
/*        and power_flag.						      */	
/* JL, April 2002                                                             */
static int try_swap (float t, float *cost, float *bb_cost, float *timing_cost, 
					 float *pwr_cost, float rlim, int *pins_on_block,
					 int place_cost_type, float **old_region_occ_x, 
					 float **old_region_occ_y, int num_regions, boolean fixed_pins,
					 enum e_place_algorithm place_algorithm, int power_flag, 
					 float timing_tradeoff, float power_tradeoff,
					 float inverse_prev_bb_cost, float inverse_prev_timing_cost,
					 float inverse_prev_pwr_cost, float *delay_cost);

static void check_place (float bb_cost, float timing_cost, int place_cost_type, 
						 int num_regions, enum e_place_algorithm place_algorithm,
						 float delay_cost);

static float starting_t (float *cost_ptr, float *bb_cost_ptr, float *timing_cost_ptr,
						 float *pwr_cost_ptr, int *pins_on_block, int place_cost_type, float **old_region_occ_x,
						 float **old_region_occ_y, int num_regions, boolean fixed_pins,
						 struct s_annealing_sched annealing_sched, int max_moves, float rlim,
						 enum e_place_algorithm place_algorithm, int power_flag, float timing_tradeoff,
						 float power_tradeoff, float inverse_prev_bb_cost, float inverse_prev_timing_cost,
						 float inverse_prev_pwr_cost, float *delay_cost_ptr);

static void update_t (float *t, float std_dev, float rlim, float success_rat,
					  struct s_annealing_sched annealing_sched); 

static void update_rlim (float *rlim, float success_rat);

static int exit_crit (float t, float cost, struct s_annealing_sched
					  annealing_sched);

static int count_connections(void);

static void compute_net_pin_index_values(void);

static double get_std_dev (int n, double sum_x_squared, double av_x);

static void free_fast_cost_update_structs (void); 

static float recompute_bb_cost (int method, int place_cost_type, int 
								num_regions);

static float comp_td_point_to_point_delay (int inet, int ipin);

static void update_td_cost(int *from_list, int *to_list, int num_of_pins, int move_height);


static void comp_delta_td_cost(int *from_list, int *to_list, 
							   int num_of_pins, float *delta_timing, 
							   float *delta_delay, int move_height);

static void comp_td_costs (float *timing_cost, float *connection_delay_sum);

static int assess_swap (float delta_c, float t);

static void find_to (int x_from, int y_from, int type, float rlim, 
					 int *x_to, int *y_to, int move_height);

static void get_non_updateable_bb (int inet, struct s_bb *bb_coord_new);

static void update_bb (int inet, struct s_bb *bb_coord_new, struct s_bb 
					   *bb_edge_new, int xold, int yold, int xnew, int ynew);

static int find_affected_nets (
							   int *nets_to_update, 
							   int *net_block_moved,
							   int *offset_y,
							   int *from_list, 
							   int *to_list, 
							   int move_height,
							   int num_of_pins);

static float get_net_cost (int inet, struct s_bb *bb_ptr);

static float nonlinear_cong_cost (int num_regions);

static void update_region_occ (int inet, struct s_bb*coords, 
							   int add_or_sub, int num_regions);

static void save_region_occ (float **old_region_occ_x, 
							 float **old_region_occ_y, int num_regions);

static void restore_region_occ (float **old_region_occ_x,
								float **old_region_occ_y, int num_regions);

static void get_bb_from_scratch (int inet, struct s_bb *coords, 
								 struct s_bb *num_on_edges); 

static int carry_length(int iblk);
static int max_carry_length();
static void get_list_of_carry_heads(int *num_ptr,
                                    int **heads,
                                    int **lengths);


static void find_to_veb (int x_from, int y_from, int type, float rlim, int *x_to, int *y_to, int veb_index, int move_height);


/*****************************************************************************/



void try_place (struct s_placer_opts placer_opts,struct s_annealing_sched 
                annealing_sched, t_chan_width_dist chan_width_dist, 
				struct s_router_opts router_opts, 
				struct s_det_routing_arch det_routing_arch,
				t_segment_inf *segment_inf,
				t_timing_inf timing_inf,
				t_subblock_data *subblock_data_ptr) {
	
	/* Does almost all the work of placing a circuit.  Width_fac gives the   *
	* width of the widest channel.  Place_cost_exp says what exponent the   *
	* width should be taken to when calculating costs.  This allows a       *
	* greater bias for anisotropic architectures.  Place_cost_type          *
	* determines which cost function is used.  num_regions is used only     *
	* the place_cost_type is NONLINEAR_CONG.                                */
	
	
	int tot_iter, inner_iter, success_sum, pins_on_block[7];
	int move_lim, moves_since_cost_recompute, width_fac;
	float t,  success_rat, rlim, d_max, est_crit;
	float cost, timing_cost, bb_cost, new_bb_cost, new_timing_cost;
	float delay_cost, new_delay_cost,  place_delay_value;
	float inverse_prev_bb_cost, inverse_prev_timing_cost;
	float pwr_cost, new_pwr_cost;  /* Added by JL, October 2002 */
	float inverse_prev_pwr_cost;   /* Added by JL, October 2002 */
	float oldt;
	double av_cost, av_bb_cost, av_timing_cost, av_delay_cost, av_power_cost, sum_of_squares, std_dev;
	float **old_region_occ_x, **old_region_occ_y;
	char msg[BUFSIZE];
	boolean fixed_pins;  /* Can pads move or not? */
	int num_connections;
	int inet, ipin, outer_crit_iter_count, inner_crit_iter_count, inner_recompute_limit;
	float **net_slack, **net_delay;
	float crit_exponent;
	float first_rlim, final_rlim, inverse_delta_rlim;
	float **remember_net_delay_original_ptr; /*used to free net_delay if it is re-assigned*/

	float timing_cost_check, delay_cost_check;
	int i, j;
	
	remember_net_delay_original_ptr = NULL; /*prevents compiler warning*/

	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) 
	{
		/*do this before the initial placement to avoid messing up the initial placement */
		printf("Try place, num net %d\n", num_nets);		
		printf("NET_TIMING_DRIVEN_PLACE or PATH_TIMING_DRIVEN_PLACE or enable_timing_computations \n");
		alloc_lookups_and_criticalities(chan_width_dist,
										router_opts, 
										det_routing_arch, 
										segment_inf,
										timing_inf, 
										*subblock_data_ptr,
										&net_delay, &net_slack);
//		printf("Finish alloc_lookups_and_criticalities \n");	
		remember_net_delay_original_ptr = net_delay;
		
		/*#define PRINT_LOWER_BOUND*/
#ifdef PRINT_LOWER_BOUND
		/*print the crit_path, assuming delay between blocks that are*
			*block_dist apart*/
		
		if (placer_opts.block_dist <= nx)
			place_delay_value = delta_clb_to_clb[placer_opts.block_dist][0];
		else if (placer_opts.block_dist <= ny)
			place_delay_value = delta_clb_to_clb[0][placer_opts.block_dist];
		else
			place_delay_value = delta_clb_to_clb[nx][ny];
		
		printf("\nLower bound assuming delay of %g\n", place_delay_value);
		
		load_constant_net_delay (net_delay, place_delay_value);
		load_timing_graph_net_delays(net_delay);
		d_max = load_net_slack(net_slack, 0);
		
		print_critical_path("Placement_Lower_Bound.echo");
		print_sink_delays("Placement_Lower_Bound_Sink_Delays.echo");
		
		/*also print sink delays assuming 0 delay between blocks, 
			this tells us how much logic delay is on each path*/
		
		load_constant_net_delay (net_delay, 0);
		load_timing_graph_net_delays(net_delay);
		d_max = load_net_slack(net_slack, 0);
		
		print_sink_delays("Placement_Logic_Sink_Delays.echo");
#endif
		
	}
	
	width_fac = placer_opts.place_chan_width;
	if (placer_opts.pad_loc_type == FREE)
		fixed_pins = FALSE;
	else
		fixed_pins = TRUE;
	
//	printf("init_chan \n");


	init_chan (width_fac, chan_width_dist); 
	
//	printf("alloc_and_load_placement_structs \n");
	alloc_and_load_placement_structs (placer_opts.place_cost_type, 
									  placer_opts.num_regions, placer_opts.place_cost_exp,
									  &old_region_occ_x, &old_region_occ_y, placer_opts); 
	

	/* Initial Placement*/
	printf("initial Placement \n");
	initial_placement (placer_opts.pad_loc_type, placer_opts.pad_loc_file);
	
//	printf("init_draw_coords \n");
	init_draw_coords ((float) width_fac);
	
	
	/* Storing the number of pins on each type of block makes the swap routine *
		* slightly more efficient.                                                */
	
	pins_on_block[CLB] = pins_per_clb;
	pins_on_block[MEMORY] = pins_per_memory;
	pins_on_block[DSP] = pins_per_dsp;
	pins_on_block[EMB] = pins_per_dsp;	//shouldn't use now
	pins_on_block[OUTPAD] = 1;
	pins_on_block[INPAD] = 1;
	
	/* Gets initial cost and loads bounding boxes. */	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
		
		
//		printf("comp_bb_cost\n");
		bb_cost = comp_bb_cost (NORMAL, placer_opts.place_cost_type, 
								placer_opts.num_regions);
		//printf("bb_cost = %f \n", bb_cost);
		
		crit_exponent = placer_opts.td_place_exp_first; /*this will be modified when rlim starts to change*/
		
//		printf("compute_net_pin_index_values\n");
		compute_net_pin_index_values();		

		num_connections = count_connections();
//		printf("\nThere are %d point to point connections in this circuit\n\n", num_connections);
		
		if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE) {
			for (inet = 0; inet<num_nets; inet++)  
				for (ipin=1; ipin<net[inet].num_pins; ipin++)
					timing_place_crit[inet][ipin] = 0;  /*dummy crit values*/
			
//			printf("comp_td_costs\n");
			comp_td_costs(&timing_cost, &delay_cost); //first pass gets delay_cost, which is used 			
																						//in criticality computations in the next call
																						//to comp_td_costs.
			
			place_delay_value = delay_cost / num_connections;  /*used for computing criticalities */
//			printf("load_constant_net_delay\n");
			load_constant_net_delay (net_delay, place_delay_value);
			
		}
		else 
			place_delay_value = 0;
		
		
		if (placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
			net_delay = point_to_point_delay_cost;/*this keeps net_delay up to date with      *
			*the same values that the placer is using  *
			*point_to_point_delay_cost is computed each*
			*time that comp_td_costs is called, and is *
			*also updated after any swap is accepted   */
		}
		
		
		load_timing_graph_net_delays(net_delay);
//		printf("load_net_slack\n");
		d_max = load_net_slack(net_slack, 0);
//		printf("load_criticalities \n");
		load_criticalities( placer_opts,  net_slack, d_max, crit_exponent);
		outer_crit_iter_count = 1;
		
		/*now we can properly compute costs  */
		//printf("comp_td_costs \n");
		comp_td_costs(&timing_cost, &delay_cost); /*also puts proper values into point_to_point_delay_cost*/
		//printf("timing_cost %f, delay_cost %f \n", timing_cost, delay_cost);
		
		inverse_prev_timing_cost = 1/timing_cost;
		inverse_prev_bb_cost = 1/bb_cost;
		cost = 1; /*our new cost function uses normalized values of           */
		/*bb_cost and timing_cost, the value of cost will be reset  */
		/*to 1 at each temperature when *_TIMING_DRIVEN_PLACE is true*/
	}
	else { /*BOUNDING_BOX_PLACE*/
		cost = bb_cost = comp_bb_cost (NORMAL, placer_opts.place_cost_type, 
									   placer_opts.num_regions);
		timing_cost = 0;
		delay_cost = 0;
		place_delay_value = 0;
		outer_crit_iter_count =0;
		num_connections = 0;
		d_max = 0;
		crit_exponent = 0;
		
		inverse_prev_timing_cost = 0; /*inverses not used */
		inverse_prev_bb_cost = 0;
	}	
/* Added by JL, October 2002 */
if (placer_opts.power_flag == 1)
{
	pwr_cost = comp_bb_cost(POWER, placer_opts.place_cost_type,
							placer_opts.num_regions);
	inverse_prev_bb_cost = 1.0/bb_cost;
	inverse_prev_pwr_cost = 1.0/pwr_cost;
}

move_lim = (int) (annealing_sched.inner_num * pow(num_blocks,1.3333));
//printf ("move_lim %d\n", move_lim);
if (placer_opts.inner_loop_recompute_divider != 0)
inner_recompute_limit = (int) (0.5 + (float) move_lim / 
							   (float) placer_opts.inner_loop_recompute_divider);
else  /*don't do an inner recompute */
inner_recompute_limit = move_lim + 1;


/* Sometimes I want to run the router with a random placement.  Avoid *
* using 0 moves to stop division by 0 and 0 length vector problems,  *
* by setting move_lim to 1 (which is still too small to do any       *
							* significant optimization).                                         */

if (move_lim <= 0) 
move_lim = 1;  

rlim = (float) max (nx, ny);

first_rlim = rlim; /*used in timing-driven placement for exponent computation*/
final_rlim = 1;
inverse_delta_rlim = 1 / (first_rlim - final_rlim);

printf("starting_t \n");
t = starting_t (&cost, &bb_cost, &timing_cost, &pwr_cost,
				pins_on_block, placer_opts.place_cost_type,
				old_region_occ_x, old_region_occ_y, placer_opts.num_regions,
				fixed_pins, annealing_sched, move_lim, rlim,
				placer_opts.place_algorithm, placer_opts.power_flag,
				placer_opts.timing_tradeoff,
				placer_opts.power_tradeoff,
				inverse_prev_bb_cost, inverse_prev_timing_cost,
				inverse_prev_pwr_cost, &delay_cost);

tot_iter = 0;
moves_since_cost_recompute = 0;
printf("Initial Placement Cost: %g bb_cost: %g td_cost: %g delay_cost: %g pwr_cost %g\n\n",
	   cost, bb_cost, timing_cost, delay_cost, pwr_cost); 

#ifndef SPEC
printf("%8s %9s %10s %10s %10s %10s %10s %10s %9s %8s %7s %7s %10s %7s\n", "T", "Cost", "Av BB Cost", 
	   "Av TD Cost", "Av Tot Del", "Av Pr Cost", "P to P Del", "d_max", "Ac Rate", "Std Dev", "R limit", "Exp", 
	   "Tot. Moves", "Alpha");
printf("%8s %9s %10s %10s %10s %10s %10s %10s %9s %8s %7s %7s %10s %7s\n", "--------", "---------", 
	   "----------", "----------", "----------", "----------", "----------", "----------", "---------", "--------",  
	   "-------", "-------","----------", "-------");
#endif

sprintf(msg,"Initial Placement.  Cost: %g  BB Cost: %g  TD Cost: %g  Delay Cost: %g"
		" Power Cost: %g\t d_max %g Channel Factor: %d",
		cost, bb_cost, timing_cost, delay_cost, pwr_cost, d_max, width_fac);


update_screen(MAJOR, msg, PLACEMENT, FALSE);


////////////////////////////////////////////////
while (exit_crit(t, cost, annealing_sched) == 0) 
{
	////////////////////////////////////////////////
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
		cost = 1;
	}
	
    av_cost = 0.;
    av_bb_cost = 0.;
    av_power_cost = 0.;
    av_delay_cost = 0.;
    av_timing_cost = 0.;
    sum_of_squares = 0.;
    success_sum = 0;
//	printf ("While exit_crit , outer_crit_iter_count %d \n", outer_crit_iter_count);
	
    if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) 
	{
		
		if (outer_crit_iter_count >= placer_opts.recompute_crit_iter ||
			placer_opts.inner_loop_recompute_divider != 0) {
#ifdef VERBOSE
			printf("Outer Loop Recompute Criticalities\n");
#endif
			place_delay_value = delay_cost / num_connections;
			
			if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE)
				load_constant_net_delay (net_delay, place_delay_value);
			/*note, for path_based, the net delay is not updated since it is current,
				*because it accesses point_to_point_delay array */
		
			//printf("exit_crit: load_timing_graph_net_delays\n");
			load_timing_graph_net_delays(net_delay);
			d_max = load_net_slack(net_slack, 0);			
			//printf("exit_crit: load_criticalities\n");
			load_criticalities( placer_opts, net_slack, d_max, crit_exponent);
			/*recompute costs from scratch, based on new criticalities*/
			//printf("exit_crit: comp_td_costs\n");
			comp_td_costs(&timing_cost, &delay_cost); 
			outer_crit_iter_count = 0;
		} 
		outer_crit_iter_count ++;
		
		/*at each temperature change we update these values to be used     */
		/*for normalizing the tradeoff between timing and wirelength (bb)  */
		inverse_prev_bb_cost = 1/bb_cost;
		inverse_prev_timing_cost = 1/timing_cost;
    }
	
    /* Changed by JL, October 2002 */
    if (placer_opts.power_flag == 1)
    {
		/*at each temperature change we update these values to be used     */
		/*for normalizing the tradeoff between timing and wirelength (bb)  */
		inverse_prev_pwr_cost = 1/pwr_cost;
    }
	
    inner_crit_iter_count = 1;
	
	//printf ("move_lim 2 %d\n", move_lim);
	//move_lim = 10000;

    for (inner_iter=0; inner_iter < move_lim; inner_iter++) 
	{
		if (try_swap(t, &cost, &bb_cost, &timing_cost, &pwr_cost,
					 rlim, pins_on_block, placer_opts.place_cost_type,
					 old_region_occ_x, old_region_occ_y, placer_opts.num_regions,
					 fixed_pins, placer_opts.place_algorithm, placer_opts.power_flag,
					 placer_opts.timing_tradeoff, placer_opts.power_tradeoff,
					 inverse_prev_bb_cost, inverse_prev_timing_cost, 
					 inverse_prev_pwr_cost, &delay_cost) == 1) {
			success_sum++;
			av_cost += cost;
			av_bb_cost += bb_cost;
			av_power_cost += pwr_cost;
			av_timing_cost += timing_cost;
			av_delay_cost += delay_cost;
			sum_of_squares += cost * cost;
		}
		//printf("inner_iter %d, outer_crit_iter_count %d \n", inner_iter, outer_crit_iter_count);
		
		if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
			placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE)
		{
			
			if (inner_crit_iter_count >= inner_recompute_limit 
				&& inner_iter != move_lim-1) { /*on last iteration don't recompute*/
				
				inner_crit_iter_count = 0;

				#ifdef VERBOSE
								printf("Inner Loop Recompute Criticalities\n");
				#endif

				if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE) {
					place_delay_value = delay_cost / num_connections;
					load_constant_net_delay (net_delay, place_delay_value);
				}
				
				//printf("exit_crit, try_swap: load_timing_graph_net_delays\n");
				load_timing_graph_net_delays(net_delay);
				//printf("exit_crit, try_swap: load_net_slack\n");
				d_max = load_net_slack(net_slack, 0);				
				load_criticalities( placer_opts, net_slack, d_max, crit_exponent);
				comp_td_costs(&timing_cost, &delay_cost); 
			}
			inner_crit_iter_count ++;
		}
		


		#ifdef VERBOSE
		printf("/nt = %g  cost = %g   bb_cost = %g timing_cost = %g move = %d dmax = %g\n", t, cost, bb_cost, timing_cost, inner_iter, d_max);
		if (fabs(bb_cost - comp_bb_cost(CHECK, placer_opts.place_cost_type, 
										placer_opts.num_regions)) > bb_cost * ERROR_TOL) 
		exit(1);
		#endif 
    }
	//printf("inner_crit_iter_count %d \n", inner_crit_iter_count);

	/* Lines below prevent too much round-off error from accumulating *
	* in the cost over many iterations.  This round-off can lead to  *
	* error checks failing because the cost is different from what   *
	* you get when you recompute from scratch.                       */



	moves_since_cost_recompute += move_lim;
	if (moves_since_cost_recompute > MAX_MOVES_BEFORE_RECOMPUTE) 
	{
		new_bb_cost = recompute_bb_cost (NORMAL, placer_opts.place_cost_type, 
										placer_opts.num_regions);       
		if (fabs(new_bb_cost - bb_cost) > bb_cost * ERROR_TOL) {
			printf("Error in try_place:  new_bb_cost = %g, old bb_cost = %g.\n",
				new_bb_cost, bb_cost);
			exit (1);
		}
		bb_cost = new_bb_cost;
		
		/* Added by JL, October 2002 */
		if (placer_opts.power_flag == 1)
		{
			new_pwr_cost = recompute_bb_cost (POWER, placer_opts.place_cost_type,
											placer_opts.num_regions);
			if (fabs(new_pwr_cost - pwr_cost) > pwr_cost * ERROR_TOL) 
			{
				printf("Error in try_place:  new_pwr_cost = %g, old pwr_cost = %g.\n", new_pwr_cost, pwr_cost);
				/*           exit (1); */
			}
			pwr_cost = new_pwr_cost;
		}
		
		
		if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
			placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
			comp_td_costs(&new_timing_cost, &new_delay_cost);
			if (fabs(new_timing_cost - timing_cost) > timing_cost * ERROR_TOL) {
				printf("Error in try_place:  new_timing_cost = %g, old timing_cost = %g.\n",
					new_timing_cost, timing_cost);
				exit (1);
			}
			if (fabs(new_delay_cost - delay_cost) > delay_cost * ERROR_TOL) {
				printf("Error in try_place:  new_delay_cost = %g, old delay_cost = %g.\n",
					new_delay_cost, delay_cost);
				exit (1);
			}
			timing_cost = new_timing_cost;
		}
		
		if (placer_opts.place_algorithm ==BOUNDING_BOX_PLACE) {
			cost = new_bb_cost;
		}
		moves_since_cost_recompute = 0;
	}

	tot_iter += move_lim;
	success_rat = ((float) success_sum)/ move_lim;
	if (success_sum == 0) {
		av_cost = cost;
		av_bb_cost = bb_cost;
		av_power_cost = pwr_cost;
		av_timing_cost = timing_cost;
		av_delay_cost = delay_cost;
	}
	else {
		av_cost /= success_sum;
		av_bb_cost /= success_sum;
		av_power_cost /= success_sum;
		av_timing_cost /= success_sum;
		av_delay_cost /= success_sum;
	}
	std_dev = get_std_dev (success_sum, sum_of_squares, av_cost);

	/*
	#ifndef SPEC
	printf("%8.4g %9.5g %10.5g %10.5g %10.5g %10.5g %10.5g %10.4g %9.4g %8.3g %7.4g %7.4g %10d  ",t, av_cost, 
		av_bb_cost, av_timing_cost, av_delay_cost, av_power_cost, place_delay_value, d_max, success_rat, std_dev, 
		rlim, crit_exponent,tot_iter);
	#endif
	*/
	oldt = t;  /* for finding and printing alpha. */
	update_t (&t, std_dev, rlim, success_rat, annealing_sched);
	/*
	#ifndef SPEC
	printf("%7.4g\n",t/oldt);
	#endif

	sprintf(msg,"Cost: %g  BB Cost %g  TD Cost %g  PWR Cost %g Temperature: %g  d_max: %g",
			cost, bb_cost, timing_cost, pwr_cost, t, d_max);
	update_screen(MINOR, msg, PLACEMENT, FALSE);
	*/
	update_rlim (&rlim, success_rat);

	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
		crit_exponent = (1 - (rlim - final_rlim) * inverse_delta_rlim)*
		(placer_opts.td_place_exp_last - placer_opts.td_place_exp_first) + 
		placer_opts.td_place_exp_first;
	}
	#ifdef VERBOSE 
	dump_clbs();
	#endif
	} // While


	//printf("finished while exit_crit, bb_cost%8.4g, timing_cost %8.4g, delay_cost %8.4g\n", bb_cost, timing_cost, delay_cost);
	t = 0;   /* freeze out */
	av_cost = 0.;
	av_bb_cost = 0.;
	av_power_cost = 0.;
	av_timing_cost = 0.;
	sum_of_squares = 0.;
	av_delay_cost = 0.;
	success_sum = 0;

	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) 
	{
		/*at each temperature change we update these values to be used     */
		/*for normalizing the tradeoff between timing and wirelength (bb)  */

		//printf("update the value to be used \n");
		if (outer_crit_iter_count >= placer_opts.recompute_crit_iter ||
			placer_opts.inner_loop_recompute_divider != 0) {
			
	#ifdef VERBOSE
			printf("Outer Loop Recompute Criticalities\n");
	#endif
			place_delay_value = delay_cost / num_connections;
			
			if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE)
				load_constant_net_delay (net_delay, place_delay_value);
			//printf("update the value to be used : load_timing_graph_net_delays\n");
			load_timing_graph_net_delays(net_delay);
			//printf("update the value to be used : load_net_slack\n");
			d_max = load_net_slack(net_slack, 0);
			//printf("update the value to be used : load_criticalities\n");
			load_criticalities( placer_opts, net_slack, d_max, crit_exponent); 
			/*recompute criticaliies */
			//printf("update the value to be used : comp_td_costs\n");
			comp_td_costs(&timing_cost, &delay_cost);  
			//printf("update the value to be used : comp_td_costs, timing_cost %8.4g, delay_cost %8.4g\n", timing_cost, delay_cost);

			outer_crit_iter_count = 0;
		}
		outer_crit_iter_count ++;
		
		inverse_prev_bb_cost = 1/(bb_cost);
		inverse_prev_timing_cost = 1/(timing_cost);
	}

	/* Added by JL, October 2002 */
	if (placer_opts.power_flag == 1)
	{
		inverse_prev_pwr_cost = 1/pwr_cost;
	}

	inner_crit_iter_count = 1;
//	printf ("move_lim 2 %d\n", move_lim);
	
	for (inner_iter=0; inner_iter < move_lim; inner_iter++) 
	{	
		if (try_swap(t, &cost, &bb_cost, &timing_cost, &pwr_cost, 
					rlim, pins_on_block, placer_opts.place_cost_type, 
					old_region_occ_x, old_region_occ_y, placer_opts.num_regions,
					fixed_pins, placer_opts.place_algorithm, placer_opts.power_flag,
					placer_opts.timing_tradeoff, placer_opts.power_tradeoff,
					inverse_prev_bb_cost, inverse_prev_timing_cost, 
					inverse_prev_pwr_cost, &delay_cost) == 1) {
			success_sum++;
			av_cost += cost;
			av_bb_cost += bb_cost;
			av_power_cost += pwr_cost;
			av_delay_cost += delay_cost;
			av_timing_cost += timing_cost;
			sum_of_squares += cost * cost;
			
			if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
				placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
				
				if (inner_crit_iter_count >= inner_recompute_limit
					&& inner_iter != move_lim-1) {
					
					inner_crit_iter_count = 0;
	#ifdef VERBOSE
					printf("Inner Loop Recompute Criticalities\n");
	#endif
					if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE) {
						place_delay_value = delay_cost / num_connections;
						load_constant_net_delay (net_delay, place_delay_value);
					}
//					printf("exit_crit, try_swap : load_timing_graph_net_delays 2\n");
					load_timing_graph_net_delays(net_delay);
//					printf("exit_crit, try_swap : load_net_slack 2\n");
					d_max = load_net_slack(net_slack, 0);
//					printf("exit_crit, try_swap : load_criticalities 2\n");
					load_criticalities( placer_opts, net_slack, d_max, crit_exponent);
//					printf("exit_crit, try_swap : comp_td_costs 2\n");
					comp_td_costs(&timing_cost, &delay_cost); 
				}
				inner_crit_iter_count ++;
			}
		}
	#ifdef VERBOSE 
		printf("t = %g  cost = %g   move = %d\n",t, cost, tot_iter);
	#endif
	}
	tot_iter += move_lim;
	success_rat = ((float) success_sum) / move_lim;
	if (success_sum == 0) {
		av_cost = cost;
		av_bb_cost = bb_cost;
		av_power_cost = pwr_cost;
		av_delay_cost = delay_cost;
		av_timing_cost = timing_cost;
	}
	else {
		av_cost /= success_sum;
		av_bb_cost /= success_sum;
		av_power_cost /= success_sum;
		av_delay_cost /= success_sum;
		av_timing_cost /= success_sum;
	}

	std_dev = get_std_dev (success_sum, sum_of_squares, av_cost);


	#ifndef SPEC
	printf("%8.4g %9.5g %10.5g %10.5g %10.5g %10.5g %10.5g %10.4g %9.4g %8.3g %7.4g %7.4g %10d  ",t, av_cost, 
		av_bb_cost, av_timing_cost, av_delay_cost, av_power_cost, place_delay_value, d_max, success_rat, std_dev, 
		rlim, crit_exponent,tot_iter);
	#endif

	#ifdef VERBOSE 
	dump_clbs();
	#endif
	printf("------------------------------------------------------------------------------------------- \n");
	printf("check_place, timing_cost %8.4g , delay_cost %8.4g\n", timing_cost, delay_cost);
	check_place(bb_cost, timing_cost, placer_opts.place_cost_type, placer_opts.num_regions,
				placer_opts.place_algorithm, delay_cost);
	//printf("finish check_place \n");

	if (placer_opts.enable_timing_computations &&
		placer_opts.place_algorithm == BOUNDING_BOX_PLACE) {
		/*need this done since the timing data has not been kept up to date*
		*in bounding_box mode */
		for (inet = 0; inet<num_nets; inet++)  
			for (ipin=1; ipin<net[inet].num_pins; ipin++)
				timing_place_crit[inet][ipin] = 0;  /*dummy crit values*/
		comp_td_costs(&timing_cost, &delay_cost); /*computes point_to_point_delay_cost*/
	}

	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {
		net_delay = point_to_point_delay_cost;/*this makes net_delay up to date with    *
		*the same values that the placer is using*/	

		
		load_timing_graph_net_delays(net_delay);
		
		est_crit = load_net_slack(net_slack, 0);
		
	#ifdef PRINT_SINK_DELAYS
		printf("print_sink_delay\n");
		//print_sink_delays("Placement_Sink_Delays.echo");
	#endif
	#ifdef PRINT_NET_SLACKS
		printf("print_net_slack\n");
		//print_net_slack("Placement_Net_Slacks.echo", net_slack);
	#endif
	#ifdef PRINT_PLACE_CRIT_PATH
		printf("print_critical_path\n");
		//print_critical_path("Placement_Crit_Path.echo");
	#endif
		printf("Placement Estimated Crit Path Delay: %g\n\n", est_crit);
	}

//sprintf(msg,"Placement. Cost: %g  bb_cost: %g td_cost: %g Channel Factor: %d d_max: %g", cost, bb_cost, timing_cost, width_fac, d_max);
sprintf(msg,"Placement. Cost: %g  ", cost );
printf("Placement. Cost: %g  bb_cost: %g  td_cost: %g  delay_cost: %g.\n",
	   cost, bb_cost, timing_cost,delay_cost);
update_screen(MAJOR, msg, PLACEMENT, FALSE);
#ifdef SPEC
printf ("Total moves attempted: %d.0\n", tot_iter);
#endif

if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
	placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
	placer_opts.enable_timing_computations) {
	
	net_delay = remember_net_delay_original_ptr;
	
	free_placement_structs (placer_opts.place_cost_type, placer_opts.num_regions,
							old_region_occ_x, old_region_occ_y, placer_opts); 
	free_lookups_and_criticalities( &net_delay, &net_slack);
	
}

	printf("finished try_place \n");
}

static int count_connections() {
	/*only count non-global connections*/
	
	int count, inet;
	
	count = 0;
	
	for (inet = 0; inet < num_nets; inet ++) {
		
		if (is_global[inet]) 
			continue;
		
		count += (net[inet].num_pins - 1);
	}
	return(count);
}

static void compute_net_pin_index_values() {
	/*computes net_pin_index array, this array allows us to quickley*/
	/*find what pin on the net a block pin corresponds to */
	
	int inet, netpin, blk, iblk, ipin, ppp;
	int i;
	
	/*initialize values to OPEN */
	for (iblk=0; iblk<num_blocks; iblk++) 
	{
		ppp = MAX(pins_per_clb,pins_per_memory);
		ppp = MAX(ppp,pins_per_dsp);

		for (i = 1 ; i<number_module; i++)
		{
			ppp = MAX(ppp,constraint[i].num_pins);			
		}
		
		for (ipin=0; ipin < ppp; ipin++)
			net_pin_index[iblk][ipin] = OPEN;
	}
	for (inet=0; inet<num_nets; inet++) {
		
		if (is_global[inet]) 
			continue;
		
		for(netpin=0;netpin<net[inet].num_pins;netpin++) 
		{
			blk=net[inet].blocks[netpin];
			if (block[blk].type == INPAD)
				net_pin_index[blk][0] = 0; /*there is only one block pin,so it is 0, and it */
			/*is driving the net since this is an INPAD*/
			else if (block[blk].type == OUTPAD)
			{ 
				net_pin_index[blk][0] = netpin;  /*there is only one block pin,it is 0*/
				/*
				if (blk == 0)
				{
					printf("blk : %s, inet %d (%s), netpin %d \n", block[blk].name, inet, net[inet].name, netpin);
					//exit(0);
				}
				*/
			}
			else {
				net_pin_index[blk][net[inet].blk_pin[netpin]] = netpin;
				/*
				if (block[blk].type == EMB)
				{
					printf("net_pin_index[%d][%d] = %d,  inet %d \n", blk, net[inet].blk_pin[netpin], netpin, inet);
				}
				*/


			}  
		}  
	}
}


static double get_std_dev (int n, double sum_x_squared, double av_x) {
	
	/* Returns the standard deviation of data set x.  There are n sample points, *
	* sum_x_squared is the summation over n of x^2 and av_x is the average x.   *
	* All operations are done in double precision, since round off error can be *
	* a problem in the initial temp. std_dev calculation for big circuits.      */
	
	double std_dev;
	
	if (n <= 1) 
		std_dev = 0.;
	else 
		std_dev = (sum_x_squared - n * av_x * av_x) / (double) (n - 1);
	
	if (std_dev > 0.)        /* Very small variances sometimes round negative */
		std_dev = sqrt (std_dev);
	else
		std_dev = 0.;
	
	return (std_dev);
}


static void update_rlim (float *rlim, float success_rat) {
	
	/* Update the range limited to keep acceptance prob. near 0.44.  Use *
	* a floating point rlim to allow gradual transitions at low temps.  */
	
	float upper_lim;
	
	*rlim = (*rlim) * (1. - 0.44 + success_rat);
	upper_lim = max(nx,ny);
	*rlim = min(*rlim,upper_lim);
	*rlim = max(*rlim,1.);  
	/* *rlim = (float) nx; */
}


static void update_t (float *t, float std_dev, float rlim, 
					  float success_rat, struct s_annealing_sched annealing_sched) {
	
	/* Update the temperature according to the annealing schedule selected. */
	
	/*  float fac; */
	
	if (annealing_sched.type == USER_SCHED) {
		*t = annealing_sched.alpha_t * (*t);
	}
	
	/* Old standard deviation based stuff is below.  This bogs down hopelessly *
	* for big circuits (alu4 and especially bigkey_mod).                      */
	/* #define LAMBDA .7  */
	/* ------------------------------------ */
	/* else if (std_dev == 0.) {  
    *t = 0.;
	}
else {
		fac = exp (-LAMBDA*(*t)/std_dev);
		fac = max(0.5,fac);   
		*t = (*t) * fac;
}   */
	/* ------------------------------------- */
	
	else {                             /* AUTO_SCHED */
		if (success_rat > 0.96) {
			*t = (*t) * 0.5; 
		}
		else if (success_rat > 0.8) {
			*t = (*t) * 0.9;
		}
		else if (success_rat > 0.15 || rlim > 1.) {
			*t = (*t) * 0.95;
		}
		else {
			*t = (*t) * 0.8; 
		}
	}
}


static int exit_crit (float t, float cost, struct s_annealing_sched 
					  annealing_sched) {
	
	/* Return 1 when the exit criterion is met.                        */
	
	if (annealing_sched.type == USER_SCHED) {
		if (t < annealing_sched.exit_t) {
			return(1);
		}
		else {
			return(0);
		}
	} 
	
	/* Automatic annealing schedule */
	
	if (t < 0.005 * cost / num_nets) {
		return(1);
	}
	else {
		return(0);
	}
}


static float starting_t (float *cost_ptr, float *bb_cost_ptr, float *timing_cost_ptr,
						 float *pwr_cost_ptr, int *pins_on_block, int place_cost_type, float **old_region_occ_x,
						 float **old_region_occ_y, int num_regions, boolean fixed_pins,
						 struct s_annealing_sched annealing_sched, int max_moves, float rlim,
						 enum e_place_algorithm place_algorithm, int power_flag,
						 float timing_tradeoff, float power_tradeoff,
						 float inverse_prev_bb_cost, float inverse_prev_timing_cost, 
						 float inverse_prev_pwr_cost, float *delay_cost_ptr) {
	
	/* Finds the starting temperature (hot condition).              */
	
	int i, num_accepted, move_lim, tmp;
	double std_dev, av, sum_of_squares;  /* Double important to avoid round off */
	
	if (annealing_sched.type == USER_SCHED) 
		return (annealing_sched.init_t);  
	
	move_lim = min (max_moves, num_blocks);
	
	num_accepted = 0;
	av = 0.;
	sum_of_squares = 0.;
	
	/* Try one move per block.  Set t high so essentially all accepted. */
	//printf("net_pin_index 2 %d \n", net_pin_index[0][0]);	
	for (i=0;i<move_lim;i++) {
		if (try_swap (1.e30, cost_ptr, bb_cost_ptr, timing_cost_ptr, 
					  pwr_cost_ptr, rlim, pins_on_block, place_cost_type,
					  old_region_occ_x, old_region_occ_y, num_regions, 
					  fixed_pins, place_algorithm, power_flag,
					  timing_tradeoff, power_tradeoff,
					  inverse_prev_bb_cost, inverse_prev_timing_cost,
					  inverse_prev_pwr_cost, delay_cost_ptr) == 1) {
			num_accepted++; 
			av += *cost_ptr;
			sum_of_squares += *cost_ptr * (*cost_ptr);
		}   
	}   
	
	/* Initial Temp = 20*std_dev. */
	
	if (num_accepted != 0)
		av /= num_accepted;
	else 
		av = 0.;
	
	std_dev = get_std_dev (num_accepted, sum_of_squares, av);
	
#ifdef DEBUG
	if (num_accepted != move_lim) {
		printf("Warning:  Starting t: %d of %d configurations accepted.\n",
			   num_accepted, move_lim);
	}
#endif
	
#ifdef VERBOSE
    printf("std_dev: %g, average cost: %g, starting temp: %g\n",
		   std_dev, av, 20. * std_dev);
#endif
	
	return (20. * std_dev); 
	/* return (15.225523	); */
}


static int try_swap (float t, float *cost, float *bb_cost, float *timing_cost, 
                     float *pwr_cost, float rlim, int *pins_on_block, 
					 int place_cost_type, float **old_region_occ_x, 
					 float **old_region_occ_y, int num_regions, boolean fixed_pins,
					 enum e_place_algorithm place_algorithm, int power_flag,
					 float timing_tradeoff, float power_tradeoff, 
					 float inverse_prev_bb_cost, float inverse_prev_timing_cost,
					 float inverse_prev_pwr_cost, float *delay_cost) {
	
	/* Picks some block and moves it to another spot.  If this spot is   *
	* occupied, switch the blocks.  Assess the change in cost function  *
	* and accept or reject the move.  If rejected, return 0.  If        *
	* accepted return 1.  Pass back the new value of the cost function. * 
	* rlim is the range limiter.  pins_on_block gives the number of     *
	* pins on each type of block (improves efficiency).                 */
	
	int move_height;
	int b_from, x_to, y_to, x_from, y_from, b_to; 
	int off_from, k, inet, keep_switch, io_num, num_of_pins;
	int num_nets_affected, bb_index,i, j, count;
	float delta_c, bb_delta_c, timing_delta_c, pwr_delta_c, delay_delta_c, newcost;
	static struct s_bb *bb_coord_new = NULL;
	static struct s_bb *bb_edge_new = NULL;
	static int *nets_to_update = NULL, *net_block_moved = NULL;
	static int *offset_y = NULL;
	int ppp, tmp;
	int from_list[MAX_CARRY_CHAIN_LENGTH];
	int to_list[MAX_CARRY_CHAIN_LENGTH];

	float timing_cost_check, delay_cost_check;
	int iconstraint;

	float check_bb_cost;

	int temp_width, temp_height, temp_iconstraint, temp_index, temp_x, temp_y;	
	int to_index, from_index;

	int temp_empty;
	int temp_occ;
	
	/* Allocate the local bb_coordinate storage, etc. only once. */
	
	ppp = MAX(pins_per_clb*MAX_CARRY_CHAIN_LENGTH,pins_per_memory);
	ppp = MAX(ppp,pins_per_dsp);

	for (iconstraint = 1; iconstraint < number_module; iconstraint ++)
	{
		ppp = MAX(ppp,constraint[iconstraint].num_pins);		
	}
	
	if (bb_coord_new == NULL) {
		bb_coord_new = (struct s_bb *) my_malloc (2 * ppp * 
												  sizeof (struct s_bb));
		bb_edge_new = (struct s_bb *) my_malloc (2 * ppp *
												 sizeof (struct s_bb));
		nets_to_update = (int *) my_malloc (2 * ppp * sizeof (int));
		net_block_moved = (int *) my_malloc (2 * ppp * sizeof (int));
		offset_y = (int *)my_malloc(2*ppp*sizeof(int));
	}
	
	b_from = my_irand(num_blocks - 1);
	while((block[b_from].carry_body)&&(!block[b_from].carry_head) || (block[b_from].fixed == 1)) {
	//while((block[b_from].carry_body)&&(!block[b_from].carry_head)) {
		//get another block to swap , since fixed block and carry body can't be moved
		b_from = my_irand(num_blocks - 1);
	}

	//printf ("try swap\n");
	
	/* If the pins are fixed we never move them from their initial    *
		* random locations.  The code below could be made more efficient *
		* by using the fact that pins appear first in the block list,    *
		* but this shouldn't cause any significant slowdown and won't be *
		* broken if I ever change the parser so that the pins aren't     *
		* necessarily at the start of the block list.                    */
	
	if (fixed_pins == TRUE) {
		while ((block[b_from].type != CLB) &&
			   (block[b_from].type != DSP) &&
			   (block[b_from].type != EMB) &&
			   (block[b_from].type != MEMORY)) {
			b_from = my_irand(num_blocks - 1);
			while((block[b_from].carry_body)&&(!block[b_from].carry_head)) {
				b_from = my_irand(num_blocks - 1);
			}
		}
	}

	//printf("Placing block with module_index = %d, b_from %d (%s) \n", block[b_from].module_index, b_from, block[b_from].name);
	/*Added by Yu Chi Wai to move veb */


	if (block[b_from].type == EMB && number_module > 1)
	//if (block[b_from].module_index > 0)
	{
		//the block is inside a veb
		//printf("Placing block with module_index = %d, b_from %d, block.fixed %d \n", block[b_from].module_index, b_from, block[b_from].fixed);
		//printf("constraint_lower_corner = %d, height %d, width %d \n", constraint[block[b_from].module_index].constraint_lower_corner, constraint[block[b_from].module_index].height, constraint[block[b_from].module_index].width);
		if ( (constraint[block[b_from].module_index].constraint_lower_corner == 1) 
			&& (constraint[block[b_from].module_index].height != -1) 
			&& (constraint[block[b_from].module_index].width != -1) )
			//&& constraint[block[b_from].module_index].fixed == 1)
		{
			//the veb which the block is in has fixed lower left corner and area
			//the block can only move inside the veb			
			x_from = block[b_from].x;
			y_from = block[b_from].y;
			
			//printf ("veb height %d\n",  constraint[block[b_from].module_index].height);
			//printf ("veb width %d\n",  constraint[block[b_from].module_index].width);
			//printf ("veb index %d\n",  block[b_from].module_index);

			//printf ("x_from %d\n",  x_from);
			//printf ("y_from %d\n",  y_from);
						
			find_to_veb (x_from, y_from, block[b_from].type, rlim, &x_to, &y_to, block[b_from].module_index, move_height);
			
			//printf ("find x_to %d\n",  x_to);
			//printf ("find y_to %d\n",  y_to);
			//printf("move block (%s) from  (%d, %d) to (%d, %d), carry_body %d \n", block[b_from].name, x_from, y_from, x_to, y_to, block[b_from].carry_body);
															
			if (block[b_from].type == EMB) 
			{
				from_list[0] = clb[x_from][y_from].u.block;
				
				if (clb[x_to ][y_to].occ) 
				{
					to_list[0] = clb[x_to][y_to].u.block;		
				}
				else 
				{
					to_list[0] = EMPTY;
				}
						
				io_num = -1;  // not used 
				move_height = 1;

				//printf("from_list[0] = %d (%s) , to_list[0] = %d  (%s)\n", from_list[0], block[from_list[0]].name, to_list[0], block[to_list[0]].name);

			}
		}
	}
	else
	{		
		if (block[b_from].carry_head) {
			move_height = block[b_from].carry_head ;
		} else {
			move_height = 1;
		}
		
		x_from = block[b_from].x;
		y_from = block[b_from].y;

		/*
		if (block[b_from].type == CLB) 
		{
			printf ("CLB block %d (%s), carrybody %d, carry_head %d, x_from %d, y_from %d, height %d\n",b_from, block[b_from].name, block[b_from].carry_body, block[b_from].carry_head,  x_from, y_from, move_height);
		}

		if (block[b_from].type == DSP)
		{
			printf ("DSP block %d (%s), carrybody %d, carry_head %d, x_from %d, y_from %d, height %d\n",b_from, block[b_from].name, block[b_from].carry_body, block[b_from].carry_head,  x_from, y_from, move_height);
		}
		if (block[b_from].type == INPAD)
		{
			printf ("INPAD block %d (%s), carrybody %d, carry_head %d, x_from %d, y_from %d, height %d\n",b_from, block[b_from].name, block[b_from].carry_body, block[b_from].carry_head,  x_from, y_from, move_height);
		}
		if (block[b_from].type == OUTPAD)
		{
			printf ("OUTPAD block %d (%s), carrybody %d, carry_head %d, x_from %d, y_from %d, height %d\n",b_from, block[b_from].name, block[b_from].carry_body, block[b_from].carry_head,  x_from, y_from, move_height);
		}
		if (block[b_from].type == MEMORY)
		{
			printf ("MEMORY block %d (%s), carrybody %d, carry_head %d, x_from %d, y_from %d, height %d\n",b_from, block[b_from].name, block[b_from].carry_body, block[b_from].carry_head,  x_from, y_from, move_height);
		}
		*/
		
		//printf ("y_from %d\n",  y_from);
				

		find_to (x_from, y_from, block[b_from].type, rlim, &x_to, &y_to, move_height);
/*
		if (clb[x_from][y_from].module_index != clb[x_to][y_to].module_index)
		{			
			printf("Error moving normal block to veb, from_module_index %d, to module_index %d\n", clb[x_from][y_from].module_index, clb[x_to][y_to].module_index);
		}
*/		
		if (move_height >= MAX_CARRY_CHAIN_LENGTH) {
			printf("Move height is %d which is larger than MAX_CARRY_CHAIN_LENGTH.  You can change this at the top of place.c\n");
			exit(1);
		}
		/*
		printf("type %d , move height %d \n", block[b_from].type, move_height);
		printf ("find x_to %d y_to %d\n",  x_to, y_to);
		*/
		//printf ("find y_to %d\n",  y_to);
		

		if (block[b_from].type == CLB) 
		{
			for(i=0;i<move_height;i++) {
				from_list[i] = clb[x_from][y_from+i].u.block;

				//printf("x_to %d, y_to %d , to block %d\n", x_to, y_to, clb[x_to][y_to+i].u.block);

				if (clb[x_to][y_to+i].type == EMB)
				{	
					printf("x_to %d, y_to %d , to block %d\n", x_to, y_to, clb[x_to][y_to+i].u.block);
					printf("error shouldn't choose EMB to move \n");
					exit(1);
				}

				if (clb[x_to][y_to+i].occ) 
				{
					to_list[i] = clb[x_to][y_to+i].u.block;
	
					b_to = to_list[i];  /* not used */
				} else {
					to_list[i] = EMPTY;
				}
			} 
			io_num = -1;  /* not used */

			//printf ("type CLB %d\n",  block[b_from].type);
			
		} else if ((block[b_from].type == DSP) ||
				(block[b_from].type == MEMORY)) {
			
			if (move_height != 1) {
				printf("Error: move height %d not implemented for non-CLB's\n",move_height);
				exit(1);
			}
			from_list[0] = b_from;
			
			if (clb[x_to][y_to].occ == 1) {
				to_list[0] = clb[x_to][y_to].u.block;    
				b_to = to_list[0];
			} else {
				to_list[0] = EMPTY;
			}
			io_num = -1;  /* not used */
			//printf ("type DSP %d\n",  block[b_from].type);
		} else if (block[b_from].type == EMB)
		{
			//nothing to do here , since do it before
		} else {  /* IO */
			if (move_height != 1) {
				printf("Error: move height %d not implemented for non-CLB's\n",move_height);
				exit(1);
			}						

			from_list[0] = b_from;
			io_num = my_irand(io_rat - 1);			

			if (io_num >= clb[x_to][y_to].occ) {  /* Moving to an empty location */
				b_to = EMPTY;
				to_list[0] = EMPTY;
			}
			else {          /* Swapping two blocks */
				b_to = *(clb[x_to][y_to].u.io_blocks+io_num);
				to_list[0] = b_to;
			}
			//printf ("type IO %d \n",  block[b_from].type);
		}
	}

/* Make the switch in order to make computing the new bounding *
* box simpler.  If the cost increase is too high, switch them *
* back.  (block data structures switched, clbs not switched   *
		  * until success of move is determined.)                       */


if ((block[b_from].type == CLB) ||
	(block[b_from].type == DSP) ||	
	(block[b_from].type == MEMORY)) 
{
	
    if (y_to + move_height - 1 > ny) {
		printf("Internal error 1 in placement algorithm.  Should never happen\n"); 
		//printf("y_to %d, move_height %d\n", y_to, move_height); 
		exit(0);
    }
	
    for(i=0;i<move_height;i++) {
		if (from_list[i] >= num_blocks) {
			printf("Internal error 2 in placement algorithm.  Should never happen\n"); exit(0);
		}
		if (clb[x_to][y_to+i].occ == 1) {         
			if (to_list[i] >= num_blocks) {
				printf("Internal error 2 in placement algorithm.  Should never happen\n"); exit(0);
			}
			
			block[from_list[i]].x = x_to;
			block[to_list[i]].x = x_from;
			block[from_list[i]].y = y_to+i;
			block[to_list[i]].y = y_from+i;

		} else {
			block[from_list[i]].x = x_to;
			block[from_list[i]].y = y_to+i;

//		if (from_list[i] == 2099 )
//		{
//			printf("move block %d from (%d, %d) to (%d, %d) \n", from_list[i], x_from, y_from, x_to, y_to);
//		}	

		}
    }	
}
else if (block[b_from].type == EMB) 
{
		
	if (clb[x_to  ][y_to ].occ == 1) 
	{
		block[b_from].x = x_to;
		block[b_from].y = y_to;
		block[to_list[0]].x = x_from;
		block[to_list[0]].y = y_from;
	}
	else
	{
		block[b_from].x = x_to;
		block[b_from].y = y_to;
	}	
}
else {   /* io block was selected for moving */

	if (move_height > 1) {
		printf("Error: move height > 1 not defined for I/O blocks yet\n");
		exit(1);
	}
	if (b_to == EMPTY){
		block[b_from].x = x_to;
		block[b_from].y = y_to;
	}
	else {          /* Swapping two blocks */
	block[b_to].x = x_from;
	block[b_to].y = y_from;
	block[b_from].x = x_to;
	block[b_from].y = y_to;



	//printf("------Move IO block, b_from %d, b_to %d, x_from %d, x_to %d, y_from %d, y_to %d\n", b_from, b_to, x_from, x_to, y_from, y_to);
	}
}

		/* Now update the cost function.  May have to do major optimizations *
		* here later.                                                       */

		/* I'm using negative values of temp_net_cost as a flag, so DO NOT   *
		* use cost functions that can go negative.                          */

		delta_c = 0;                    /* Change in cost due to this swap. */
		bb_delta_c = 0;
		timing_delta_c = 0;
		pwr_delta_c = 0;

		if (block[b_from].type == EMB)
		{	
			num_of_pins = constraint[block[b_from].module_index].num_pins;  
		}
		else
		{
			num_of_pins = pins_on_block[block[b_from].type];  
		}


		num_nets_affected = find_affected_nets (nets_to_update,
												net_block_moved,
												offset_y,
												from_list,
												to_list,
												move_height,
												num_of_pins);

		
		if (place_cost_type == NONLINEAR_CONG) {
			save_region_occ (old_region_occ_x, old_region_occ_y, num_regions);
		}

		//printf("block[%d].type = %d (%s), num_of_pins %d , num_nets_affected %d\n", b_from, block[b_from].type, block[b_from].name, num_of_pins, num_nets_affected);

		bb_index = 0;               /* Index of new bounding box. */

		for (k=0;k<num_nets_affected;k++) 
		{
			
				inet = nets_to_update[k];
				
				/* If we swapped two blocks connected to the same net, its bounding box *
				* doesn't change.                                                      */
			
				if (net_block_moved[k] == FROM_AND_TO) 
					continue;
				
				if (net[inet].num_pins <= SMALL_NET) {
					get_non_updateable_bb (inet, &bb_coord_new[bb_index]);
				}
				else {
					
					if (net_block_moved[k] == FROM) 
						update_bb (inet, &bb_coord_new[bb_index], &bb_edge_new[bb_index],
								x_from, y_from + offset_y[k], x_to, y_to+offset_y[k]);      
					else
						update_bb (inet, &bb_coord_new[bb_index], &bb_edge_new[bb_index],
								x_to, y_to + offset_y[k], x_from, y_from + offset_y[k]);      
				}
			
				if (place_cost_type != NONLINEAR_CONG) {
					temp_net_cost[inet] = get_net_cost (inet, &bb_coord_new[bb_index]);
					
					/* added by JL, April 2002 */
					if (power_flag == 1)
						pwr_delta_c += (temp_net_cost[inet] - net_cost[inet])*net[inet].density;
					
					bb_delta_c += temp_net_cost[inet] - net_cost[inet];
				}
				else {
					/* Rip up, then replace with new bb. */					
					update_region_occ (inet, &bb_coords[inet], -1, num_regions);  
					update_region_occ (inet, &bb_coord_new[bb_index],1, num_regions);
				}
			
				bb_index++;
		}   


		//printf("finished update region occ \n");
		if (place_cost_type == NONLINEAR_CONG) 
		{
				printf("NON linear cong \n");
				newcost = nonlinear_cong_cost(num_regions);
				bb_delta_c = newcost - *bb_cost;
		}

		if (place_algorithm == NET_TIMING_DRIVEN_PLACE ||
			place_algorithm == PATH_TIMING_DRIVEN_PLACE) 
		{   
				//printf("comp_delta_td_cost \n");
				/*in this case we redefine delta_c as a combination of timing and bb.  *
				*additionally, we normalize all values, therefore delta_c is in       *
				*relation to 1*/
				
				
				comp_delta_td_cost(from_list, to_list, num_of_pins, &timing_delta_c,	
								&delay_delta_c, move_height);
				
				//printf("finished comp_delta_td_cost \n");
				if (power_flag == 1)
				{
					delta_c = timing_tradeoff*timing_delta_c*inverse_prev_timing_cost + (1-timing_tradeoff)*
					((1.0-power_tradeoff)*bb_delta_c*inverse_prev_bb_cost +
						(power_tradeoff)*pwr_delta_c*inverse_prev_pwr_cost);
				}
				else
				{
					delta_c = timing_tradeoff*timing_delta_c*inverse_prev_timing_cost +
					(1-timing_tradeoff)*bb_delta_c*inverse_prev_bb_cost;
				} 
		}
		else 
		{
				delta_c = bb_delta_c;
		}

		//printf("check keep_switch \n");
		keep_switch = assess_swap (delta_c, t); 

		/* 1 -> move accepted, 0 -> rejected. */ 
		//printf("keep_switch %d \n", keep_switch);
		
//		if (from_list[0] == 2099)
//		{
//			if (keep_switch)
//				printf("block %d move accepted, num_net_affected %d\n", from_list[0], num_nets_affected);
//
//		}
//		if (to_list[0] == 2099)
//		{
//			if (keep_switch)
//				printf("move block %d to block %d accept, num_net_affected %d \n", from_list[0], to_list[0], num_nets_affected);
//
//		}	


		if (keep_switch) 
		{			

			*cost = *cost + delta_c;    
			//printf("original bb_cost %f \n", *bb_cost);
			*bb_cost = *bb_cost + bb_delta_c;
			
			//printf("				 bb_cost %g, bb_delta_c %g \n", *bb_cost, bb_delta_c);
			/*
			printf("1 check bb_cost %g \n", comp_bb_cost(CHECK, place_cost_type, num_regions));
			printf("2 check bb_cost %g \n---------- \n", comp_bb_cost(NORMAL, place_cost_type, num_regions));
			*/
			*pwr_cost = *pwr_cost + pwr_delta_c;
			
			
			if (place_algorithm == NET_TIMING_DRIVEN_PLACE ||
				place_algorithm == PATH_TIMING_DRIVEN_PLACE) {      
				/*update the point_to_point_timing_cost and point_to_point_delay_cost 
					values from the temporary values*/
				*timing_cost = *timing_cost + timing_delta_c;

				//printf("timing_cost %g, timing_delta_c %g \n", *timing_cost, timing_delta_c);

				*delay_cost = *delay_cost + delay_delta_c;
				
				//printf("updated_td_cost \n");
				update_td_cost(from_list, to_list, num_of_pins, move_height);
				
				
			}
			
			/* update net cost functions and reset flags. */
			
			bb_index = 0;
			
			for (k=0;k<num_nets_affected;k++) {
				inet = nets_to_update[k];
				
				/* If we swapped two blocks connected to the same net, its bounding box *
				* doesn't change.                                                      */
				
				if (net_block_moved[k] == FROM_AND_TO) {
					temp_net_cost[inet] = -1;  
					continue;
				}
				
				bb_coords[inet] = bb_coord_new[bb_index];
				if (net[inet].num_pins > SMALL_NET) 
					bb_num_on_edges[inet] = bb_edge_new[bb_index];
				
				bb_index++;
				
				net_cost[inet] = temp_net_cost[inet];
				temp_net_cost[inet] = -1;  
				
			}
			
			/* Update Clb data structures since we kept the move. */
			//printf("Update clb \n");	
			if ((block[b_from].type == CLB) ||
				(block[b_from].type == DSP) ||				
				(block[b_from].type == MEMORY)) {
				
				for(i=0;i<move_height;i++) {
					if (to_list[i] == EMPTY) {
						clb[x_to][y_to+i].u.block = from_list[i];

						clb[x_from][y_from+i].module_index = clb[x_to][y_to+i].module_index;					
						clb[x_to][y_to+i].module_index = block[from_list[i]].module_index;	//copy the module_index to destination clb
						

						clb[x_from][y_from+i].occ = 0; 
						clb[x_from][y_from+i].empty = 1;
						clb[x_to][y_to+i].occ = 1;				
						clb[x_to][y_to+i].empty = 0;
					} else {
						clb[x_from][y_from+i].module_index = clb[x_to][y_to+i].module_index;					
						clb[x_to][y_to+i].module_index = block[from_list[i]].module_index;	//copy the module_index to destination clb

						clb[x_from][y_from+i].u.block = to_list[i]; 
						clb[x_to][y_to+i].u.block = from_list[i];
						clb[x_from][y_from+i].occ = 1; 
						clb[x_from][y_from+i].empty = 0; 
						clb[x_to][y_to+i].occ = 1;
						clb[x_to][y_to+i].empty = 0;
					}
				}
			}
			else if (block[b_from].type == EMB)
			{
				//swap all the CLB for 2 EMBs
				//only one element in the to_list
				//assume all the EMBs are in the same size
				if (to_list[0] == EMPTY) 
				{					
					to_index = clb[x_to+i][y_to+j].module_index;
					//printf("to list is EMPTY \n");
					for (i=0 ; i < constraint[block[b_from].module_index].width ; i++)
					{
						for (j=0 ; j < constraint[block[b_from].module_index].height ; j++)
						{
							clb[x_to+i][y_to+j].u.block = from_list[0];
							clb[x_from + i][y_from+j].module_index = clb[x_to+i][y_to+j].module_index;					
							clb[x_to + i][y_to + j].module_index = block[from_list[0]].module_index;	//copy the module_index to destination clb

							temp_iconstraint = clb[x_from + i][y_from+j].iconstraint;							
							clb[x_from + i][y_from+j].iconstraint = clb[x_to + i][y_to+j].iconstraint;
							clb[x_to + i][y_to+j].iconstraint =  temp_iconstraint;				

							/*
							clb[x_from + i][y_from+j].occ = 0; 
							clb[x_from + i][y_from+j].empty = 1;
							clb[x_to + i][y_to+j].occ = 1;				
							clb[x_to + i][y_to+j].empty = 0;
							*/

							temp_empty = clb[x_from + i][y_from+j].empty;
							clb[x_from + i][y_from+j].empty= clb[x_to + i][y_to+j].empty;
							clb[x_to + i][y_to+j].empty = temp_empty;

							temp_occ = clb[x_from + i][y_from+j].occ;
							clb[x_from + i][y_from+j].occ = clb[x_to + i][y_to+j].occ;
							clb[x_to + i][y_to+j].occ = temp_occ;
						}
					}
					
					//handle the constraint
					from_index = block[b_from].module_index;
					//to_index = block[to_list[0]].module_index;
					temp_width = constraint[from_index].width;
					temp_height = constraint[from_index].height;
					temp_x = constraint[from_index].lx;
					temp_y = constraint[from_index].ly;					
					

					constraint[from_index].width = constraint[to_index].width;
					constraint[from_index].height = constraint[to_index].height;
					constraint[from_index].lx = constraint[to_index].lx;
					constraint[from_index].ly = constraint[to_index].ly;					
	
	

					constraint[to_index].width = temp_width;
					constraint[to_index].height = temp_height;
			
					constraint[to_index].lx = temp_x;
					constraint[to_index].ly = temp_y;					
					
					//printf("swap from %d to %d \n", from_index, to_index);
					//block[b_from].module_index = to_index;
					//handle the block 			
					/*
					temp_x = block[b_from].x;
					temp_y = block[b_from].y;

					block[b_from].x = block[to_list[0]].x;
					block[b_from].y = block[to_list[0]].y;
					block[to_list[0]].x = temp_x;
					block[to_list[0]].y = temp_y;
					*/

				}
				else
				{

					//swap all the CLB for 2 EMBs
					//only one element in the to_list
					//assume all the EMBs are in the same size
					
					//printf("to list is not EMPTY \n");
					for (i=0 ; i < constraint[block[b_from].module_index].width ; i++)
					{
						for (j=0 ; j < constraint[block[b_from].module_index].height ; j++)
						{
							clb[x_from + i][y_from+j].u.block = to_list[0]; 
							clb[x_to+i][y_to+j].u.block = from_list[0];
							temp_index = clb[x_from + i][y_from+j].module_index;
							clb[x_from + i][y_from+j].module_index = clb[x_to+i][y_to+j].module_index;					
							clb[x_to + i][y_to + j].module_index = temp_index;	//copy the module_index to destination clb

							temp_iconstraint = clb[x_from + i][y_from+j].iconstraint;							
							clb[x_from + i][y_from+j].iconstraint = clb[x_to + i][y_to+j].iconstraint;
							clb[x_to + i][y_to+j].iconstraint =  temp_iconstraint;		

							//temp_rr_build = clb[x_from + i][y_from+j].rr_build;					
							
							temp_empty = clb[x_from + i][y_from+j].empty;
							clb[x_from + i][y_from+j].empty= clb[x_to + i][y_to+j].empty;
							clb[x_to + i][y_to+j].empty = temp_empty;

							temp_occ = clb[x_from + i][y_from+j].occ;
							clb[x_from + i][y_from+j].occ = clb[x_to + i][y_to+j].occ;
							clb[x_to + i][y_to+j].occ = temp_occ;
							/*
							clb[x_from + i][y_from+j].empty = 0;							
							clb[x_to + i][y_to+j].empty = 0;									

							if (i == 0 && j ==0)
							{
								clb[x_from + i][y_from+j].occ = 1; 
								clb[x_to + i][y_to+j].occ = 1;				
							}
							else
							{
								clb[x_from + i][y_from+j].occ = 0; 
								clb[x_to + i][y_to+j].occ = 0;			
							}*/

							//printf("from CLB[%d][%d], to CLB[%d][%d] \n", x_from + i , y_from + j , x_to +i , y_to +j);

						}					
					}
					from_index = block[b_from].module_index;
					to_index = block[to_list[0]].module_index;

					//block[b_from].module_index = to_index;
					//block[to_list[0]].module_index = from_index;


					//handle the constraint
//					from_index = block[b_from].module_index;
//					to_index = block[to_list[0]].module_index;
			//printf("swap from %d, type %d to %d, type %d \n", from_index, constraint[from_index].type, to_index, constraint[to_index].type);					

					temp_width = constraint[from_index].width;
					temp_height = constraint[from_index].height;
					temp_x = constraint[from_index].lx;
					temp_y = constraint[from_index].ly;					

					constraint[from_index].width = constraint[to_index].width;
					constraint[from_index].height = constraint[to_index].height;
					constraint[from_index].lx = constraint[to_index].lx;
					constraint[from_index].ly = constraint[to_index].ly;					
					
					constraint[to_index].width = temp_width;
					constraint[to_index].height = temp_height;

					constraint[to_index].lx = temp_x;
					constraint[to_index].ly = temp_y;
					
/*
					b_to = to_list[0];
					printf("number of module %d\n", number_module);
					printf("from_index width = %d , height = %d \n", constraint[from_index].width , constraint[from_index].height);
					printf("to_index width = %d , height = %d \n", constraint[to_index].width , constraint[to_index].height);
					printf("from lx = %d , ly %d \n", constraint[from_index].lx, constraint[from_index].ly);
					printf("to lx = %d , ly %d \n", constraint[to_index].lx, constraint[to_index].ly);
					printf("from block.x  = %d , y %d \n",  block[b_from].x,  block[b_from].y);
					printf("to block.x  = %d , y %d \n",  block[b_to].x,  block[b_to].y);
					printf("clb[x_from ][y_from].module_index = %d \n", clb[x_from ][y_from].module_index);
					printf("clb[x_to ][y_to].module_index = %d \n", clb[x_to ][y_to].module_index);
					printf("clb[x_from ][y_from].iconstraint = %d \n", clb[x_from ][y_from].iconstraint);
					printf("clb[x_to ][y_to].iconstraint = %d \n", clb[x_to ][y_to].iconstraint);

					printf("----------------------------------------------------------------------------\n");
*/
										
/*
					//handle the block 					
					temp_x = block[b_from].x;
					temp_y = block[b_from].y;

					block[b_from].x = block[to_list[0]].x;
					block[b_from].y = block[to_list[0]].y;
					block[to_list[0]].x = temp_x;
					block[to_list[0]].y = temp_y;
*/
				}

			}			
			else {     /* io block was selected for moving */
				
				/* Get the "sub_block" number of the b_from block. */
				//printf("Move io block accept \n");

				for (off_from=0;;off_from++) {
					if (clb[x_from][y_from].u.io_blocks[off_from] == from_list[0]) break;
				}
				if (b_to != EMPTY) {   /* Swapped two blocks. */
					clb[x_to][y_to].u.io_blocks[io_num] = from_list[0];
					clb[x_from][y_from].u.io_blocks[off_from] = to_list[0];
				}
				else {                 /* Moved to an empty location */

						clb[x_to][y_to].u.io_blocks[clb[x_to][y_to].occ] = from_list[0];  
						clb[x_to][y_to].occ++;   
						for  (k=off_from;k<clb[x_from][y_from].occ-1;k++) { /* prevent gap  */
							clb[x_from][y_from].u.io_blocks[k] =          /* in io_blocks */
							clb[x_from][y_from].u.io_blocks[k+1];
						}
						clb[x_from][y_from].occ--;
				}
			}  
		}  

		else {    /* Move was rejected.  */

			//printf("Move reject, bb_cost %g\n", *bb_cost);	
			/* Reset the net cost function flags first. */
			for (k=0;k<num_nets_affected;k++) {
				inet = nets_to_update[k];
				temp_net_cost[inet] = -1;
			}

			/* Restore the block data structures to their state before the move. */


			for(i=0;i<move_height;i++) {
				if (to_list[i] == EMPTY) {
					block[from_list[i]].x = x_from;
					block[from_list[i]].y = y_from + i;
				} else {
					block[to_list[i]].x = x_to;
					block[to_list[i]].y = y_to + i;
					block[from_list[i]].x = x_from;
					block[from_list[i]].y = y_from + i;
				}
			}

			/* Restore the region occupancies to their state before the move. */
			if (place_cost_type == NONLINEAR_CONG) {
				restore_region_occ (old_region_occ_x, old_region_occ_y, num_regions);
			}			
		}

		//check_bb_cost = comp_bb_cost(CHECK, place_cost_type, num_regions);
		//printf("				 bb_cost %g \n", *bb_cost);
		//printf("bb_cost %g, check bb_cost %g, different  %g \n", *bb_cost, check_bb_cost, (*bb_cost - check_bb_cost));
		//	printf("2 check bb_cost %g \n---------- \n", comp_bb_cost(NORMAL, place_cost_type, num_regions));

		return(keep_switch);
}


static void save_region_occ (float **old_region_occ_x, 
							 float **old_region_occ_y, int num_regions) {
	
	/* Saves the old occupancies of the placement subregions in case the  *
	* current move is not accepted.  Used only for NONLINEAR_CONG.       */
	
	int i, j;
	
	for (i=0;i<num_regions;i++) { 
		for (j=0;j<num_regions;j++) { 
			old_region_occ_x[i][j] = place_region_x[i][j].occupancy; 
			old_region_occ_y[i][j] = place_region_y[i][j].occupancy; 
		} 
	} 
}


static void restore_region_occ (float **old_region_occ_x, 
								float **old_region_occ_y, int num_regions) {
	
	/* Restores the old occupancies of the placement subregions when the  *
	* current move is not accepted.  Used only for NONLINEAR_CONG.       */
	
	int i, j;
	
	for (i=0;i<num_regions;i++) {
		for (j=0;j<num_regions;j++) {
			place_region_x[i][j].occupancy = old_region_occ_x[i][j];
			place_region_y[i][j].occupancy = old_region_occ_y[i][j];
		}
	}
}


static int find_affected_nets (
							   int *nets_to_update, 
							   int *net_block_moved,
							   int *offset_y,
							   int *from_list, 
							   int *to_list, 
							   int move_height,
							   int num_of_pins) {
	
	int i, k, inet, affected_index, count;
	
	affected_index = 0;
	
	for(i=0;i<move_height;i++) {
		
		for (k=0;k<num_of_pins;k++) {
			
			inet = block[from_list[i]].nets[k];
			
			if (inet == OPEN) 
				continue;
			
			if (is_global[inet]) 
				continue;
			
			/* This is here in case the same block connects to a net twice. */
			
			if (temp_net_cost[inet] > 0.)  
				continue;
			
			nets_to_update[affected_index] = inet;
			net_block_moved[affected_index] = FROM;
			
			offset_y[affected_index] = i;
			
			affected_index++;
			temp_net_cost[inet] = 1.;         /* Flag to say we've marked this net. */
			
		}
	}
	
	for(i=0;i<move_height;i++) {
		if (to_list[i] != EMPTY) {
			
			for (k=0;k<num_of_pins;k++) {
				inet = block[to_list[i]].nets[k];
				
				if (inet == OPEN) 
					continue;
				
				if (is_global[inet]) 
					continue;
				
				if (temp_net_cost[inet] > 0.) {         /* Net already marked. */
					for (count=0;count<affected_index;count++) {
						if (nets_to_update[count] == inet) {
							if (net_block_moved[count] == FROM) 
								net_block_moved[count] = FROM_AND_TO;
							break;
						}
					}
					
					
					if (count > affected_index) {
						printf("Error in find_affected_nets -- count = %d, affected index = %d.\n", count, affected_index);
						exit (1);
					}
				}
				
				else {           /* Net not marked yet. */
					
					nets_to_update[affected_index] = inet;
					net_block_moved[affected_index] = TO;
					offset_y[affected_index] = i;
					affected_index++;
					temp_net_cost[inet] = 1.;    /* Flag means we've  marked net. */
				}
			}
		}
	}

return (affected_index);
}


static void find_to (int x_from, int y_from, int type, float rlim, 
					 int *x_to, int *y_to, int move_height) {
	
	/* Returns the point to which I want to swap, properly range limited. *
	* rlim must always be between 1 and nx (inclusive) for this routine  *
	* to work.                                                           */
	
	int x_rel, y_rel, iside, iplace, rlx, rly, ok, dest_block, i;

	int count;
	int trylim;
	
	rlx = min(nx,rlim);   /* Only needed when nx < ny. */
	rly = min (ny,rlim);  /* Added rly for aspect_ratio != 1 case. */

	trylim = (int) rlim * num_blocks;
	//rlx = nx;
	//rly = ny;
	
#ifdef DEBUG
	if (rlx < 1 || rlx > nx) {
		printf("Error in find_to: rlx = %d\n",rlx);
		exit(1);
	}
#endif
	
	ok = 0;
	count = 0;
	while (!ok) 
	{ 
		if ( (type == CLB) )// || (type == EMB))
		{
			count ++ ;

			x_rel = my_irand (2*rlx*move_height);    
			y_rel = my_irand (2*rly*move_height);
			*x_to = x_from - rlx + x_rel;
			*y_to = y_from - rly + y_rel;

			while (*x_to > nx) *x_to = *x_to - nx;    /* better spectral props. */
			while (*x_to < 1) *x_to = *x_to + nx;     /* than simple min, max   */
			while (*y_to < 1) *y_to = *y_to + (ny - move_height + 1);
			while (*y_to > (ny-move_height+1)) *y_to = *y_to - (ny-move_height+1); 

						
			//if (count >2000000)
			if (count >trylim)
			{				
				count = 0;
				//if the block can't find a suitable move fry trylim, then increase the rlx, rly to increase the possiblity of move
				//most probably , the block is surrounded by some carry chain
				rlx ++;
				rly ++;
/*
				printf("count %d, x_from %d, y_from %d, r_lim %g, x_to %d, y_to %d, type %d, move_height%d, rlx %d, rly %d, y_rel %d, x_rel %d , module_index %d\n",count ,  x_from, y_from, rlim, *x_to, *y_to, type, move_height, rlx, rly, x_rel, y_rel, clb[x_from][y_from].module_index);				
				printf("occ %d, type %d , empty %d \n",clb[*x_to][*y_to].occ, clb[*x_to][*y_to].type, clb[*x_to][*y_to].empty);
				dest_block = clb[*x_to][*y_to].u.block;
				printf("dest_block %d, type %d, carry_body %d, fixed %d, module_index %d \n", dest_block, block[dest_block].type,block[dest_block].carry_body, block[clb[*x_to][*y_to].u.block].fixed, clb[*x_to][*y_to].module_index);
				*/
			}

			/*
			if (count >1000010)
				exit(1);
			*/			
			ok = (clb[*x_to][*y_to].type == CLB);			
			//ok = (clb[*x_to][*y_to].type == CLB) & (clb[*x_to][*y_to].type != EMB);			
			
			if (ok) {                    
				for(i=0;i<move_height;i++) {
					
					if (clb[*x_to][*y_to+i].occ) {						
						
						dest_block = clb[*x_to][*y_to+i].u.block;
						
						ok = ok & (block[dest_block].type == CLB);
						ok = ok & (block[dest_block].carry_body == 0);											
						if (clb[*x_to][*y_to].occ)
							ok = ok & (block[clb[*x_to][*y_to].u.block].fixed == 0);
												
						ok = ok & (clb[x_from][y_from].module_index == clb[*x_to][*y_to].module_index);																	
					}					
					else
					{
						//it is not ok to place on clb occupied by veb
						ok = ok & (clb[*x_to][*y_to+i].empty == 1);				
					}
				}
			}
/*
			if (count >1000000)
			{
				printf("OK = %d \n", ok);
			}
			*/
		}
		else if (type == DSP) {
			if (move_height != 1) {
				printf("Error: move heights > 1 for DSP blocks not supported yet\n");
				exit(0);
			}
			x_rel = my_irand (4*rlx);    
			y_rel = my_irand (4*rly);
			*x_to = x_from - 2*rlx + x_rel;
			*y_to = y_from - 2*rly + y_rel;
			if (*x_to > nx) *x_to = *x_to - nx;    /* better spectral props. */       
			if (*x_to < 1) *x_to = *x_to + nx;     /* than simple min, max   */
			if (*y_to > ny) *y_to = *y_to - ny;    /* clipping.              */
			if (*y_to < 1) *y_to = *y_to + ny;
			if (*x_to > nx) *x_to = *x_to - nx;    /* better spectral props. */       
			if (*x_to < 1) *x_to = *x_to + nx;     /* than simple min, max   */
			if (*y_to > ny) *y_to = *y_to - ny;    /* clipping.              */
			if (*y_to < 1) *y_to = *y_to + ny;
			ok = (clb[*x_to][*y_to].type == DSP);
		}
		else if (type == MEMORY) {
			if (move_height != 1) {
				printf("Error: move heights > 1 for memory blocks not supported yet\n");
				exit(0);
			}
			x_rel = my_irand (4*rlx);    
			y_rel = my_irand (4*rly);
			*x_to = x_from - 2*rlx + x_rel;
			*y_to = y_from - 2*rly + y_rel;
			if (*x_to > nx) *x_to = *x_to - nx;    /* better spectral props. */
			if (*x_to < 1) *x_to = *x_to + nx;     /* than simple min, max   */
			if (*y_to > ny) *y_to = *y_to - ny;    /* clipping.              */
			if (*y_to < 1) *y_to = *y_to + ny;
			if (*x_to > nx) *x_to = *x_to - nx;    /* better spectral props. */
			if (*x_to < 1) *x_to = *x_to + nx;     /* than simple min, max   */
			if (*y_to > ny) *y_to = *y_to - ny;    /* clipping.              */
			if (*y_to < 1) *y_to = *y_to + ny;
			ok = (clb[*x_to][*y_to].type == MEMORY);
		}
		else {                 /* io_block to be moved. */
			ok = 1;
			if (move_height != 1) {
				printf("Error: move heights > 1 for io blocks not supported yet\n");
				exit(0);
			}
			
			if (rlx >= nx) {
				iside = my_irand(3);
				/*                              *
				*       +-----1----+           *
				*       |          |           *
				*       |          |           *
				*       0          2           *
				*       |          |           *
				*       |          |           *
				*       +-----3----+           *
				*                              */
				switch (iside) {
					case 0:
						iplace = my_irand (ny-1) + 1;
						*x_to = 0;
						*y_to = iplace;
						break;
					case 1:
						iplace = my_irand (nx-1) + 1;
						*x_to = iplace;
						*y_to = ny+1;
						break;
					case 2:
						iplace = my_irand (ny-1) + 1;
						*x_to = nx+1;
						*y_to = iplace;
						break;
					case 3:
						iplace = my_irand (nx-1) + 1;
						*x_to = iplace;
						*y_to = 0;
						break;
					default:
						printf("Error in find_to.  Unexpected io swap location.\n");
						exit (1);
				}
			}
			else {   /* rlx is less than whole chip */
				if (x_from == 0) {
					iplace = my_irand (2*rly);
					*y_to = y_from - rly + iplace;
					*x_to = x_from;
					if (*y_to > ny) {
						*y_to = ny + 1;
						*x_to = my_irand (rlx - 1) + 1;
					}
					else if (*y_to < 1) {
						*y_to = 0;
						*x_to = my_irand (rlx - 1) + 1;
					}
				}
				else if (x_from == nx+1) {
					iplace = my_irand (2*rly);
					*y_to = y_from - rly + iplace;
					*x_to = x_from;
					if (*y_to > ny) {
						*y_to = ny + 1;
						*x_to = nx - my_irand (rlx - 1); 
					}
					else if (*y_to < 1) {
						*y_to = 0;
						*x_to = nx - my_irand (rlx - 1);
					}
				}
				else if (y_from == 0) {
					iplace = my_irand (2*rlx);
					*x_to = x_from - rlx + iplace;
					*y_to = y_from;
					if (*x_to > nx) {
						*x_to = nx + 1;
						*y_to = my_irand (rly - 1) + 1;
					}
					else if (*x_to < 1) {
						*x_to = 0;
						*y_to = my_irand (rly -1) + 1;
					}
				}
				else {  /* *y_from == ny + 1 */
					iplace = my_irand (2*rlx);
					*x_to = x_from - rlx + iplace;
					*y_to = y_from;
					if (*x_to > nx) {
						*x_to = nx + 1;
						*y_to = ny - my_irand (rly - 1);
					}
					else if (*x_to < 1) {
						*x_to = 0;
						*y_to = ny - my_irand (rly - 1);
					}
				}
			}    /* End rlx if */
		}    /* end type if */
		if ((x_from == *x_to) && (y_from == *y_to)) {
			ok = 0;
		}
	} 

#ifdef DEBUG
if (*x_to < 0 || *x_to > nx+1 || *y_to < 0 || *y_to > ny+1) {
	printf("Error in routine find_to:  (x_to,y_to) = (%d,%d)\n",
		   *x_to, *y_to);
	exit(1);
}

if (type == CLB) {
	if (clb[*x_to][*y_to].type != CLB) {
        printf("Error: Moving CLB to illegal type block at (%d,%d)\n",
			   *x_to,*y_to);
        exit(1);
	}
} else if (type == DSP) { 
	if (clb[*x_to][*y_to].type != DSP) {
        printf("Error: Moving DSP to illegal type block at (%d,%d)\n",
			   *x_to,*y_to);
        exit(1);
	}
} else if (type == MEMORY) { 
	if (clb[*x_to][*y_to].type != MEMORY) {
        printf("Error: Moving MEMORY to illegal type block at (%d,%d)\n",
			   *x_to,*y_to);
        exit(1);
	}
} else {
	if (clb[*x_to][*y_to].type != IO) {
        printf("Error: Moving IO block to illegal type location at "
			   "(%d,%d)\n", *x_to, *y_to);
        exit(1);
	}
}
#endif

/* printf("(%d,%d) moved to (%d,%d)\n",x_from,y_from,*x_to,*y_to); */
}


static int assess_swap (float delta_c, float t) {
	
	/* Returns: 1 -> move accepted, 0 -> rejected. */ 
	
	int accept;
	float prob_fac, fnum;
	
	if (delta_c <= 0) {
		
#ifdef SPEC          /* Reduce variation in final solution due to round off */
		fnum = my_frand();
#endif
		
		accept = 1;
		return(accept);
	}
	
	if (t == 0.) 
		return(0);
	
	fnum = my_frand();
	prob_fac = exp(-delta_c/t);
	if (prob_fac > fnum) {
		accept = 1;
	}
	else {
		accept = 0;
	}
	return(accept);
}


static float recompute_bb_cost (int method, int place_cost_type, int 
								num_regions) {
	
	/* Recomputes the cost to eliminate roundoff that may have accrued.  *
	* This routine does as little work as possible to compute this new  *
	* cost.                                                             */
	
	int i, j, inet;
	float cost;
	
	cost = 0;
	
	/* Initialize occupancies to zero if regions are being used. */
	
	if (place_cost_type == NONLINEAR_CONG) {
		for (i=0;i<num_regions;i++) {
			for (j=0;j<num_regions;j++) {
				place_region_x[i][j].occupancy = 0.;
				place_region_y[i][j].occupancy = 0.;
			}
		}
	}    
	
	for (inet=0;inet<num_nets;inet++) {     /* for each net ... */
		
		if (is_global[inet] == FALSE) {    /* Do only if not global. */
			
			/* Bounding boxes don't have to be recomputed; they're correct. */ 
			
			if (place_cost_type != NONLINEAR_CONG) {
				if (method == POWER)
					cost += net_cost[inet]*net[inet].density;
				else
					cost += net_cost[inet];
			} 
			else {      /* Must be nonlinear_cong case. */
				update_region_occ (inet, &bb_coords[inet], 1, num_regions);
			} 
		}
	}

if (place_cost_type == NONLINEAR_CONG) {
    cost = nonlinear_cong_cost (num_regions);
}

return (cost);
}


static float comp_td_point_to_point_delay (int inet, int ipin) {
	
	/*returns the delay of one point to point connection */
	
	int source_block, sink_block;
	int delta_x, delta_y;
	enum e_block_types source_type, sink_type;
	float delay_source_to_sink;
	
	
	delay_source_to_sink = 0.;
	//printf("comp_td_point_to_point_delay, inet %d , ipin %d\n", inet, ipin);
	source_block = net[inet].blocks[0];
	//printf("source_block %d \n", source_block);
	source_type = block[source_block].type;
		
	sink_block = net[inet].blocks[ipin];
	//printf("sink_block %d \n", sink_block);
	sink_type = block[sink_block].type;

	//printf(" INPAD %d, OUTPAD %d, CLB %d , DSP %d, EMB %d MEMORY %d \n", INPAD, OUTPAD, CLB, DSP, EMB, MEMORY);
	//printf ("sink_block %d , type %d, source_block %d, type %d \n", sink_block, sink_type, source_block, source_type);	
	//printf ("sink_block.x %d ,  sink_block.y %d \n source_block.x %d, source_block.y %d \n", block[sink_block].x, block[sink_block].y, block[source_block].x, block[source_block].y);
	
	//printf ("sink_block %d, type %d,  sink_block.x %d ,  sink_block.y %d \n ", sink_block, sink_type,  block[sink_block].x, block[sink_block].y);

	delta_x = abs(block[sink_block].x - block[source_block].x);
	delta_y = abs(block[sink_block].y - block[source_block].y);

	//printf("delta_x %d,  delta_y %d \n", delta_x, delta_y);
	
	if ((source_type == CLB) ||
		(source_type == DSP) ||
		(source_type == EMB) ||
		(source_type == MEMORY)) 
	{
		if ((sink_type == CLB) ||
			(sink_type == DSP) ||
			(sink_type == EMB) ||
			(sink_type == MEMORY)) {      
			delay_source_to_sink = delta_clb_to_clb[delta_x][delta_y];

			//printf("delay_source %f \n", delay_source_to_sink);
		}
		else if (sink_type == OUTPAD) {
			//printf ("OUTPAD: sink_block %d , type %d, source_block %d, type %d \n", sink_block, sink_type, source_block, source_type);
	
			delay_source_to_sink = delta_clb_to_outpad[delta_x][delta_y];

			//printf("OUTPAD : delay_source %f \n", delay_source_to_sink);
		}
		else {
			printf("xError in comp_td_point_to_point_delay in place.c, bad sink_type %d.\n", sink_type);
			printf("source_block: %d, inet: %d, sink_block: %d, ipin: %d, source_type: %d.\n", source_block, inet, sink_block, ipin, source_type);
			printf("CLB: %d, DSP: %d, MEMORY: %d, INPAD: %d", CLB, DSP, MEMORY, INPAD);
			exit(1);
		}
	}
	else if (source_type == INPAD) {
		if ((sink_type == CLB) ||
			(sink_type == DSP) ||
			(sink_type == EMB) ||
			(sink_type == MEMORY)) {
			delay_source_to_sink = delta_inpad_to_clb[delta_x][delta_y];
		}
		else if (sink_type == OUTPAD) {
			delay_source_to_sink = delta_inpad_to_outpad[delta_x][delta_y];
		}
		else {
			printf("yError in comp_td_point_to_point_delay in place.c, bad sink_type %d.\n", sink_type);
			printf("source_block: %d, inet: %d, sink_block: %d, ipin: %d, source_type: %d.\n", source_block, inet, sink_block, ipin, source_type);
			printf("CLB: %d, DSP: %d, MEMORY: %d, INPAD: %d", CLB, DSP, MEMORY, INPAD);
			exit(1);
		}
	}
	else {
		printf("Error in comp_td_point_to_point_delay in place.c, bad source_type\n");
		exit(1);
	}

	//printf("delay_source %f \n", delay_source_to_sink);

	if (delay_source_to_sink < 0) {
		printf("Error in comp_td_point_to_point_delay in place.c, bad delay_source_to_sink value\n");
		exit(1);
	}
	
	if (delay_source_to_sink < 0.) {
		printf("Error in comp_td_point_to_point_delay in place.c, delay is less than 0\n");
		exit(1);
	}
	
	return(delay_source_to_sink);
}



static void update_td_cost(int *from_list, int *to_list, int num_of_pins, int move_height) {
	/*update the point_to_point_timing_cost values from the temporary */
	/*values for all connections that have changed */  
	
	int blkpin, net_pin, inet, ipin, i;
	int count, check;
    
	for(i=0;i<move_height;i++) {
		
		for (blkpin=0; blkpin<num_of_pins; blkpin++) {
			
			inet = block[from_list[i]].nets[blkpin];
			
			if (inet == OPEN) 
				continue;
			
			if (is_global[inet]) 
				continue;
			
			net_pin = net_pin_index[from_list[i]][blkpin];
			
			if (net_pin != 0) { 
				
				/*the following "if" prevents the value from being updated twice*/
				check = 1;
				for (count = 0; count < move_height ; count ++)
				{
					if (net[inet].blocks[0] == to_list[count] || net[inet].blocks[0] == from_list[count]) 
						check = 0;
				}

				//if (net[inet].blocks[0] != to_list[i] && net[inet].blocks[0] != from_list[i]) 
				if (check == 1)
				{
					//printf("from_list : temp_point_to_point_delay_cost,  pp_delay %8.4g\n",  comp_td_point_to_point_delay(inet, net_pin));
					point_to_point_delay_cost[inet][net_pin] = temp_point_to_point_delay_cost[inet][net_pin];

					//printf("from_list : temp_point_to_point_delay_cost, inet %d (%s), net_pin %d, delay_cost %8.4g , pp_delay %8.4g\n", inet, net[inet].name, net_pin, temp_point_to_point_delay_cost[inet][net_pin], comp_td_point_to_point_delay(inet, net_pin));
					temp_point_to_point_delay_cost[inet][net_pin] = -1;
					
					point_to_point_timing_cost[inet][net_pin] = 
					temp_point_to_point_timing_cost[inet][net_pin];
					temp_point_to_point_timing_cost[inet][net_pin] = -1;
/*
					if (inet == 2401)
					{						
						printf("**** 1 point_to_point_timing_cost[%d][%d] %g \n",  inet, net_pin, point_to_point_timing_cost[inet][net_pin]);
					}
					*/
				}
			}
			else { /*this net is being driven by a moved block, recompute */
				/*all point to point connections on this net.*/
				for (ipin=1; ipin<net[inet].num_pins; ipin++) {
					
					//printf("from_list: temp_point_to_point_delay_cost,  pp_delay %8.4g\n", comp_td_point_to_point_delay(inet, ipin));
					point_to_point_delay_cost[inet][ipin] =  temp_point_to_point_delay_cost[inet][ipin];

					//printf("from_list: temp_point_to_point_delay_cost, inet %d (%s), ipin %d, delay_cost %8.4g , pp_delay %8.4g\n", inet, net[inet].name,ipin, temp_point_to_point_delay_cost[inet][ipin], comp_td_point_to_point_delay(inet, ipin));

					temp_point_to_point_delay_cost[inet][ipin] = -1;
					
					point_to_point_timing_cost[inet][ipin] = 
						temp_point_to_point_timing_cost[inet][ipin];
					temp_point_to_point_timing_cost[inet][ipin] = -1;
/*
					if (inet == 2401)
					{						
						printf("**** 2 point_to_point_timing_cost[%d][%d] %g \n",  inet, ipin, point_to_point_timing_cost[inet][ipin]);
					}
					*/
				}
			}
		}
	}

for(i=0;i<move_height;i++) {
    if (to_list[i] != EMPTY) {
		for (blkpin=0; blkpin<num_of_pins; blkpin++) {
			
			inet = block[to_list[i]].nets[blkpin];
			
			if (inet == OPEN) 
				continue;
			
			if (is_global[inet]) 
				continue;
			
			net_pin = net_pin_index[to_list[i]][blkpin];
			
			if (net_pin != 0) { 
				
				/*the following "if" prevents the value from being updated 2x*/

				check = 1;
				for (count = 0; count < move_height ; count ++)
				{
					if (net[inet].blocks[0] == to_list[count] || net[inet].blocks[0] == from_list[count]) 
						check = 0;
				}

				if (check ==1)
				//if (net[inet].blocks[0] != to_list[i] && net[inet].blocks[0] != from_list[i]) 
				{					
					//printf("to_list : temp_point_to_point_delay_cost,  pp_delay %8.4g\n",  comp_td_point_to_point_delay(inet, net_pin));
					point_to_point_delay_cost[inet][net_pin] = temp_point_to_point_delay_cost[inet][net_pin];

					//printf("temp_point_to_point_delay_cost, inet %d, net_pin %d, delay_cost %8.8g , pp_delay %8.8g\n", inet, net_pin, temp_point_to_point_delay_cost[inet][net_pin], comp_td_point_to_point_delay(inet, net_pin));

					temp_point_to_point_delay_cost[inet][net_pin] = -1;
					
					point_to_point_timing_cost[inet][net_pin] = 
						temp_point_to_point_timing_cost[inet][net_pin];
					temp_point_to_point_timing_cost[inet][net_pin] = -1;
/*
					if (inet == 2401)
					{						
						printf("**** 3 point_to_point_timing_cost[%d][%d] %g \n",  inet, net_pin, point_to_point_timing_cost[inet][net_pin]);
					}
					*/
				}
			}
			else { /*this net is being driven by a moved block, recompute */
				/*all point to point connections on this net.*/
				for (ipin=1; ipin<net[inet].num_pins; ipin++) {
					
					//printf("to_list: temp_point_to_point_delay_cost,  pp_delay %8.4g\n", comp_td_point_to_point_delay(inet, ipin));

					point_to_point_delay_cost[inet][ipin] =  temp_point_to_point_delay_cost[inet][ipin];

					//printf("temp_point_to_point_delay_cost, inet %d, ipin %d, delay_cost %8.8g , pp_delay %8.8g\n", inet, ipin, temp_point_to_point_delay_cost[inet][ipin], comp_td_point_to_point_delay(inet, ipin));

					temp_point_to_point_delay_cost[inet][ipin] = -1;
					
					point_to_point_timing_cost[inet][ipin] = 
						temp_point_to_point_timing_cost[inet][ipin];
					temp_point_to_point_timing_cost[inet][ipin] = -1;
/*
					if (inet == 2401)
					{						
						printf("**** 4 point_to_point_timing_cost[%d][%d] %g \n",  inet, ipin, point_to_point_timing_cost[inet][ipin]);
					}
*/
				}
			}
		}
    }
}
}


static void comp_delta_td_cost(int *from_list, int *to_list, 
							   int num_of_pins, float *delta_timing, 
							   float *delta_delay, int move_height) {
	
	/*a net that is being driven by a moved block must have all of its  */
	/*sink timing costs recomputed. A net that is driving a moved block */
	/*must only have the timing cost on the connection driving the input*/
	/*pin computed*/
	
	int inet, i,k, net_pin, ipin;
	float delta_timing_cost, delta_delay_cost, temp_delay;
	int check, count;
	
	delta_timing_cost = 0.;
	delta_delay_cost = 0.;
	
	
	//printf("comp_delta_td_cost, num_of_pins %d, move_height %d, net_pin_index %d \n", num_of_pins, move_height, net_pin_index[0][0]);
	for(i=0;i<move_height;i++) 
	{
		
		for (k=0;k<num_of_pins;k++) 
		{ 
			inet = block[from_list[i]].nets[k];
			
			if (inet == OPEN) 
				continue;
			
			if (is_global[inet]) 
				continue;
			
			net_pin = net_pin_index[from_list[i]][k];

			//printf("from_list %d (%s), inet %d (%s), net_pin %d \n", from_list[i], block[from_list[i]].name, inet, net[inet].name, net_pin);
			
			if (net_pin > 0) 
			{ /*this net is driving a moved block               */
				
				/*if this net is being driven by a block that has moved, we do not  */
				/*need to compute the change in the timing cost (here) since it will*/
				/*be computed in the fanout of the net on  the driving block, also  */     
				/*computing it here would double count the change, and mess up the  */
				/*delta_timing_cost value */ 

				check = 1;
				for (count = 0; count < move_height ; count ++)
				{
					if (net[inet].blocks[0] == to_list[count] || net[inet].blocks[0] == from_list[count]) 
						check = 0;
				}

				//printf("Check %d\n", check);
				//if (net[inet].blocks[0] != to_list[i] && net[inet].blocks[0] != from_list[i]) 
				if (check ==1)
				{
				
					temp_delay = comp_td_point_to_point_delay(inet, net_pin);
					//printf("temp_delay %g \n", temp_delay);
					
					temp_point_to_point_delay_cost[inet][net_pin] = temp_delay;
					temp_point_to_point_timing_cost[inet][net_pin] =
						timing_place_crit[inet][net_pin] * temp_delay;
					
					delta_delay_cost += temp_point_to_point_delay_cost[inet][net_pin] -
						point_to_point_delay_cost[inet][net_pin];
					
					delta_timing_cost += temp_point_to_point_timing_cost[inet][net_pin] - 
						point_to_point_timing_cost[inet][net_pin];
					
					//if (delta_timing_cost > 1.0)
					/*
					if (inet==2401)
					{							
						printf("k %d , move_height %d\n", k, move_height);
						printf("---1, iblk %d,  temp_delay %g , delta_timing_cost %g, point_to_point_timing_cost[%d][%d] %g, temp_point_to_point_timing_cost[%d][%d] %g\n", 
							from_list[i], temp_delay, delta_timing_cost, point_to_point_timing_cost[inet][net_pin],inet, net_pin, temp_point_to_point_timing_cost[inet][net_pin], inet, net_pin);
					}
*/
				}
			}
			else { /*this net is being driven by a moved block, recompute */
				/*all point to point connections on this net.*/
				//printf("net %d, num_pins %d \n", inet , net[inet].num_pins);
				for (ipin=1; ipin<net[inet].num_pins; ipin++) {
					
					temp_delay = comp_td_point_to_point_delay(inet, ipin);
					//printf("temp_delay %g \n", temp_delay);
					
					temp_point_to_point_delay_cost[inet][ipin] = temp_delay;
					temp_point_to_point_timing_cost[inet][ipin] =
						timing_place_crit[inet][ipin] * temp_delay;
					
					delta_delay_cost += temp_point_to_point_delay_cost[inet][ipin] -
						point_to_point_delay_cost[inet][ipin];
					
					delta_timing_cost += temp_point_to_point_timing_cost[inet][ipin] - 
						point_to_point_timing_cost[inet][ipin];

					//if (delta_timing_cost > 1.0)
					/*
					if (inet==2401)
					{						
						printf("k %d , move_height %d\n", k, move_height);
						printf("---2, iblk %d,  temp_delay %g , delta_timing_cost %g, point_to_point_timing_cost[%d][%d] %g, temp_point_to_point_timing_cost[%d][%d] %g\n", 
							from_list[i], temp_delay, delta_timing_cost, point_to_point_timing_cost[inet][ipin],inet, ipin, temp_point_to_point_timing_cost[inet][ipin], inet, ipin);
					}					
					*/
				}
			}
		}
	}

	//printf("moving height %d\n", move_height);
		for(i=0;i<move_height;i++) 
		{
			//printf("to_list[%d] = %d\n", i, to_list[i]);
			if (to_list[i] != EMPTY) 
			{
				for (k=0;k<num_of_pins;k++) { 
					inet = block[to_list[i]].nets[k];
					
					if (inet == OPEN) 
						continue;
					
					if (is_global[inet]) 
						continue;
					
					net_pin = net_pin_index[to_list[i]][k];

					//printf("moving height net_pin %d \n", net_pin );
					
					if (net_pin > 0) { /*this net is driving a moved block*/
						
						/*if this net is being driven by a block that has moved, we do not */
						/*need to compute the change in the timing cost (here) since it was*/
						/*computed in the fanout of the net on  the driving block, also    */     
						/*computing it here would double count the change, and mess up the */
						/*delta_timing_cost value */ 

						check = 1;
						for (count = 0; count < move_height ; count ++)
						{
							if (net[inet].blocks[0] == to_list[count] || net[inet].blocks[0] == from_list[count]) 
								check = 0;
						}

						if (check == 1)
						//if (net[inet].blocks[0] != to_list[i] && net[inet].blocks[0] != from_list[i]) 
						{
								
							temp_delay = comp_td_point_to_point_delay(inet, net_pin);
							//printf("temp_delay %d \n", temp_delay );
							temp_point_to_point_delay_cost[inet][net_pin] = temp_delay;
							temp_point_to_point_timing_cost[inet][net_pin] =
								timing_place_crit[inet][net_pin] * temp_delay;
							delta_delay_cost += temp_point_to_point_delay_cost[inet][net_pin] -
								point_to_point_delay_cost[inet][net_pin];
							delta_timing_cost += temp_point_to_point_timing_cost[inet][net_pin] - 
								point_to_point_timing_cost[inet][net_pin];

								//if (delta_timing_cost > 1.0)
							/*
								if(inet == 2401)
								{						
									printf("---3, iblk %d,  temp_delay %g , delta_timing_cost %g, point_to_point_timing_cost[%d][%d] %g, temp_point_to_point_timing_cost[%d][%d] %g\n", 
									from_list[i], temp_delay, delta_timing_cost, point_to_point_timing_cost[inet][net_pin],inet, net_pin, temp_point_to_point_timing_cost[inet][net_pin], inet, net_pin);
								}
							*/	
						}
					}
					else { /*this net is being driven by a moved block, recompute */
						/*all point to point connections on this net.*/
						for (ipin=1; ipin<net[inet].num_pins; ipin++) {
							
							temp_delay = comp_td_point_to_point_delay(inet, ipin);
							
							temp_point_to_point_delay_cost[inet][ipin] = temp_delay;
							temp_point_to_point_timing_cost[inet][ipin] =
								timing_place_crit[inet][ipin] * temp_delay;
							
							
							delta_delay_cost += temp_point_to_point_delay_cost[inet][ipin] -
								point_to_point_delay_cost[inet][ipin];
							delta_timing_cost += temp_point_to_point_timing_cost[inet][ipin] - 
								point_to_point_timing_cost[inet][ipin];

									//if (delta_timing_cost > 1.0)
							/*
									if(inet == 2401)
									{						
										printf("---4, iblk %d,  temp_delay %g , delta_timing_cost %g, point_to_point_timing_cost[%d][%d] %g, temp_point_to_point_timing_cost[%d][%d] %g\n", 
										from_list[i], temp_delay, delta_timing_cost, point_to_point_timing_cost[inet][ipin],inet, ipin, temp_point_to_point_timing_cost[inet][ipin], inet, ipin);
									}									
									*/
						}
					}
				}
			}
		}

		*delta_timing = delta_timing_cost;
		*delta_delay = delta_delay_cost;
}

static void comp_td_costs (float *timing_cost, float *connection_delay_sum) {
	/*computes the cost (from scratch) due to the delays and criticalities*
	*on all point to point connections, we define the timing cost of     *
	*each connection as criticality*delay */
	
	int inet , ipin;
	float loc_timing_cost, loc_connection_delay_sum, temp_delay_cost, temp_timing_cost;
	
	loc_timing_cost = 0.;
	loc_connection_delay_sum = 0.;
	
	for (inet=0;inet<num_nets;inet++) {     /* for each net ... */
		if ( (is_global[inet] == FALSE)  )
		{    /* Do only if not global. */
			
			for (ipin=1; ipin<net[inet].num_pins; ipin++) {
				
				temp_delay_cost = comp_td_point_to_point_delay(inet, ipin);
				
				temp_timing_cost = temp_delay_cost *timing_place_crit[inet][ipin];
/*
				if (inet == 2401)
				{
					printf("comp_td_point_to_point_delay, inet %d, ipin %d, temp_delay_cost %8.4g, temp_timing_cost %8.4g \n", inet, ipin, temp_delay_cost, temp_timing_cost);
				}
*/				
				loc_connection_delay_sum += temp_delay_cost;
				point_to_point_delay_cost[inet][ipin] = temp_delay_cost;
				temp_point_to_point_delay_cost[inet][ipin] = -1; /*undefined*/
				
				point_to_point_timing_cost[inet][ipin] = temp_timing_cost;
				temp_point_to_point_timing_cost[inet][ipin] = -1; /*undefined*/
				loc_timing_cost += temp_timing_cost;
			}
		}
	}

*timing_cost = loc_timing_cost;
*connection_delay_sum = loc_connection_delay_sum;
}


static float comp_bb_cost (int method, int place_cost_type, int num_regions) 
{
	
	/* Finds the cost from scratch.  Done only when the placement   *
	* has been radically changed (i.e. after initial placement).   *
	* Otherwise find the cost change incrementally.  If method     *
	* check is NORMAL, we find bounding boxes that are updateable  *
	* for the larger nets.  If method is CHECK, all bounding boxes *
	* are found via the non_updateable_bb routine, to provide a    *
	* cost which can be used to check the correctness of the       *
	* other routine.                                               */
	
	int i, j, k; 
	float cost;
	
	cost = 0;
	
	/* Initialize occupancies to zero if regions are being used. */
	
	if (place_cost_type == NONLINEAR_CONG) 
	{
		printf("comp_bb_cost : NONLINEAR_CONG \n");
		for (i=0;i<num_regions;i++) 
		{
			for (j=0;j<num_regions;j++) 
			{
				place_region_x[i][j].occupancy = 0.;
				place_region_y[i][j].occupancy = 0.;
			}
		}
	}

	//printf("comp_bb_cost : num_nets %d \n", num_nets);
	
	for (k=0;k<num_nets;k++) 
	{     /* for each net ... */
		
		if ( (is_global[k] == FALSE)) 
		{    /* Do only if not global. */
			
			/* Small nets don't use incremental updating on their bounding boxes, *
			* so they can use a fast bounding box calculator.                    */
			//fprintf("net[%d].num_pins %d\n", k, net[k].num_pins);

			if (net[k].num_pins > SMALL_NET && 
				((method == NORMAL) || (method == POWER))) {
				//printf("get_bb_from_scratch \n");
				get_bb_from_scratch (k, &bb_coords[k], &bb_num_on_edges[k]);
			}
			else {
				//printf("get_non_updateable_bb, k %d, num_pins %d \n", k, net[k].num_pins);
				get_non_updateable_bb (k, &bb_coords[k]);
			}
			
			if (place_cost_type != NONLINEAR_CONG) 
			{
				net_cost[k] = get_net_cost(k, &bb_coords[k]);
				if (method == POWER)
				{
					/* Added by JL, October 2002 */
					cost += net_cost[k]*net[k].density;
					//printf("Not NONLINEAR_CONF cost %f, net_cost[%d] %f \n", cost, k, net_cost[k]);
				}
				else
				{
					cost += net_cost[k];
					//printf("Not NONLINEAR_CONF cost %f, net_cost[%d] %f \n", cost, k, net_cost[k]);
				}
			}
			else {      /* Must be nonlinear_cong case. */
				update_region_occ (k, &bb_coords[k], 1, num_regions);
				//printf("NONLINEAR_CONF cost %f, net_cost[%d] %f \n", cost, k, net_cost[k]);
			}
		}
	}

	if (place_cost_type == NONLINEAR_CONG) {
		cost = nonlinear_cong_cost (num_regions);
	}

	return (cost);
}


static float nonlinear_cong_cost (int num_regions) {
	
	/* This routine computes the cost of a placement when the NONLINEAR_CONG *
	* option is selected.  It assumes that the occupancies of all the       *
	* placement subregions have been properly updated, and simply           *
	* computes the cost due to these occupancies by summing over all        *
	* subregions.  This will be inefficient for moves that don't affect     *
	* many subregions (i.e. small moves late in placement), esp. when there *
	* are a lot of subregions.  May recode later to update only affected    *
	* subregions.                                                           */
	
	float cost, tmp;
	int i, j;
	
	cost = 0.;
	
	for (i=0;i<num_regions;i++) {
		for (j=0;j<num_regions;j++) {
			
			/* Many different cost metrics possible.  1st try:  */
			
			if (place_region_x[i][j].occupancy < place_region_x[i][j].capacity) {
				cost += place_region_x[i][j].occupancy * 
				place_region_x[i][j].inv_capacity;
			}
			else {  /* Overused region -- penalize. */
				
				tmp = place_region_x[i][j].occupancy * 
				place_region_x[i][j].inv_capacity;
				cost += tmp * tmp;
			}
			
			if (place_region_y[i][j].occupancy < place_region_y[i][j].capacity) {
				cost += place_region_y[i][j].occupancy * 
				place_region_y[i][j].inv_capacity;
			}
			else {  /* Overused region -- penalize. */
				
				tmp = place_region_y[i][j].occupancy * 
				place_region_y[i][j].inv_capacity;
				cost += tmp * tmp;
			}
			
		}
	}

return (cost);
}


static void update_region_occ (int inet, struct s_bb *coords, 
							   int add_or_sub, int num_regions) {
	
	/* Called only when the place_cost_type is NONLINEAR_CONG.  If add_or_sub *
	* is 1, this uses the new net bounding box to increase the occupancy     *
	* of some regions.  If add_or_sub = - 1, it decreases the occupancy      *
	* by that due to this bounding box.                                      */
	
	float net_xmin, net_xmax, net_ymin, net_ymax, crossing; 
	float inv_region_len, inv_region_height;
	float inv_bb_len, inv_bb_height;
	float overlap_xlow, overlap_xhigh, overlap_ylow, overlap_yhigh;
	float y_overlap, x_overlap, x_occupancy, y_occupancy;
	int imin, imax, jmin, jmax, i, j;
	
	if (net[inet].num_pins > 50) {
		crossing = 2.7933 + 0.02616 * (net[inet].num_pins - 50);
	}
	else {  
		crossing = cross_count[net[inet].num_pins-1];
	}
	
	net_xmin = coords->xmin - 0.5;
	net_xmax = coords->xmax + 0.5;
	net_ymin = coords->ymin - 0.5;
	net_ymax = coords->ymax + 0.5;
	
	/* I could precompute the two values below.  Should consider this. */
	
	inv_region_len = (float) num_regions / (float) nx;
	inv_region_height = (float) num_regions / (float) ny;
	
	/* Get integer coordinates defining the rectangular area in which the *
		* subregions have to be updated.  Formula is as follows:  subtract   *
		* 0.5 from net_xmin, etc. to get numbers from 0 to nx or ny;         *
		* divide by nx or ny to scale between 0 and 1; multiply by           *
		* num_regions to scale between 0 and num_regions; and truncate to    *
		* get the final answer.                                              */
	
	imin = (int) (net_xmin - 0.5) * inv_region_len;
	imax = (int) (net_xmax - 0.5) * inv_region_len;
	imax = min (imax, num_regions - 1);       /* Watch for weird roundoff */
	
	jmin = (int) (net_ymin - 0.5) * inv_region_height;
	jmax = (int) (net_ymax - 0.5) * inv_region_height;
	jmax = min (jmax, num_regions - 1);       /* Watch for weird roundoff */
	
	inv_bb_len = 1. / (net_xmax - net_xmin);
	inv_bb_height = 1. / (net_ymax - net_ymin);
	
	/* See RISA paper (ICCAD '94, pp. 690 - 695) for a description of why *
		* I use exactly this cost function.                                  */
	
	for (i=imin;i<=imax;i++) {
		for (j=jmin;j<=jmax;j++) {
			overlap_xlow = max (place_region_bounds_x[i],net_xmin);
			overlap_xhigh = min (place_region_bounds_x[i+1],net_xmax);
			overlap_ylow = max (place_region_bounds_y[j],net_ymin);
			overlap_yhigh = min (place_region_bounds_y[j+1],net_ymax);
			
			x_overlap = overlap_xhigh - overlap_xlow;
			y_overlap = overlap_yhigh - overlap_ylow;
			
#ifdef DEBUG
			
			if (x_overlap < -0.001) {
				printf ("Error in update_region_occ:  x_overlap < 0"
						"\n inet = %d, overlap = %g\n", inet, x_overlap);
			}
			
			if (y_overlap < -0.001) {
				printf ("Error in update_region_occ:  y_overlap < 0"
						"\n inet = %d, overlap = %g\n", inet, y_overlap);
			}
#endif
			
			
			x_occupancy = crossing * y_overlap * x_overlap * inv_bb_height *
				inv_region_len;
			y_occupancy = crossing * x_overlap * y_overlap * inv_bb_len *
				inv_region_height;
			
			place_region_x[i][j].occupancy += add_or_sub * x_occupancy;
			place_region_y[i][j].occupancy += add_or_sub * y_occupancy;
		}
	}
	
}


static void free_place_regions (int num_regions) {
	
	/* Frees the place_regions data structures needed by the NONLINEAR_CONG *
	* cost function.                                                       */
	
	free_matrix (place_region_x, 0, num_regions-1, 0, sizeof (struct 
															  s_place_region));
	
	free_matrix (place_region_y, 0, num_regions-1, 0, sizeof (struct 
															  s_place_region));
	
	free (place_region_bounds_x);
	free (place_region_bounds_y);
}


static void free_placement_structs (int place_cost_type, int num_regions,
									float **old_region_occ_x, float **old_region_occ_y, 
									struct s_placer_opts placer_opts) {
	
	/* Frees the major structures needed by the placer (and not needed       *
	* elsewhere).   */
	
	int inet;
	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {   
		for (inet=0; inet<num_nets; inet++) {
			/*add one to the address since it is indexed from 1 not 0 */
			
			point_to_point_delay_cost[inet] ++;
			free(point_to_point_delay_cost[inet]);
			
			point_to_point_timing_cost[inet] ++;
			free(point_to_point_timing_cost[inet]);
			
			temp_point_to_point_delay_cost[inet] ++;
			free(temp_point_to_point_delay_cost[inet]);
			
			temp_point_to_point_timing_cost[inet] ++;
			free(temp_point_to_point_timing_cost[inet]);
		}
		free(point_to_point_delay_cost);
		free(temp_point_to_point_delay_cost);
		
		free(point_to_point_timing_cost);
		free(temp_point_to_point_timing_cost);
		
		free_matrix (net_pin_index, 0, num_blocks-1, 0, sizeof(int));
	}
	
	
	free (net_cost);
	free (temp_net_cost);
	free (bb_num_on_edges);
	free (bb_coords);
	
	net_cost = NULL;         /* Defensive coding. */
	temp_net_cost = NULL;
	bb_num_on_edges = NULL;
	bb_coords = NULL;
	
	free_unique_pin_list ();
	
	if (place_cost_type == NONLINEAR_CONG) {
		free_place_regions (num_regions);
		free_matrix (old_region_occ_x, 0, num_regions-1,0, sizeof (float));
		free_matrix (old_region_occ_y, 0, num_regions-1,0, sizeof (float));
	}
	
	else if (place_cost_type == LINEAR_CONG) {
		free_fast_cost_update_structs ();
	}
}


static void alloc_and_load_placement_structs (int place_cost_type, 
											  int num_regions, float place_cost_exp, float ***old_region_occ_x, 
											  float ***old_region_occ_y, struct s_placer_opts placer_opts) {
	
	/* Allocates the major structures needed only by the placer, primarily for *
	* computing costs quickly and such.                                       */
	
	int inet, ipin, ppp;
	int i;
	
	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {
		/*allocate structures associated with timing driven placement */
		/* [0..num_nets-1][1..num_pins-1]  */
		point_to_point_delay_cost = (float **) my_malloc (num_nets * sizeof (float*));
		temp_point_to_point_delay_cost = (float **) my_malloc (num_nets * sizeof (float*));
		
		point_to_point_timing_cost = (float **) my_malloc (num_nets * sizeof (float*));
		temp_point_to_point_timing_cost = (float **) my_malloc (num_nets * sizeof (float*));
		
		ppp = MAX(pins_per_clb,pins_per_memory);
		ppp = MAX(ppp,pins_per_dsp);

		for(i =1; i<number_module; i++)
		{
			ppp = MAX(ppp,constraint[i].num_pins);
		}
		
		net_pin_index = (int **)alloc_matrix(0, num_blocks-1, 0, ppp, sizeof (int));
		
		for (inet=0; inet<num_nets; inet++) {
			
			/* in the following, subract one so index starts at *
			* 1 instead of 0 */
			point_to_point_delay_cost[inet] = 
			(float *) my_malloc ((net[inet].num_pins-1) * sizeof (float));
			point_to_point_delay_cost[inet] --;
			
			temp_point_to_point_delay_cost[inet] = 
				(float *) my_malloc ((net[inet].num_pins-1) * sizeof (float));
			temp_point_to_point_delay_cost[inet] --;
			
			point_to_point_timing_cost[inet] = 
				(float *) my_malloc ((net[inet].num_pins-1) * sizeof (float));
			point_to_point_timing_cost[inet] --;
			
			temp_point_to_point_timing_cost[inet] = 
				(float *) my_malloc ((net[inet].num_pins-1) * sizeof (float));
			temp_point_to_point_timing_cost[inet] --;
		}
		for (inet=0; inet<num_nets; inet++) {
			for (ipin=1; ipin<net[inet].num_pins; ipin++) {
				point_to_point_delay_cost[inet][ipin] = 0;
				temp_point_to_point_delay_cost[inet][ipin] = 0;
			}
		}
	}
	
	
	
	
	
	net_cost = (float *) my_malloc (num_nets * sizeof (float));
	temp_net_cost = (float *) my_malloc (num_nets * sizeof (float));
	
	/* Used to store costs for moves not yet made and to indicate when a net's   *
		* cost has been recomputed. temp_net_cost[inet] < 0 means net's cost hasn't *
		* been recomputed.                                                          */
	
	for (inet=0;inet<num_nets;inet++)
		temp_net_cost[inet] = -1.; 
	
	bb_coords = (struct s_bb *) my_malloc (num_nets * sizeof(struct s_bb));
	bb_num_on_edges = (struct s_bb *) my_malloc (num_nets * sizeof(struct s_bb));
	
	/* Get a list of pins with no duplicates. */
	
	alloc_and_load_unique_pin_list ();
	
	/* Allocate storage for subregion data, if needed. */
	
	if (place_cost_type == NONLINEAR_CONG) {
		alloc_place_regions (num_regions);
		load_place_regions(num_regions);
		*old_region_occ_x = (float **) alloc_matrix (0, num_regions-1,0,
													 num_regions-1, sizeof (float));
		*old_region_occ_y = (float **) alloc_matrix (0, num_regions-1,0,
													 num_regions-1, sizeof (float));
	}
	else {   /* Shouldn't use them; crash hard if I do!   */
		*old_region_occ_x = NULL;
		*old_region_occ_y = NULL;
	}

if (place_cost_type == LINEAR_CONG) {
    alloc_and_load_for_fast_cost_update (place_cost_exp);
}
}


static void alloc_place_regions (int num_regions) {
	
	/* Allocates memory for the regional occupancy, cost, etc. counts *
	* kept when we're using the NONLINEAR_CONG placement cost        *
	* function.                                                      */
	
	place_region_x = (struct s_place_region **) alloc_matrix (0, num_regions-1,
															  0, num_regions-1, sizeof (struct s_place_region));
	
	place_region_y = (struct s_place_region **) alloc_matrix (0, num_regions-1,
															  0, num_regions-1, sizeof (struct s_place_region));
	
	place_region_bounds_x = (float *) my_malloc ((num_regions+1) * 
												 sizeof (float));
	
	place_region_bounds_y = (float *) my_malloc ((num_regions+1) * 
												 sizeof (float));
}


static void load_place_regions (int num_regions) {
	
	/* Loads the capacity values in each direction for each of the placement *
	* regions.  The chip is divided into a num_regions x num_regions array. */
	
	int i, j, low_block, high_block, rnum;
	float low_lim, high_lim, capacity, fac, block_capacity;
	float len_fac, height_fac;
	
	/* First load up horizontal channel capacities.  */ 
	
	for (j=0;j<num_regions;j++) {
		capacity = 0.;
		low_lim = (float) j / (float) num_regions * ny + 1.;
		high_lim = (float) (j+1) / (float) num_regions * ny;
		low_block = floor (low_lim);
		low_block = max (1,low_block); /* Watch for weird roundoff effects. */
		high_block = ceil (high_lim);
		high_block = min(high_block, ny);
		
		block_capacity = (chan_width_x[low_block - 1] + 
						  chan_width_x[low_block])/2.;
		if (low_block == 1) 
			block_capacity += chan_width_x[0]/2.;
		
		fac = 1. - (low_lim - low_block);
		capacity += fac * block_capacity;
        
		for (rnum=low_block+1;rnum<high_block;rnum++) {
			block_capacity = (chan_width_x[rnum-1] + chan_width_x[rnum]) / 2.;
			capacity += block_capacity;
		}
		
		block_capacity = (chan_width_x[high_block-1] +
						  chan_width_x[high_block]) / 2.;
		if (high_block == ny) 
			block_capacity += chan_width_x[ny]/2.;
		
		fac = 1. - (high_block - high_lim);
		capacity += fac * block_capacity;
		
		for (i=0;i<num_regions;i++) {
			place_region_x[i][j].capacity = capacity;
			place_region_x[i][j].inv_capacity = 1. / capacity;
			place_region_x[i][j].occupancy = 0.;
			place_region_x[i][j].cost = 0.;
		}
	}
	
	/* Now load vertical channel capacities.  */
	
	for (i=0;i<num_regions;i++) {
		capacity = 0.;
		low_lim = (float) i / (float) num_regions * nx + 1.;
		high_lim = (float) (i+1) / (float) num_regions * nx;
		low_block = floor (low_lim);
		low_block = max (1,low_block); /* Watch for weird roundoff effects. */
		high_block = ceil (high_lim);
		high_block = min(high_block, nx);
		
		block_capacity = (chan_width_y[low_block - 1] +
						  chan_width_y[low_block])/2.;
		if (low_block == 1)
			block_capacity += chan_width_y[0]/2.;
		
		fac = 1. - (low_lim - low_block);
		capacity += fac * block_capacity;
		
		for (rnum=low_block+1;rnum<high_block;rnum++) {
			block_capacity = (chan_width_y[rnum-1] + chan_width_y[rnum]) / 2.;
			capacity += block_capacity;
		}
		
		block_capacity = (chan_width_y[high_block-1] +
						  chan_width_y[high_block]) / 2.;
		if (high_block == nx)
			block_capacity += chan_width_y[nx]/2.;
		
		fac = 1. - (high_block - high_lim);
		capacity += fac * block_capacity;
		
		for (j=0;j<num_regions;j++) {
			place_region_y[i][j].capacity = capacity; 
			place_region_y[i][j].inv_capacity = 1. / capacity;
			place_region_y[i][j].occupancy = 0.;
			place_region_y[i][j].cost = 0.; 
		}
	}
	
	/* Finally set up the arrays indicating the limits of each of the *
		* placement subregions.                                          */
	
	len_fac = (float) nx / (float) num_regions;
	height_fac = (float) ny / (float) num_regions;
	
	place_region_bounds_x[0] = 0.5;
	place_region_bounds_y[0] = 0.5;
	
	for (i=1;i<=num_regions;i++) {
		place_region_bounds_x[i] = place_region_bounds_x[i-1] + len_fac;
		place_region_bounds_y[i] = place_region_bounds_y[i-1] + height_fac;
	}
}


static void free_unique_pin_list (void) {
	
	/* Frees the unique pin list structures.                               */
	
	int any_dup, inet;
	
	any_dup = 0;
	
	for (inet=0;inet<num_nets;inet++) {
		if (duplicate_pins[inet] != 0) {
			free (unique_pin_list[inet]);
			any_dup = 1;
		}
	}
	
	if (any_dup != 0) 
		free (unique_pin_list);
	
	free (duplicate_pins);
}


static void alloc_and_load_unique_pin_list (void) {
	
	/* This routine looks for multiple pins going to the same block in the *
	* pinlist of each net.  If it finds any, it marks that net as having  *
	* duplicate pins, and creates a new pinlist with no duplicates.  This *
	* is then used by the updatable bounding box calculation routine for  *
	* efficiency.                                                         */
	
	int inet, ipin, bnum, num_dup, any_dups, offset;
	int *times_listed;  /* [0..num_blocks-1]: number of times a block is   *
		* listed in the pinlist of a net.  Temp. storage. */
	
	duplicate_pins = my_calloc (num_nets, sizeof(int));
	times_listed = my_calloc (num_blocks, sizeof(int)); 
	any_dups = 0;
	
	for (inet=0;inet<num_nets;inet++) {
		
		num_dup = 0;
		
		for (ipin=0;ipin<net[inet].num_pins;ipin++) {
			bnum = net[inet].blocks[ipin];
			times_listed[bnum]++;
			if (times_listed[bnum] > 1) 
				num_dup++;
		}
		
		if (num_dup > 0) {   /* Duplicates found.  Make unique pin list. */
			duplicate_pins[inet] = num_dup;
			
			//printf("duplicate_pin[%d] %d \n", inet ,duplicate_pins[inet]);
			if (any_dups == 0) {        /* This is the first duplicate found */
				unique_pin_list = (int **) my_calloc (num_nets, sizeof(int *));
				any_dups = 1;
			}
			
			unique_pin_list[inet] = my_malloc ((net[inet].num_pins - num_dup) *
											   sizeof(int));
			
			offset = 0;
			for (ipin=0;ipin<net[inet].num_pins;ipin++) { 
				bnum = net[inet].blocks[ipin];
				if (times_listed[bnum] != 0) {
					times_listed[bnum] = 0;
					unique_pin_list[inet][offset] = bnum;
					offset++;
				}
			}
		}

else {          /* No duplicates found.  Reset times_listed. */
for (ipin=0;ipin<net[inet].num_pins;ipin++) {
	bnum = net[inet].blocks[ipin];
	times_listed[bnum] = 0;
}
}
	}

free ((void *) times_listed);
}


static void get_bb_from_scratch (int inet, struct s_bb *coords, 
								 struct s_bb *num_on_edges) {
	
	/* This routine finds the bounding box of each net from scratch (i.e.    *
	* from only the block location information).  It updates both the       *
* coordinate and number of blocks on each edge information.  It         *
* should only be called when the bounding box information is not valid. */
	
	int ipin, bnum, x, y, xmin, xmax, ymin, ymax;
	int xmin_edge, xmax_edge, ymin_edge, ymax_edge;
	int n_pins;
	int *plist;
	
	/* I need a list of blocks to which this net connects, with no block listed *
		* more than once, in order to get a proper count of the number on the edge *
		* of the bounding box.                                                     */
	
	if (duplicate_pins[inet] == 0) {
		plist = net[inet].blocks;
		n_pins = net[inet].num_pins;
	}
	else {
		plist = unique_pin_list[inet];
		n_pins = net[inet].num_pins - duplicate_pins[inet];
	}
	
	x = block[plist[0]].x;
	y = block[plist[0]].y;
	
	x = max(min(x,nx),1);   
	y = max(min(y,ny),1);
	
	xmin = x;
	ymin = y;
	xmax = x;
	ymax = y;
	xmin_edge = 1;
	ymin_edge = 1;
	xmax_edge = 1;
	ymax_edge = 1;
	
	for (ipin=1;ipin<n_pins;ipin++) {
		
		bnum = plist[ipin];
		x = block[bnum].x;
		y = block[bnum].y;
		
		/* Code below counts IO blocks as being within the 1..nx, 1..ny clb array. *
			* This is because channels do not go out of the 0..nx, 0..ny range, and   *
			* I always take all channels impinging on the bounding box to be within   *
			* that bounding box.  Hence, this "movement" of IO blocks does not affect *
			* the which channels are included within the bounding box, and it         *
			* simplifies the code a lot.                                              */
		
		x = max(min(x,nx),1);   
		y = max(min(y,ny),1);
		
		if (x == xmin) {  
			xmin_edge++;
		}
		if (x == xmax) {  /* Recall that xmin could equal xmax -- don't use else */
			xmax_edge++;
		}
		else if (x < xmin) {
			xmin = x;
			xmin_edge = 1;
		}
		else if (x > xmax) {
			xmax = x;
			xmax_edge = 1;
		}
		
		if (y == ymin) {
			ymin_edge++;
		}
		if (y == ymax) {
			ymax_edge++;
		}
		else if (y < ymin) {
			ymin = y;
			ymin_edge = 1;
		}
		else if (y > ymax) {
			ymax = y;
			ymax_edge = 1;
		}
	}

/* Copy the coordinates and number on edges information into the proper   *
* structures.                                                            */

coords->xmin = xmin;
coords->xmax = xmax;
coords->ymin = ymin;
coords->ymax = ymax;

num_on_edges->xmin = xmin_edge;
num_on_edges->xmax = xmax_edge;
num_on_edges->ymin = ymin_edge;
num_on_edges->ymax = ymax_edge;
}


static float get_net_cost (int inet, struct s_bb *bbptr) {
	
	/* Finds the cost due to one net by looking at its coordinate bounding  *
	* box.                                                                 */
	
	float ncost, crossing;
	
	/* Get the expected "crossing count" of a net, based on its number *
	* of pins.  Extrapolate for very large nets.                      */
	
	if (net[inet].num_pins > 50) {
		crossing = 2.7933 + 0.02616 * (net[inet].num_pins - 50); 
		/*    crossing = 3.0;    Old value  */
	}
	else {
		crossing = cross_count[net[inet].num_pins-1]; 
	}
	
	/* Could insert a check for xmin == xmax.  In that case, assume  *
		* connection will be made with no bends and hence no x-cost.    *
		* Same thing for y-cost.                                        */
	
	/* Cost = wire length along channel * cross_count / average      *
		* channel capacity.   Do this for x, then y direction and add.  */
	
	ncost = (bbptr->xmax - bbptr->xmin + 1) * crossing *
		chanx_place_cost_fac[bbptr->ymax][bbptr->ymin-1]; 
	
	ncost += (bbptr->ymax - bbptr->ymin + 1) * crossing *
		chany_place_cost_fac[bbptr->xmax][bbptr->xmin-1];
	
	return(ncost);
}


static void get_non_updateable_bb (int inet, struct s_bb *bb_coord_new) {
	
	/* Finds the bounding box of a net and stores its coordinates in the  *
	* bb_coord_new data structure.  This routine should only be called   *
	* for small nets, since it does not determine enough information for *
	* the bounding box to be updated incrementally later.                *
	* Currently assumes channels on both sides of the CLBs forming the   *
	* edges of the bounding box can be used.  Essentially, I am assuming *
	* the pins always lie on the outside of the bounding box.            */
	
	
	int k, xmax, ymax, xmin, ymin, x, y;
	
	x = block[net[inet].blocks[0]].x;
	y = block[net[inet].blocks[0]].y;
	
	xmin = x;
	ymin = y;
	xmax = x;
	ymax = y;
	
	for (k=1;k<net[inet].num_pins;k++) {
		x = block[net[inet].blocks[k]].x;
		y = block[net[inet].blocks[k]].y;
		
		if (x < xmin) {
			xmin = x;
		}
		else if (x > xmax) {
			xmax = x;
		}
		
		if (y < ymin) {
			ymin = y;
		}
		else if (y > ymax ) {
			ymax = y;
		}
	}
	
	/* Now I've found the coordinates of the bounding box.  There are no *
		* channels beyond nx and ny, so I want to clip to that.  As well,   *
		* since I'll always include the channel immediately below and the   *
		* channel immediately to the left of the bounding box, I want to    *
		* clip to 1 in both directions as well (since minimum channel index *
												* is 0).  See route.c for a channel diagram.                        */
	
	bb_coord_new->xmin = max(min(xmin,nx),1);
	bb_coord_new->ymin = max(min(ymin,ny),1);
	bb_coord_new->xmax = max(min(xmax,nx),1);
	bb_coord_new->ymax = max(min(ymax,ny),1);
}


static void update_bb (int inet, struct s_bb *bb_coord_new, struct s_bb 
					   *bb_edge_new, int xold, int yold, int xnew, int ynew) 
{
	
	/* Updates the bounding box of a net by storing its coordinates in    *
	* the bb_coord_new data structure and the number of blocks on each   *
	* edge in the bb_edge_new data structure.  This routine should only  *
	* be called for large nets, since it has some overhead relative to   *
	* just doing a brute force bounding box calculation.  The bounding   *
	* box coordinate and edge information for inet must be valid before  *
	* this routine is called.                                            *
	* Currently assumes channels on both sides of the CLBs forming the   *
	* edges of the bounding box can be used.  Essentially, I am assuming *
	* the pins always lie on the outside of the bounding box.            */
	
	/* IO blocks are considered to be one cell in for simplicity. */
	
	xnew = max(min(xnew,nx),1);
	ynew = max(min(ynew,ny),1);
	xold = max(min(xold,nx),1);
	yold = max(min(yold,ny),1);
	
	/* Check if I can update the bounding box incrementally. */ 
	
	if (xnew < xold) 
	{                          /* Move to left. */
		
		/* Update the xmax fields for coordinates and number of edges first. */
		
		if (xold == bb_coords[inet].xmax) 
		{       /* Old position at xmax. */
			if (bb_num_on_edges[inet].xmax == 1) {
				get_bb_from_scratch (inet, bb_coord_new, bb_edge_new);
				return;
			}
			else {
				bb_edge_new->xmax = bb_num_on_edges[inet].xmax - 1;
				bb_coord_new->xmax = bb_coords[inet].xmax; 
			}
		}

		else 
		{              /* Move to left, old postion was not at xmax. */
			bb_coord_new->xmax = bb_coords[inet].xmax; 
			bb_edge_new->xmax = bb_num_on_edges[inet].xmax;
		}

	/* Now do the xmin fields for coordinates and number of edges. */

		if (xnew < bb_coords[inet].xmin) 
		{    /* Moved past xmin */
			bb_coord_new->xmin = xnew;
			bb_edge_new->xmin = 1;
		}
		else if (xnew == bb_coords[inet].xmin) 
		{   /* Moved to xmin */
			bb_coord_new->xmin = xnew;
			bb_edge_new->xmin = bb_num_on_edges[inet].xmin + 1;
		}
		else 
		{                                  /* Xmin unchanged. */
			bb_coord_new->xmin = bb_coords[inet].xmin;
			bb_edge_new->xmin = bb_num_on_edges[inet].xmin;
		}
	}    /* End of move to left case. */


	else if (xnew > xold) 
	{             /* Move to right. */

		/* Update the xmin fields for coordinates and number of edges first. */

		if (xold == bb_coords[inet].xmin) 
		{   /* Old position at xmin. */
			if (bb_num_on_edges[inet].xmin == 1) 
			{
				get_bb_from_scratch (inet, bb_coord_new, bb_edge_new);
				return;
			}
			else 
			{
				bb_edge_new->xmin = bb_num_on_edges[inet].xmin - 1;
				bb_coord_new->xmin = bb_coords[inet].xmin;
			}
		}
		else 
		{                /* Move to right, old position was not at xmin. */
			bb_coord_new->xmin = bb_coords[inet].xmin;
			bb_edge_new->xmin = bb_num_on_edges[inet].xmin;
		}

		/* Now do the xmax fields for coordinates and number of edges. */

		if (xnew > bb_coords[inet].xmax) 
		{    /* Moved past xmax. */
			bb_coord_new->xmax = xnew;
			bb_edge_new->xmax = 1;   
		} 
		else if (xnew == bb_coords[inet].xmax) 
		{   /* Moved to xmax */
			bb_coord_new->xmax = xnew;
			bb_edge_new->xmax = bb_num_on_edges[inet].xmax + 1;
		} 
		else 
		{                                  /* Xmax unchanged. */ 
			bb_coord_new->xmax = bb_coords[inet].xmax; 
			bb_edge_new->xmax = bb_num_on_edges[inet].xmax;   
		}
	}    /* End of move to right case. */

	else 
	{          /* xnew == xold -- no x motion. */
		bb_coord_new->xmin = bb_coords[inet].xmin;
		bb_coord_new->xmax = bb_coords[inet].xmax;
		bb_edge_new->xmin = bb_num_on_edges[inet].xmin;
		bb_edge_new->xmax = bb_num_on_edges[inet].xmax;
	}

	/* Now account for the y-direction motion. */

	if (ynew < yold) 
	{                  /* Move down. */

	/* Update the ymax fields for coordinates and number of edges first. */

		if (yold == bb_coords[inet].ymax) 
		{       /* Old position at ymax. */
			if (bb_num_on_edges[inet].ymax == 1) 
			{
				get_bb_from_scratch (inet, bb_coord_new, bb_edge_new);
				return;
			}
			else 
			{
				bb_edge_new->ymax = bb_num_on_edges[inet].ymax - 1;
				bb_coord_new->ymax = bb_coords[inet].ymax;
			}
		}     
		else 
		{              /* Move down, old postion was not at ymax. */
			bb_coord_new->ymax = bb_coords[inet].ymax;
			bb_edge_new->ymax = bb_num_on_edges[inet].ymax;
		}     

		/* Now do the ymin fields for coordinates and number of edges. */

		if (ynew < bb_coords[inet].ymin) 
		{    /* Moved past ymin */
			bb_coord_new->ymin = ynew;
			bb_edge_new->ymin = 1;
		}     
		else if (ynew == bb_coords[inet].ymin) 
		{   /* Moved to ymin */
			bb_coord_new->ymin = ynew;
			bb_edge_new->ymin = bb_num_on_edges[inet].ymin + 1;
		}     
		else 
		{                                  /* ymin unchanged. */
			bb_coord_new->ymin = bb_coords[inet].ymin;
			bb_edge_new->ymin = bb_num_on_edges[inet].ymin;
		}     
	}    /* End of move down case. */
	else if (ynew > yold) 
	{             /* Moved up. */	
		/* Update the ymin fields for coordinates and number of edges first. */

		if (yold == bb_coords[inet].ymin) 
		{   /* Old position at ymin. */
			if (bb_num_on_edges[inet].ymin == 1) 
			{
				get_bb_from_scratch (inet, bb_coord_new, bb_edge_new);
				return;
			}
			else 
			{
				bb_edge_new->ymin = bb_num_on_edges[inet].ymin - 1;
				bb_coord_new->ymin = bb_coords[inet].ymin;
			}
		}     
		else 
		{                /* Moved up, old position was not at ymin. */
			bb_coord_new->ymin = bb_coords[inet].ymin;
			bb_edge_new->ymin = bb_num_on_edges[inet].ymin;
		}     
		/* Now do the ymax fields for coordinates and number of edges. */
		if (ynew > bb_coords[inet].ymax) 
		{    /* Moved past ymax. */
			bb_coord_new->ymax = ynew;
			bb_edge_new->ymax = 1;
		}     
		else if (ynew == bb_coords[inet].ymax) 
		{   /* Moved to ymax */
			bb_coord_new->ymax = ynew;
			bb_edge_new->ymax = bb_num_on_edges[inet].ymax + 1;
		}     
		else 
		{                                  /* ymax unchanged. */
			bb_coord_new->ymax = bb_coords[inet].ymax;
			bb_edge_new->ymax = bb_num_on_edges[inet].ymax;
		}     
	}    /* End of move up case. */
	else 
	{          /* ynew == yold -- no y motion. */
		bb_coord_new->ymin = bb_coords[inet].ymin;
		bb_coord_new->ymax = bb_coords[inet].ymax;
		bb_edge_new->ymin = bb_num_on_edges[inet].ymin;
		bb_edge_new->ymax = bb_num_on_edges[inet].ymax;
	}
}


static void initial_placement (enum e_pad_loc_type pad_loc_type,
							   char *pad_loc_file) {  
	
	/* Randomly places the blocks to create an initial placement.     */
	
	struct s_pos {int x; int y;} *pos; 
	int i, j, k, count, iblk, choice, tsize, isubblk;
	int *carry_head_list;
	int *carry_length_list;
	int num_carry_chains;
	int lastx, lasty;
	int *located;

	int check_valid;


	
	tsize = max(nx*ny, 2*(nx+ny));
	pos = (struct s_pos *) my_malloc(tsize*sizeof(struct s_pos));
	
	/* Initialize all occupancy to zero. */
	
	
	if (max_carry_length() > ny) {
		printf("Maximum carry chain length is %d.  Please re-run\n",max_carry_length());
		printf("on an architecture with ny >= this number.\n");
		exit(1);
	}
	
	for (i=0;i<=nx+1;i++) {
		for (j=0;j<=ny+1;j++) {			
				clb[i][j].occ = 0;
				clb[i][j].empty = 1;
				clb[i][j].module_index = 0;							
		}
	}


	located = (int *)my_malloc(sizeof(int)*num_blocks);
	for(i=0;i<num_blocks;i++) located[i] = 0;

	/*First place all the veb  */
	initial_constraint_placement(located);
	/* Second place CLB's that are part of a carry chain */	
	carry_head_list = (int *)my_malloc(sizeof(int)*num_blocks);
	carry_length_list = (int *)my_malloc(sizeof(int)*num_blocks);
	
	
	get_list_of_carry_heads(&num_carry_chains,&carry_head_list, &carry_length_list);
	lastx = 1;
	lasty = 1;
	for(i=0;i<num_carry_chains;i++) {
		j = carry_head_list[i];		
		if (located[j] == 0)
		{
			//the carry chain is not located before in initial_constraint_placement
			if (block[j].fixed != 1)
			{
				//the block is not constrained and doesn't have fix position
				if (lasty + carry_length_list[i] > ny+1) {
					lastx = lastx + 1;
					lasty = 1;
					if (lastx >= nx) {
						printf("Error placing carry chains.  Re-run with a larger array\n");
						exit(1);
					}
				}
				
				//search a position to place a carry chain	
				do {
					check_valid = 1;
					for (k = 0; k < carry_length_list[i]; k ++)
					{
						if (clb[lastx][lasty + k].type != CLB)
						{
							check_valid = 0;
							break;
						}
					}

					//search y until CLB					
					if (check_valid == 0)
					{						
						//	printf("finding y , lastx %d, lasty %d \n", lastx, lasty);	
						do
						{														
							lasty++;	
							//check the carry chain exit the boundary of FPGA or not
							if (lasty + carry_length_list[i] > ny+1) 
							{
								//change to next column if outside the boundary
								lastx = lastx + 1;
								lasty = 1;
								if (lastx >= nx) 
								{
									printf("Error placing carry chains.  Re-run with a larger array\n");
									exit(1);
								}
							}
						}while (clb[lastx][lasty].type != CLB);
						
						
					}														
						
				}while(check_valid!= 1);
																	
				if (lastx >= nx) {
					printf("Error placing carry chains.  Re-run with a larger array\n");
					exit(1);
				}
				
				while(j!=-1) 
				{
					//printf("Starting with block %d (%s) in location %d %d, module_index %d, block.carry_body = %d \n",j,block[j].name,lastx,lasty, block[j].module_index, block[j].carry_body);
					clb[lastx][lasty].occ = 1;
					clb[lastx][lasty].empty = 0;
					clb[lastx][lasty].u.block = j;
					clb[lastx][lasty].module_index = block[j].module_index;
					located[j] = 1;
					lasty ++;
					j = block[j].carryout_cluster;
				}
			}
		}
	}
	
	free(carry_length_list);
	free(carry_head_list);
	
	
	/* List of locations for CLB's */
	
	count = 0;
	for (i=1;i<=nx;i++) {
		for (j=1;j<=ny;j++) {
			if (clb[i][j].type == CLB) {
				if (clb[i][j].occ == 0) 
				{
					if (clb[i][j].module_index ==0)
					{
						pos[count].x = i;
						pos[count].y = j;
						count++;
					}
				}
			}
		}
	}

	for (iblk=0;iblk<num_blocks;iblk++) {
		if (block[iblk].type == CLB) {     /* only place CLBs in center */
			if (located[iblk] == 0 ) {   /* only if this CLB was not placed as part of a carry chain */
				choice = my_irand(count - 1); 
				clb[pos[choice].x][pos[choice].y].u.block = iblk;
				clb[pos[choice].x][pos[choice].y].occ = 1;
				clb[pos[choice].x][pos[choice].y].empty = 0;
				
				clb[pos[choice].x][pos[choice].y].module_index = block[iblk].module_index;
	
				
				//printf("Starting with block %d (%s) in location %d %d\n",iblk,block[iblk].name,pos[choice].x,pos[choice].y);
				/* Ensure randomizer doesn't pick this block again */
				pos[choice] = pos[count-1];   /* overwrite used block position */
				count--;
			}
		}
	}

free(located);

/* List of locations for DSP's */

printf("Placing DSP \n");
count = 0;
for (i=1;i<=nx;i++) {
    for (j=1;j<=ny;j++) {
        if (clb[i][j].type == DSP) {
			//printf("clb[%d][%d] is DSP \n", i, j);
			pos[count].x = i;
			pos[count].y = j;
			count++;
		}
	}
}

//printf("count %d \n", count);
for (iblk=0;iblk<num_blocks;iblk++) {
    if (block[iblk].type == DSP) {
		choice = my_irand(count - 1); 
		clb[pos[choice].x][pos[choice].y].u.block = iblk;
		clb[pos[choice].x][pos[choice].y].occ = 1;
		
		/* Ensure randomizer doesn't pick this block again */
		pos[choice] = pos[count-1];   /* overwrite used block position */
		count--;
    }
}

/* List of locations for memories */

count = 0;
for (i=1;i<=nx;i++) {
    for (j=1;j<=ny;j++) {
        if (clb[i][j].type == MEMORY) {
			pos[count].x = i;
			pos[count].y = j;
			count++;
		}
	}
}

for (iblk=0;iblk<num_blocks;iblk++) {
    if (block[iblk].type == MEMORY) {
		choice = my_irand(count - 1); 
		clb[pos[choice].x][pos[choice].y].u.block = iblk;
		clb[pos[choice].x][pos[choice].y].occ = 1;
		
		/* Ensure randomizer doesn't pick this block again */
		pos[choice] = pos[count-1];   /* overwrite used block position */
		count--;
    }
}

/* Now do the io blocks around the periphery */

if (pad_loc_type == USER) {
    read_user_pad_loc (pad_loc_file);
} 
else {           /* place_randomly. */
count = 0;
for (i=1;i<=nx;i++) {
	pos[count].x = i;
	pos[count].y = 0;
	pos[count+1].x = i;
	pos[count+1].y = ny + 1;
	count += 2;
}

for (j=1;j<=ny;j++) {
	pos[count].x = 0;
	pos[count].y = j;
	pos[count+1].x = nx + 1;
	pos[count+1].y = j;
	count += 2;
}

for (iblk=0;iblk<num_blocks;iblk++) {
	if (block[iblk].type == INPAD || block[iblk].type == OUTPAD) {
		choice = my_irand (count - 1); 
		isubblk = clb[pos[choice].x][pos[choice].y].occ;
		clb[pos[choice].x][pos[choice].y].u.io_blocks[isubblk] = iblk;
		clb[pos[choice].x][pos[choice].y].occ++;
		if (clb[pos[choice].x][pos[choice].y].occ == io_rat) {
			/* Ensure randomizer doesn't pick this block again */
			pos[choice] = pos[count-1];   /* overwrite used block position */
			count--;
		}
		
	} 
}
}    /* End randomly place IO blocks branch of if */

/* All the blocks are placed now.  Make the block array agree with the    *
* clb array.                                                             */

for (i=0;i<=nx+1;i++) {
    for (j=0;j<=ny+1;j++) {
		if ((clb[i][j].type == CLB) ||
			(clb[i][j].type == DSP) ||
			(clb[i][j].type == EMB) ||
			(clb[i][j].type == MEMORY)) {
			if (clb[i][j].occ == 1) {
				block[clb[i][j].u.block].x = i;
				block[clb[i][j].u.block].y = j;
			}
		}
		else {
			if (clb[i][j].type == IO) {
				for (k=0;k<clb[i][j].occ;k++) {
					block[clb[i][j].u.io_blocks[k]].x = i;
					block[clb[i][j].u.io_blocks[k]].y = j;

					//printf("blk %d, x %d, y %d \n", clb[i][j].u.io_blocks[k], block[clb[i][j].u.io_blocks[k]].x , block[clb[i][j].u.io_blocks[k]].y);
				}
			}
		}
    }
}

printf("At end of initial_placement.\n");

#ifdef VERBOSE 
printf("At end of initial_placement.\n");
dump_clbs();
#endif

free (pos);
}


static void free_fast_cost_update_structs (void) {
	
	/* Frees the structures used to speed up evaluation of the nonlinear   *
	* congestion cost function.                                           */
	
	int i;
	
	for (i=0;i<=ny;i++) 
		free (chanx_place_cost_fac[i]);
	
	free (chanx_place_cost_fac);
	
	for (i=0;i<=nx;i++) 
		free (chany_place_cost_fac[i]);
	
	free (chany_place_cost_fac);
}


static void alloc_and_load_for_fast_cost_update (float place_cost_exp) {
	
	/* Allocates and loads the chanx_place_cost_fac and chany_place_cost_fac *
	* arrays with the inverse of the average number of tracks per channel   *
	* between [subhigh] and [sublow].  This is only useful for the cost     *
	* function that takes the length of the net bounding box in each        *
	* dimension divided by the average number of tracks in that direction.  *
	* For other cost functions, you don't have to bother calling this       *
	* routine; when using the cost function described above, however, you   *
	* must always call this routine after you call init_chan and before     *
	* you do any placement cost determination.  The place_cost_exp factor   *
	* specifies to what power the width of the channel should be taken --   *
	* larger numbers make narrower channels more expensive.                 */
	
	int low, high, i;
	
	/* Access arrays below as chan?_place_cost_fac[subhigh][sublow].  Since   *
	* subhigh must be greater than or equal to sublow, we only need to       *
	* allocate storage for the lower half of a matrix.                       */
	
	chanx_place_cost_fac = (float **) my_malloc ((ny + 1) * sizeof (float *));
	for (i=0;i<=ny;i++)
		chanx_place_cost_fac[i] = (float *) my_malloc ((i + 1) * sizeof (float));
	
	chany_place_cost_fac = (float **) my_malloc ((nx + 1) * sizeof (float *));
	for (i=0;i<=nx;i++)
		chany_place_cost_fac[i] = (float *) my_malloc ((i + 1) * sizeof (float));
	
	
	/* First compute the number of tracks between channel high and channel *
		* low, inclusive, in an efficient manner.                             */
	
	chanx_place_cost_fac[0][0] = chan_width_x[0];
	
	for (high=1;high<=ny;high++) {
		chanx_place_cost_fac[high][high] = chan_width_x[high];    
		for (low=0;low<high;low++) {
			chanx_place_cost_fac[high][low] = chanx_place_cost_fac[high-1][low]
			+ chan_width_x[high];
		}
	}
	
	/* Now compute the inverse of the average number of tracks per channel *
		* between high and low.  The cost function divides by the average     *
		* number of tracks per channel, so by storing the inverse I convert   *
		* this to a faster multiplication.  Take this final number to the     *
		* place_cost_exp power -- numbers other than one mean this is no      *
		* longer a simple "average number of tracks"; it is some power of     *
		* that, allowing greater penalization of narrow channels.             */
	
	for (high=0;high<=ny;high++) 
		for (low=0;low<=high;low++) {
			chanx_place_cost_fac[high][low] = (high - low + 1.) / 
			chanx_place_cost_fac[high][low];
			chanx_place_cost_fac[high][low] = 
				pow ((double) chanx_place_cost_fac[high][low],
					 (double) place_cost_exp);
		}
			
			
			/* Now do the same thing for the y-directed channels.  First get the  *
			* number of tracks between channel high and channel low, inclusive.  */
			
			chany_place_cost_fac[0][0] = chan_width_y[0];
	
	for (high=1;high<=nx;high++) {
		chany_place_cost_fac[high][high] = chan_width_y[high];
		for (low=0;low<high;low++) {
			chany_place_cost_fac[high][low] = chany_place_cost_fac[high-1][low]
			+ chan_width_y[high];
		}
	}
	
	/* Now compute the inverse of the average number of tracks per channel * 
		* between high and low.  Take to specified power.                     */
	
	for (high=0;high<=nx;high++) 
		for (low=0;low<=high;low++) {
			chany_place_cost_fac[high][low] = (high - low + 1.) / 
			chany_place_cost_fac[high][low]; 
			chany_place_cost_fac[high][low] = 
				pow ((double) chany_place_cost_fac[high][low],
					 (double) place_cost_exp);
		}
}


static void check_place (float bb_cost, float timing_cost, int place_cost_type, 
						 int num_regions, enum e_place_algorithm place_algorithm,
						 float delay_cost) {
	
	/* Checks that the placement has not confused our data structures. *
	* i.e. the clb and block structures agree about the locations of  *
	* every block, blocks are in legal spots, etc.  Also recomputes   *
	* the final placement cost from scratch and makes sure it is      *
	* within roundoff of what we think the cost is.                   */
	
	static int *bdone; 
	int i, j, k, error=0, bnum;
	float bb_cost_check;
	float timing_cost_check, delay_cost_check;
	
	bb_cost_check = comp_bb_cost(CHECK, place_cost_type, num_regions);
	printf("bb_cost recomputed from scratch is %g.\n", bb_cost_check);
	printf("ERROR_TOL %g \n ", ERROR_TOL);
	if (fabs(bb_cost_check - bb_cost) > bb_cost * ERROR_TOL) {
		printf("Error:  bb_cost_check: %g and bb_cost: %g differ in check_place.\n",
			   bb_cost_check, bb_cost);
		printf("bb_cost normal: %g \n", comp_bb_cost(NORMAL, place_cost_type, num_regions));
		error++;
	}
	
	if (place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		place_algorithm == PATH_TIMING_DRIVEN_PLACE) {
		comp_td_costs(&timing_cost_check, &delay_cost_check);
		printf("timing_cost recomputed from scratch is %g. \n", timing_cost_check);
		if (fabs(timing_cost_check - timing_cost) > timing_cost * ERROR_TOL) {
			printf("Error:  timing_cost_check: %g and timing_cost: "
				   "%g differ in check_place.\n",
				   timing_cost_check,timing_cost);
			error++;
		}
		printf("delay_cost recomputed from scratch is %g. \n", delay_cost_check);
		if (fabs(delay_cost_check - delay_cost) > delay_cost * ERROR_TOL) {
			printf("Error:  delay_cost_check: %g and delay_cost: "
				   "%g differ in check_place.\n",
				   delay_cost_check,delay_cost);
			error++;
		}
	}
	
	bdone = (int *) my_malloc (num_blocks*sizeof(int));
	for (i=0;i<num_blocks;i++) bdone[i] = 0;
/*
	for (i=0;i<=nx+1;i++) 
		for (j=0;j<=ny+1;j++) 
		{
			printf("clb[%d][%d].u.block  = %d \n", i, j , clb[i][j].u.block);
		}
*/						 
	
	/* Step through clb array. Check it against block array. */
	for (i=0;i<=nx+1;i++) 
		for (j=0;j<=ny+1;j++) {
			if (clb[i][j].occ == 0) continue;
			if ((clb[i][j].type == CLB) ||
				(clb[i][j].type == DSP) ||
				(clb[i][j].type == EMB) ||
				(clb[i][j].type == MEMORY)) {
				bnum = clb[i][j].u.block;
				if (block[bnum].type != clb[i][j].type) {
					printf("Error:  block %d type does not match (%d,%d) type.\n",
						   bnum,i,j);
					printf("block type %d, clb type %d \n", block[bnum].type, clb[i][j].type);
					error++;
				}
				if ((block[bnum].x != i) || (block[bnum].y != j)) {
					printf("Error:  block %d location conflicts with (%d,%d)"
						   "data.\n", bnum, i, j);
					error++;
				}
				if (clb[i][j].occ > 1) {
					printf("Error: clb(%d,%d) has occupancy of %d\n",
						   i,j,clb[i][j].occ);
					error++;
				}
				bdone[bnum]++;
			}
			else {  /* IO block */
				if (clb[i][j].occ > io_rat) {
					printf("Error:  clb(%d,%d) has occupancy of %d\n",i,j,
						   clb[i][j].occ);
					error++;
				}
				for (k=0;k<clb[i][j].occ;k++) {
					bnum = clb[i][j].u.io_blocks[k];
					if ((block[bnum].type != INPAD) && block[bnum].type != OUTPAD) {
						printf("Error:  block %d type does not match clb(%d,%d) type.\n",
							   bnum,i,j);
						error++;
					}
					if ((block[bnum].x != i) || (block[bnum].y != j)) {
						printf("Error:  block %d location conflicts with clb(%d,%d)"
							   "data.\n", bnum, i, j);
						error++; 
					}
					bdone[bnum]++;
				} 
			} 
			
		}

/* Check that every block exists in the clb and block arrays somewhere. */
for (i=0;i<num_blocks;i++) 
if (bdone[i] != 1) {
	printf("Error:  block %d listed %d times in data structures.\n",
		   i,bdone[i]);
	error++;
}
free (bdone);

if (error == 0) {
    printf("\nCompleted placement consistency check successfully.\n\n");
}
else {
    printf("\nCompleted placement consistency check, %d Errors found.\n\n",
		   error);
    printf("Aborting program.\n");
    exit(1);
}
}


void read_place (char *place_file, char *net_file, char *arch_file,
                 struct s_placer_opts placer_opts, struct s_router_opts router_opts,
				 t_chan_width_dist chan_width_dist, 
				 struct s_det_routing_arch det_routing_arch, 
				 t_segment_inf *segment_inf, t_timing_inf timing_inf,
				 t_subblock_data *subblock_data_ptr) {
	
	/* Reads in a previously computed placement of the circuit.  It      *
	* checks that the placement corresponds to the current architecture *
	* and netlist file.                                                 */
	
	
	char msg[BUFSIZE];
	int chan_width_factor, num_connections, inet, ipin;
	float bb_cost, delay_cost, timing_cost, est_crit;
	float **dummy_x, **dummy_y;
	float **net_slack, **net_delay;
	float **remember_net_delay_original_ptr; /*used to free net_delay if it is re-assigned*/
	
	remember_net_delay_original_ptr = NULL; /*prevents compiler warning*/
	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {
		/*this must be called before alloc_and_load_placement_structs *
		*and parse_placement_file since it modifies the structures*/
		alloc_lookups_and_criticalities(chan_width_dist,
										router_opts, 
										det_routing_arch, 
										segment_inf,
										timing_inf, 
										*subblock_data_ptr,
										&net_delay, &net_slack);
		
		num_connections = count_connections();
		
		remember_net_delay_original_ptr = net_delay;
	}
	else {
		num_connections = 0;
	}
	
	/* First read in the placement.   */
	
	parse_placement_file (place_file, net_file, arch_file);
	
	/* Load the channel occupancies and cost factors so that:   *
		* (1) the cost check will be OK, and                       *
		* (2) the geometry will draw correctly.                    */
	
	chan_width_factor = placer_opts.place_chan_width;
	init_chan (chan_width_factor, chan_width_dist);
	
	/* NB:  dummy_x and dummy_y used because I'll never use the old_place_occ *
		* arrays in this routine.  I need the placement structures loaded for    *
		* comp_cost and check_place to work.                                     */
	
	alloc_and_load_placement_structs (placer_opts.place_cost_type, 
									  placer_opts.num_regions, 
									  placer_opts.place_cost_exp, 
									  &dummy_x,  &dummy_y, placer_opts); 
	
	/* Need cost in order to call check_place. */
	
	bb_cost = comp_bb_cost (NORMAL, placer_opts.place_cost_type, 
							placer_opts.num_regions);
	
	/* FIX THIS */
	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {
		
		for (inet = 0; inet<num_nets; inet++)  
			for (ipin=1; ipin<net[inet].num_pins; ipin++)
				timing_place_crit[inet][ipin] = 0;  /*dummy crit values*/
		
		comp_td_costs(&timing_cost, &delay_cost);  /*set up point_to_point_delay_cost*/
		
		
		net_delay = point_to_point_delay_cost;  /*this keeps net_delay up to date with the *
			*same values that the placer is using     */
		load_timing_graph_net_delays(net_delay);
		est_crit = load_net_slack(net_slack, 0);
		
		printf("Placement. bb_cost: %g  delay_cost: %g.\n\n", 
			   bb_cost, delay_cost);
#ifdef PRINT_SINK_DELAYS
		print_sink_delays("Placement_Sink_Delays.echo");
#endif
#ifdef PRINT_NET_SLACKS
		print_net_slack("Placement_Net_Slacks.echo", net_slack);
#endif
#ifdef PRINT_PLACE_CRIT_PATH
		//print_critical_path("Placement_Crit_Path.echo");
#endif
		printf("Placement Estimated Crit Path Delay: %g\n\n", est_crit);
	}
	else {
		timing_cost = 0;
		delay_cost = 0;
		printf("Placement bb_cost is %g.\n", bb_cost);
	}
	check_place (bb_cost, timing_cost, placer_opts.place_cost_type, placer_opts.num_regions, 
				 placer_opts.place_algorithm, delay_cost);
	
	free_placement_structs (placer_opts.place_cost_type, placer_opts.num_regions,
							dummy_x, dummy_y, placer_opts);
	
	if (placer_opts.place_algorithm == NET_TIMING_DRIVEN_PLACE ||
		placer_opts.place_algorithm == PATH_TIMING_DRIVEN_PLACE ||
		placer_opts.enable_timing_computations) {
		net_delay = remember_net_delay_original_ptr;
		free_lookups_and_criticalities( &net_delay, &net_slack);
	}
	
	init_draw_coords ((float) chan_width_factor);
	
	sprintf (msg, "Placement from file %s.  bb_cost %g.", place_file,
			 bb_cost);
	update_screen (MAJOR, msg, PLACEMENT, FALSE);
}

static int max_carry_length()
{
	int i, max;
	
	max = 0;
	for(i=0;i<num_blocks;i++) {
		max = MAX(max, block[i].carry_head);
	}
	return(max);
}

static void get_list_of_carry_heads(int *num_ptr,
                                    int **heads,
                                    int **lengths)
{
	int i;
	
	*num_ptr = 0;
	for(i=0;i<num_blocks;i++) {
		if (block[i].carry_head) {
			(*heads)[*num_ptr] = i;
			(*lengths)[*num_ptr] = block[i].carry_head;
			(*num_ptr)++;
		}
	}
}


static void find_to_veb (int x_from, int y_from, int type, float rlim, int *x_to, int *y_to, int veb_index, int move_height) 
{
		
	int x_rel, y_rel, iside, iplace, rlx, rly, ok, dest_block, i;	
	int rand_module;
	int rand_block;
	int dest_block_index;
				
	ok = 0;
	while (!ok) 
	{ 			
			rand_module = my_irand(number_module);
			//printf("rand_block %d (type %d), fixed= %d,lx = %d, ly = %d, from constraint %d (type %d), lx = %d , ly = %d \n", rand_module, constraint[rand_module].type, constraint[rand_module].fixed, constraint[rand_module].lx, constraint[rand_module].ly,  veb_index, constraint[veb_index].type, constraint[veb_index].lx, constraint[veb_index].ly);
			if (constraint[rand_module].type == constraint[veb_index].type)
			{				
				//printf("Dest_block %d \n", dest_block);
				
				*x_to =constraint[rand_module].lx;
				*y_to = constraint[rand_module].ly;

				ok = (clb[*x_to][*y_to].type == EMB);			
				//ok = (clb[*x_to][*y_to].occ== 1);			
				
				//printf("x_to %d, y_to %d, type = %d, occ =%d,  to type = %d  \n", *x_to, *y_to, clb[*x_to][*y_to].type, clb[*x_to][*y_to].occ, clb[x_from][y_from].type);
				
				ok = ok & (constraint[rand_module].fixed == 0);								
									
				if ((x_from == *x_to) && (y_from == *y_to)) 
				{
					ok = 0;
				}

			}																										
	} 

	//printf("ok \n");
}

