#include <string.h>
#include <stdio.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "read_netlist.h"
#include "hash.h"
#include "check_netlist.h"
#include "read_activity.h"
#include <assert.h>

 /*----------------------------------------------------------------------------
  *  Copyright 2002 By Kara K.W. Poon and Steve Wilton
  *
  * Permission to use, copy, and modify this software and its documentation is
  * hereby granted only under the following terms and conditions.  Both the
  * above copyright notice and this permission notice must appear in all copies
  * of the software, derivative works or modified versions, and any portions
  * thereof, and both notices must appear in supporting documentation.
  * Permission is granted ONLY for non-commercial use.  For
  * commercial use, please contact the authors.
  *
  * This software may be distributed (but not offered for sale or transferred
  * for compensation) to third parties, provided such third parties agree to
  * abide by the terms and conditions of this notice.
  *
  * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHORS, AS
  * WELL AS THE UNIVERSITY OF BRITISH COLUMBIA AND THE
  * UNIVERSITY OF TORONTO DISCLAIM ALL WARRANTIES WITH
  * REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL THE
  * AUTHORS OR THE UNIVERSITY OF BRITISH COLUMBIA OR THE
  * UNIVERSITY OF TORONTO BE LIABLE FOR ANY SPECIAL, DIRECT,
  * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
  * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
  * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
  * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  *----------------------------------------------------------------------------*/
/*
  VPR was first written by Vaugh Betz at the University of 
  Toronto. Descriptions on different versions of VPR is available
  in the descript.txt file.

  The power model is an extension of the original VPR. This 
  feature is added by Kara Poon at the University of British 
  Columbia as a master thesis project supervised by Dr. Steve
  Wilton.

 */

/* This source file reads an ".ac2" file.  An example .ac2 file is    *
 * shown below:                                                       *
 *     global_net_probability clk 0.5                                 *
 *     global_net_density clk 2.0                                     *
 *     intercluster_probability a 0.5                                 *
 *     intercluster_net_density a 0.2                                 *
 *     intercluster_net_probability b 0.1                             *
 *     intercluster_net_density b 0.3                                 *
 *     intercluster_net_probability z 0.5                             *
 *     intercluster_net_density z 0.5                                 *
 *                                                                    *
 *     subblock_probability z 0.5 0.1 0.0 0.0 0.5 0.2 0.5             *
 *     subblock_density z 0.2 0.3 0.0 0.0 2.0 0.4 0.5                 *
 *                                                                    *
 *     embeddedblock_probability dspblk 0.5 0.5 0.5 0.5 2 2           *
 *     embeddedblock_density dspblk 0.2 0.2 0.2 0.2 0.5 0.5           *
 *
 * Each global net in the design has a corresponding global_net_      *
 * probability/density line.  Following the keyword                   *
 * global_net_probability/density, the                                *
 * global netname is listed, followed by the probability or the       *
 * expected number of transitions per clock cycle.  Note that for a   *
 * simple clock, this is usually 2 for its transition density, but it *
 * doesn't have to be so.                                             *
 * Nets that connect clusters (ie. use the general-purpose routing)   *
 * are listed using "intercluster_net_probability/density" lines.     *
 * The format of each of these lines is the same as the               *
 * global_net_probability/density lines.                              *
 * Each subblock (LUT and/or FF) has a corresponding                  *
 * subblock_probability/density line. Following the keyword           *
 * subblock_probability/density, the subblock name                    *
 * is listed.  Then, the activities of all inputs (usually 4, but     *
 * not necessarily) are then listed, followed by the clock            *
 * probability/density Then, the activity of the node between the LUT *
 * and FF (if there is both a LUT and FF used within this subblock)   *
 * is listed.  (this  number is 0 if a FF is not used).  Finally,     *
 * the activity of the output is listed.  Note that some of this is   *
 * redundant information, but that is ok.  It makes it more general,  *
 * in case we want to change things later                             *
 * Embedded blocks (DSP and Memories are specified using              *
 * embeddedblock_probability and embeddedblock_density.  In this case *
 * the inputs are listed first, followed by the outputs, followed     *
 * by the clocks.  All of this is redundant information (since you    *
 * can get it from the nets attached to the pins) but again, it will  *
 * make it easier if we want to change things later.                  */

void read_activity(char *filename, t_subblock_data *subblock_data_ptr)
{
   char buf[BUFSIZE], *ptr, *word1, *word2,*word3;
   FILE *fp;
   int i,j,k,lut_size,match;
   int this_net;
   int action;
   int marker;
   int numsubblocks;
   int num_pins;
   int num_inputs, num_outputs, num_clocks;
   int num_dsp_inputs, num_dsp_outputs, num_dsp_clocks;
   int num_memory_inputs, num_memory_outputs, num_memory_clocks;
   int tot_pins = 0;
   float tot_activity = 0.0;

   if (num_nets == 0) {
      printf("Error reading activity file.  Must read netlist first\n");
      exit(-1);
   }
   for(i=0;i<num_nets;i++) {
      //net[i].activity = -1.0; /* uninitialized */
      net[i].probability = -1.0;
      net[i].density = -1.0;
   }
   num_dsp_inputs = 0; num_dsp_outputs = 0; num_dsp_clocks = 0;
   for(j=0;j<pins_per_dsp;j++) {
              if (dsp_is_output_pin[j]) {
                 num_dsp_outputs ++;
	      } else {
                 if (is_global_dsp_pin[j]) {
                    num_dsp_clocks ++;
		 } else {
                    num_dsp_inputs ++;
		 }
	      }
   }
   num_memory_inputs = 0; num_memory_outputs = 0; num_memory_clocks = 0;
   for(j=0;j<pins_per_memory;j++) {
              if (memory_is_output_pin[j]) {
                 num_memory_outputs ++;
	      } else {
                 if (is_global_memory_pin[j]) {
                    num_memory_clocks ++;
		 } else {
                    num_memory_inputs ++;
		 }
	      }
   }

   /* set default blocks to activity 0 */

   lut_size = subblock_data_ptr->subblock_lut_size;
   for(i=0;i<num_blocks;i++) {

        if (block[i].type == DSP) {
           num_inputs = num_dsp_inputs;
           num_outputs = num_dsp_outputs;
           num_clocks = num_dsp_clocks;
           num_pins = pins_per_dsp;

	} else if (block[i].type == MEMORY) {
           num_inputs = num_memory_inputs;
           num_outputs = num_memory_outputs;
           num_clocks = num_memory_clocks;
           num_pins = pins_per_memory;
	} else {
           num_inputs = lut_size;
           num_outputs = 1;
           num_clocks = 1;
           num_pins = lut_size + 2;
	}

        for(j=0;j<subblock_data_ptr->num_subblocks_per_block[i];j++) {
	
           subblock_data_ptr->subblock_inf[i][j].pin_probability = 
                            (float *)malloc(sizeof(float)*num_pins);
           subblock_data_ptr->subblock_inf[i][j].pin_density = 
                            (float *)malloc(sizeof(float)*num_pins);

           for(k=0;k<num_pins;k++) {
              subblock_data_ptr->subblock_inf[i][j].pin_density[k] = 0.0;
              subblock_data_ptr->subblock_inf[i][j].pin_probability[k] = 0.0;
	   }
	   subblock_data_ptr->subblock_inf[i][j].between_lut_and_ff_probability= 0.0;
	   subblock_data_ptr->subblock_inf[i][j].between_lut_and_ff_density = 0.0;				 
	}
   }

   fp = my_fopen (filename, "r", 0);
   if (fp == NULL) {
      printf("Error: could not open activity file: %s\n",filename);
      exit(-1);
   } else {
       ptr = my_fgets (buf, BUFSIZE, fp);
       linenum = 0;
       while (ptr != NULL) {
          word1 = strtok(buf,TOKENS);

          marker = strlen(word1)+1;
          action = 0;

			 if (strcmp(word1,"global_net_probability")==0) 
				{
				  action = 1;
				}
			 else if (strcmp(word1,"global_net_density")==0)
				{
				  action = 2;
				}
			 else if (strcmp(word1,"intercluster_net_probability")==0)
				{
				  action = 3;
				}
			 else if (strcmp(word1,"intercluster_net_density")==0)
				{
				  action = 4;
				}
			 else if (strcmp(word1,"subblock_probability")==0) 
				{
				  action = 5;
				} 
			 else if (strcmp(word1,"subblock_density")==0) 
				{
				  action = 6;
				} 
			 else if (strcmp(word1,"embeddedblock_probability")==0) 
				{
				  action = 7;
				} 
			 else if (strcmp(word1,"embeddedblock_density")==0) 
				{
				  action = 8;
				} 
			 
			 else {
				printf("Illegal keyword in activity file %s\n",filename);
				printf("Keyword read = %s, expected global_net_activity,\n",word1);
				printf("intercluster_net_activity, or subblock_activity\n");
				exit(0);
			 }

          switch(action) {

			 case 1:
			 case 2:
			 case 3:
			 case 4:

				/* get net name */
				
				word2 = strtok(&(buf[marker]),TOKENS);
				marker += strlen(word2) + 1;
				
				/* search for a net with this name */
				
				this_net = -1;
				for(i=0;i<num_nets;i++) {
				  if (strcmp(word2,net[i].name)==0) {
					 this_net = i;  i=num_nets+1;
	           } 
				} 
				if (this_net == -1) {
				  printf("WARNING: Activity file contains net %s, but %s is not ",word2,word2);
				  printf("in the netlist.  Ignoring.\n");
				} else { 
				
	   			   word3 = strtok(&(buf[marker]),TOKENS);
				   if ((action == 1)||(action == 3))
				   {
				 	 net[this_net].probability = atof(word3);
				   }
				   else if ((action == 2)||(action == 4))
				   {
					 net[this_net].density =  atof(word3);
				   }

				
   				   if (((action == 3)||(action == 4)) && (is_global[this_net])) {
				     printf("Error in reading activity file %s\n",filename);
				     printf("Activity file says that %s is non-global\n",word2);
				     printf("while netlist specifies this net is global\n");
				     exit(-1);
			 	   }
				

				   if (((action == 1)||(action == 2)) && (!is_global[this_net])) {
				     printf("Error in reading activity file %s\n",filename);
				     printf("Activity file says that %s is global\n",word2);
				     printf("while netlist specifies this net is not global\n");
				     exit(-1);
				   }
				}

				break;

			 case 5:
			 case 6:

                                num_inputs = lut_size;
                                num_outputs = 1;
                                num_clocks = 1;

				/* get subblock name */
				
				word2 = strtok(&(buf[marker]),TOKENS);
				marker += strlen(word2) + 1;
				
				/* search for a subblock with this name */
				
				match = 0;
				for(i=0;i<num_blocks;i++) {
				  numsubblocks = subblock_data_ptr->num_subblocks_per_block[i];
				  
				  for(j=0;j<numsubblocks;j++) {
					 
					 if (strcmp(word2,
									subblock_data_ptr->subblock_inf[i][j].name)==0) {
									
						/* Found a matching name */
						/* Read in lut_size input activities */
				  	
						for(k=0;k<lut_size;k++) {
						  
						  word3 = strtok(&(buf[marker]),TOKENS);
						  
						  marker += strlen(word3) + 1;

						  if (action == 5)
							 {
								subblock_data_ptr->subblock_inf[i][j].pin_probability[k] = atof(word3);
							 }
						  else if (action == 6)							 
							 {
								subblock_data_ptr->subblock_inf[i][j].pin_density[k] = atof(word3);
							 }
						}
							
						/* Read in clock activity */
				  
						word3 = strtok(&(buf[marker]),TOKENS);
						marker += strlen(word3) + 1;

						if (action == 5)
						  {
							 subblock_data_ptr->subblock_inf[i][j].pin_probability[lut_size+1] = atof(word3);
						  }
						else if (action == 6)		
						  {
							 subblock_data_ptr->subblock_inf[i][j].pin_density[lut_size+1] = atof(word3);
						  }

						/* Read in activity between lut and ff */
				 	
						word3 = strtok(&(buf[marker]),TOKENS);
						marker += strlen(word3) + 1;

						
						if (action == 5)
						  {
							 subblock_data_ptr->subblock_inf[i][j].between_lut_and_ff_probability = atof(word3);
						  }
						else if (action == 6)	
						  {
							 subblock_data_ptr->subblock_inf[i][j].between_lut_and_ff_density = atof(word3);
						  }
							
						/* Read in subblock output activity */
				 	
						word3 = strtok(&(buf[marker]),TOKENS);
						marker += strlen(word3) + 1;

						if (action == 5)
						  {						
							 subblock_data_ptr->subblock_inf[i][j].pin_probability[lut_size] = atof(word3);
						  }
						else if (action == 6)	
						  {
							 subblock_data_ptr->subblock_inf[i][j].pin_density[lut_size] = atof(word3);
						  }
						
						i=BIG_INT; j=BIG_INT;
						match = 1;
							
					 } /* end if */
				  } /* end for */
				}/* end for */
	  
		 
				if (match==0) {
				  printf("WARNING: Activity file contains block %s, but %s is not ",
							word2,word2);
				  printf("in the netlist.\n");
				}
				
				break;


			 case 7:
			 case 8:

				/* get embeddedblock name */
				
				word2 = strtok(&(buf[marker]),TOKENS);
				marker += strlen(word2) + 1;
				
				/* search for a subblock with this name */
				
				match = 0;
				for(i=0;i<num_blocks;i++) {

				  if (subblock_data_ptr->num_subblocks_per_block[i] > 0) {

				  if (strcmp(word2,subblock_data_ptr->subblock_inf[i][0].name)==0) {
			
					/* Found a matching name */
        if (block[i].type == DSP) {
           num_inputs = num_dsp_inputs;
           num_outputs = num_dsp_outputs;
           num_clocks = num_dsp_clocks;
           num_pins = pins_per_dsp;
	} else if (block[i].type == MEMORY) {
           num_inputs = num_memory_inputs;
           num_outputs = num_memory_outputs;
           num_clocks = num_memory_clocks;
           num_pins = pins_per_memory;
	}

					for(k=0;k<num_pins;k++) {
					   word3 = strtok(&(buf[marker]),TOKENS);
				           marker += strlen(word3) + 1;
				           if (action == 7) {
                                              if (strcmp(word3,"OPEN") == 0) {
  					         subblock_data_ptr->subblock_inf[i][0].pin_probability[k] = 0.0;
					      } else {
  					         subblock_data_ptr->subblock_inf[i][0].pin_probability[k] = atof(word3);

					      }
					   } else if (action == 8) {							 
                                              if (strcmp(word3,"OPEN") == 0) {
  					         subblock_data_ptr->subblock_inf[i][0].pin_density[k] = 0.0;
					      } else {
  					         subblock_data_ptr->subblock_inf[i][0].pin_density[k] = atof(word3);
					      }
					   }
					}


					i=BIG_INT; j=BIG_INT;
					match = 1;
				  }
				  }/* end if */
				}/* end for */
	  
		 
				if (match==0) {
				  printf("Error in reading activity file %s\n",filename);
				  printf("Activity file contains embeddedblock %s, but %s is not ",
							word2,word2);
				  printf("in the netlist.\n");
				  exit(-1);
				}
				
				break;

				
				
			 }
          ptr = my_fgets (buf, BUFSIZE, fp);
       }
       fclose(fp);

       /* Make sure all nets have an activity value */

       for(i=0;i<num_nets;i++) {
			  if ((net[i].probability < 0.0)||(net[i].density< 0.0)) {
             printf("Activity file does not contain signal %s\n",
                          net[i].name);
             exit(-1);
	  }
       }
   }

   tot_activity = 0.0;
   tot_pins = 0;
   max_activity = 0.0; 
   for(i=0;i<num_nets;i++) {
      max_activity = (net[i].density > max_activity) ? 
		     net[i].density : max_activity;
      tot_activity += net[i].density*net[i].num_pins;
      tot_pins += net[i].num_pins;
   }
   avg_activity = tot_activity/((float) tot_pins);
}




void read_function_file(char *filename, t_subblock_data *subblock_data_ptr)
{
   char buf[BUFSIZE], *ptr, *word1, *word2,*word3;
   FILE *fp;
   int i,j,k,lut_size,match;
   int this_net;
   int action;
   int marker;
   int numsubblocks;

   if (num_blocks == 0) {
      printf("Error reading function file.  Must read netlist first\n");
      exit(-1);
   }

   lut_size = subblock_data_ptr->subblock_lut_size;
   fp = my_fopen (filename, "r", 0);
   if (fp == NULL) {
      printf("Error: could not open function file: %s\n",filename);
      exit(-1);
   } else {
	  ptr = my_fgets (buf, BUFSIZE, fp);
	  while (ptr != NULL) {
		 word1 = strtok(buf,TOKENS);
		 
		 marker = strlen(word1)+1;
		 if (strcmp(word1,"subblock_function")!=0) {
			printf("Illegal keyword in activity file %s\n",filename);
			printf("Keyword read = %s, expected subblock_function.\n",word1);
			exit(0);
		 }
		 
		 word2 = strtok(&(buf[marker]),TOKENS);
		 marker += strlen(word2) + 1;
		 
		 /* search for a subblock with this name */
		 
		 match = 0;
		 for(i=0;i<num_blocks;i++) {
			numsubblocks = subblock_data_ptr->num_subblocks_per_block[i];
			
			for(j=0;j<numsubblocks;j++) {
			  
			  if (strcmp(word2,
							 subblock_data_ptr->subblock_inf[i][j].name)==0) {
				 
				 /* Found a matching name */
				 
				 /* Read in lut_size input activities */
				 word3 = strtok(&(buf[marker]),TOKENS);
				 
				 if (strlen(word3) != (1<<lut_size)) {
					printf("activity specification %s\n",word3);
					printf("is of length %d.  Was expecting a\n",strlen(word3));
					printf("a function of length %d.\n",1<<lut_size);
					exit(0);
				 }
				 
				 for(k=0;k<strlen(word3);k++) {
					subblock_data_ptr->subblock_inf[i][j].function[k] =
					  word3[k];
				 }
				 
				 i=BIG_INT; j=BIG_INT;
				 match = 1;
				 
			  }
		   }
		 }
		 
		 
		 if (match==0) {
			printf("WARNING: Activity file contains block %s, but %s is not ",
					 word2,word2);
			printf("in the netlist.\n");
		 }
		 
		 
		 ptr = my_fgets (buf, BUFSIZE, fp);  
		 
	  }
	}
	fclose(fp);
	
}


