#include <string.h>
#include <stdio.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "read_netlist.h"
#include "hash.h"
#include "check_netlist.h"
#include <assert.h>


/* This source file reads in a .net file.  A .net file is a netlist   *
 * format defined by me to allow arbitary logic blocks (clbs) to      *
 * make up a netlist.  There are three legal block types: .input,     *
 * .output, and .clb.  To define a block, you start a line with one   *
 * of these keywords; the function of each type of block is obvious.  *
 * After the block type keyword, you specify the name of this block.  *
 * The line below must start with the keyword pinlist:, and gives a   *
 * list of the nets connected to the pins of this block.  .input and  *
 * .output blocks are pads and have only one net connected to them.   *
 * The number of pins on a .clb block is specified in the arch-       *
 * itecture file, as are the equivalences between pins.  Each         *
 * clb must have the same number of pins (heterogeneous FPGAs are     *
 * currently not supported) so any pins which are not used on a clb   *
 * must be identified with the reserved word "open".  All keywords    *
 * must be lower case.                                                *
 *                                                                    *
 * The lines immediately below the pinlist line must specify the      *
 * contents of the clb.  Each .subblock line lists the name of the    *
 * subblock, followed by the clb pin number or subblock output to     *
 * which each subblock pin should connect.  Each subblock is assummed *
 * to consist of a LUT with subblock_lut_size inputs, a FF, and an    *
 * output.  The pin order is input1, input2, ... output, clock.  The  *
 * architecture file sets the number of subblocks per clb and the LUT *
 * size used.  Subblocks are used only for timing analysis.  An       *
 * example clb declaration is:                                        *
 *                                                                    *
 * .clb name_of_clb  # comment                                        *
 *  pinlist:  net_1 net_2 my_net net_of_mine open D open              *
 *  subblock: sub_1 0 1 2 3 open open open                            *
 *  subblock: sub_2 1 3 ble_0 open open 5 open                        *
 *                                                                    *
 * Notice that the output of the first subblock (ble_0) is used only  *
 * by the second subblock.  This is fine.                             *
 *                                                                    *
 * Ending a line with a backslash (\) means it is continued on the    *
 * line below.  A sharp sign (#) indicates the rest of a line is      *
 * a comment.                                                         *
 * The vpack program can be used to convert a flat blif netlist       *
 * into .net format.                                                  *
 *                                                                    *
 * V. Betz, Jan. 29, 1997.                                            */

/* A note about the way the character buffer, buf, is passed around. *
 * strtok does not make a local copy of the character string         *
 * initially passed to it, so you must make sure it always exists.   *
 * Hence, I just use one buffer declared in read_net and pass it     *
 * downstream.  Starting a new line with an automatic variable       *
 * buffer in a downstream subroutine and then returning and trying   *
 * to keep tokenizing it would cause problems since the buffer now   *
 * lies on a stale part of the stack and can be overwritten.         */


/*************************** Variables local to this module *****************/

/* Temporary storage used during parsing. */

static int *num_driver, *temp_num_pins;
static struct s_hash **hash_table;
static int temp_block_storage;

/* Used for memory chunking of everything except subblock data. */

static int chunk_bytes_avail = 0;
static char *chunk_next_avail_mem = NULL;


/* Subblock data can be accessed anywhere within this module.  Pointers to *
 * the main subblock data structures are passed back to the rest of the    *
 * program through the subblock_data_ptr structure passed to read_net.     */

static int max_subblocks_per_block;
static int subblock_lut_size;
static int subblock_num_pin;
static int subblock_num_ipin;
static int subblock_num_opin;
static t_subblock **subblock_inf;
static int *num_subblocks_per_block;

static struct s_name_convert *name_convert;
static int num_name_convert;

/* The subblock data is put in its own "chunk" so it can be freed without  *
 * hosing the other netlist data.                                          */

static int ch_subblock_bytes_avail;
static char *ch_subblock_next_avail_mem;
static struct s_linked_vptr *ch_subblock_head_ptr;

struct temporary_carry_name_info_t {
   char **cluster_name;
   char **ble_name;
   int num_cluster;
   int num_ble_per_cluster;
} temporary_carry_name_info_t;

static struct temporary_carry_name_info_t *temporary_carry_name_info = NULL;


/************************ Subroutines local to this module ******************/

static int add_net (char *ptr, enum e_pin_type type, int bnum, int blk_pnum, 
       int doall); 

static char *get_tok(char *buf, int doall, FILE *fp_net, char *veb_link_file, int pass);
static void add_io (int doall, int type, FILE *fp_net, char *buf);
static char *add_clb (int doall, FILE *fp_net, char *buf);
static char *add_veb (int doall, FILE *fp_net, char *buf, char *veb_link_file);
static char *add_veb_clb (int doall, FILE *fp_veb, char *buf, char *veb_name, int module_number, struct s_name_convert *name_convert, int num_name_convert);
static int veb_parse_name_and_pinlist (int doall, FILE *fp_veb, char *buf, char *veb_name) ;
static char *add_dsp (int doall, FILE *fp_net, char *buf);
static char *add_memory (int doall, FILE *fp_net, char *buf);
static void add_global (int doall, FILE *fp_net, char *buf);
static void init_parse(int doall);
static void free_parse (void);
static void parse_name_and_pinlist (int doall, FILE *fp_net, char *buf,int type);
static int get_pin_number (char *ptr); 

static void load_subblock_array (int doall, FILE *fp_net, char *temp_buf, 
     int num_subblocks, int bnum, int type, char *veb_name); //type : 0 - normal clb, 1 - veb clb
static void load_carryout (int doall, FILE *fp_net, char *temp_buf, 
     int bnum); 
static int carry_length(int iblk);

static void add_one_empty_subblock(int doall, int bnum);

static void set_subblock_count (int bnum, int num_subblocks); 
static char *parse_subblocks (int doall, FILE *fp_net, char *buf, int bnum, int type, char *veb_name);  //type : 0 - normal clb, 1 - veb clb
static void parse_carry_cluster_name(char * netname, char* cluster_name);	//get the carry cluster name in from the net name
static int find_carry_from_block(int net_num);
static int find_carry_to_block(int net_num);
static char * veb_convert_veb_net (char * in_string);
static char * veb_convert_top_net (char * in_string);
static void preprocess_veb_io ();

static char *add_emb (int doall, FILE *fp_net, char *buf, char *emb_link_file) ;



/********************** Subroutine definitions *******************************/




void allocate_temporary_carry_name_info(int num_cluster, int num_ble_per_cluster)
{
   int i,j;

   temporary_carry_name_info = (struct temporary_carry_name_info_t *)
           my_malloc(sizeof(temporary_carry_name_info_t));
   temporary_carry_name_info->num_cluster = num_cluster;
   temporary_carry_name_info->num_ble_per_cluster = num_ble_per_cluster;
   temporary_carry_name_info->cluster_name = (char **) my_malloc(
                  sizeof(char *)*num_cluster*num_ble_per_cluster);
   temporary_carry_name_info->ble_name = (char **) my_malloc(
                  sizeof(char *)*num_cluster*num_ble_per_cluster);
   for(i=0;i<num_cluster;i++) {
      for(j=0;j<num_ble_per_cluster;j++) {
         temporary_carry_name_info->cluster_name[i*num_ble_per_cluster+j] = (char *)NULL;
         temporary_carry_name_info->ble_name[i*num_ble_per_cluster+j] = (char *)NULL;
      }
   }
}

void free_temporary_carry_name_info()
{
   int i;

   if (temporary_carry_name_info == NULL) return;

   for(i=0;i<temporary_carry_name_info->num_ble_per_cluster *
             temporary_carry_name_info->num_cluster;i++) {
      if (temporary_carry_name_info->cluster_name[i] != (char *)NULL) {
         free(temporary_carry_name_info->cluster_name[i]);
      }
      if (temporary_carry_name_info->ble_name[i] != (char *)NULL) {
         free(temporary_carry_name_info->ble_name[i]);
      }
   }
   free(temporary_carry_name_info->cluster_name);
   free(temporary_carry_name_info->ble_name);
   free(temporary_carry_name_info);
}


char *get_temporary_carry_name_cluster_name(int icluster, int iblk)
{
   if (temporary_carry_name_info == NULL) {
      printf("Error accessing temporary_carry_name_info\n"); 
      exit(1);
   }
   if (icluster >= temporary_carry_name_info->num_cluster) {
      printf("Illegal cluster number %d in get_temporary_carry_name_cluster_name\n",
          icluster);
      exit(1);
   }
   if (iblk >= temporary_carry_name_info->num_ble_per_cluster) {
      printf("Illegal ble number %d in get_temporary_carry_name_cluster_name\n",
          icluster);
      exit(1);
   }
   return(temporary_carry_name_info->cluster_name[icluster*  
                  temporary_carry_name_info->num_ble_per_cluster + iblk]);
}

char *get_temporary_carry_name_ble_name(int icluster, int iblk)
{
   if (temporary_carry_name_info == NULL) {
      printf("Error accessing temporary_carry_name_info\n"); 
      exit(1);
   }
   if (icluster >= temporary_carry_name_info->num_cluster) {
      printf("Illegal cluster number %d in get_temporary_carry_name_ble_name\n",
          icluster);
      exit(1);
   }
   if (iblk >= temporary_carry_name_info->num_ble_per_cluster) {
      printf("Illegal ble number %d in get_temporary_carry_name_ble_name\n",
          icluster);
      exit(1);
   }
   return(temporary_carry_name_info->ble_name[icluster*  
                  temporary_carry_name_info->num_ble_per_cluster + iblk]);
}

void set_temporary_carry_name_cluster_name(int icluster, int iblk, char *cluster_name)
{
   if (temporary_carry_name_info == NULL) {
      printf("Error accessing temporary_carry_name_info\n"); 
      exit(1);
   }
   if (icluster >= temporary_carry_name_info->num_cluster) {
      printf("Illegal cluster number %d in set_temporary_carry_name_info\n",
          icluster);
      exit(1);
   }
   if (iblk >= temporary_carry_name_info->num_ble_per_cluster) {
      printf("Illegal ble number %d in set_temporary_carry_name_info\n",
          icluster);
      exit(1);
   }

   temporary_carry_name_info->cluster_name[icluster*temporary_carry_name_info->num_ble_per_cluster + iblk] =
           (char *)my_malloc(1+(int)strlen(cluster_name));
   strcpy(temporary_carry_name_info->cluster_name[icluster*temporary_carry_name_info->num_ble_per_cluster + iblk],
          cluster_name);
}

void set_temporary_carry_name_ble_name(int icluster, int iblk, char *ble_name)
{
   if (temporary_carry_name_info == NULL) {
      printf("Error accessing temporary_carry_name_info\n"); 
      exit(1);
   }
   if (icluster >= temporary_carry_name_info->num_cluster) {
      printf("Illegal cluster number %d in set_temporary_carry_name_info\n",
          icluster);
      exit(1);
   }
   if (iblk >= temporary_carry_name_info->num_ble_per_cluster) {
      printf("Illegal ble number %d in set_temporary_carry_name_info\n",
          icluster);
      exit(1);
   }

   temporary_carry_name_info->ble_name[icluster*temporary_carry_name_info->num_ble_per_cluster + iblk] =
           (char *)my_malloc(1+(int)strlen(ble_name));
   strcpy(temporary_carry_name_info->ble_name[icluster*temporary_carry_name_info->num_ble_per_cluster + iblk],
          ble_name);
}


static int carry_length(int iblk)
{
   int j,cnt;
   int current_iblk;

   //printf ("Carry head : %d (%s) \n", iblk, block[iblk].name);
   if (block[iblk].carryout_cluster == -1) {return(1);}

   cnt = 1;
   j = iblk;
   
//   printf ("Carry body : \n");
   while(block[j].carryout_cluster != -1) {
		current_iblk = j;
		j = block[j].carryout_cluster;
//		printf("\t %d (%s) \n", j, block[j].name);
		//check the cin of the next carry chain is the same as the cout of this carry body or not
		//if (block[j].carryin_cluster == current_iblk)
		//{		
			cnt++;
		//}
//		else
//		{
//			printf ("**Found carryout not equal to carryin of current_block %d (%s), carryout block %d (%s) \n", current_iblk, block[current_iblk].name, j, block[j].name);
//			return(cnt);
//		}
   }
   return(cnt);
}


static int find_cluster_with_name(t_subblock_data *subblock_data_ptr, char *name)
{
   int i;

   for(i=0;i<num_blocks;i++) {
      if (strcmp(block[i].name,name)==0) return(i);
   }
   return(-1);
}

static int find_block_inside_cluster_with_name(t_subblock_data *subblock_data_ptr,
                                               int icluster,char *name)
{
   int i;

   for(i=0;i<subblock_data_ptr->num_subblocks_per_block[icluster];i++) {
      if (strcmp(subblock_data_ptr->subblock_inf[icluster][i].name,name)==0) return(i);
   }
   return(-1);
}

static int find_carry_from_block(int net_num)
{
   //find the block which connects its cout to the target cin
   int i;   
   int block_cout_pin = SLICE_COUT;; //is the cout pin of each block

   for(i=0;i<num_blocks;i++) {
      if (block[i].nets[block_cout_pin] == net_num) 
		  return(i);
   }
   return(-1);
}

static int find_carry_to_block(int net_num)
{
   //find the block which connects its cin from the target cout
   int i;      
   int block_cin_pin = SLICE_CIN;;	// is the cin pin of each block

   for(i=0;i<num_blocks;i++) {
      if (block[i].nets[block_cin_pin] == net_num) 
		  return(i);
   }
   return(-1);
}




void update_carryouts(t_subblock_data *subblock_data_ptr) 
{
   int i,j;
   //char *dest_block_name, *dest_cluster_name;
   //int dest_block_num, dest_cluster_num, found_exit;
   int *tmp;

   int has_cin;
   int has_cout;
   int found_exit;

   //int block_cin_pin = SLICE_CIN;	//pin 32 is the cin pin of each block
   //int block_cout_pin = SLICE_COUT; //pin 14 is the cout pin of each block

   int block_cin_pin = SLICE_CIN;	//the cin pin of each block
   int block_cout_pin = SLICE_COUT; //the cout pin of each block

   int subblock_cin_ipin = I_CIN; //ipin 1 is the cin pin of each subblock input
   int subblock_cin_pin = B_CIN; //pin 2 is the cin pin of each subblock I/O
   int subblock_cout_opin = O_COUT; //opin 11 is the cout pin of each block output
   int subblock_cout_pin = B_COUT; //pin 3 is the cout pin of each block I/O

   int from_block_num, to_block_num;   
   int from_subblock_num, to_subblock_num;


	printf("------------------------------------------------------------------------------------------\n");
	printf("update_carryouts\n");
   for(i=0;i<num_blocks;i++) {
      block[i].carryin_cluster = -1;
      block[i].carryin_ble = -1;
      block[i].carryout_cluster = -1;
      block[i].carryout_ble = -1;
	  
      for(j=0;j<subblock_data_ptr->num_subblocks_per_block[i];j++) 	  
	  {
		  //printf("num_subblocks_per_block[%d] = %d, j =%d \n", i, subblock_data_ptr->num_subblocks_per_block[i], j);
         subblock_data_ptr->subblock_inf[i][j].carryin_cluster = -1;
         subblock_data_ptr->subblock_inf[i][j].carryin_ble = -1;
         subblock_data_ptr->subblock_inf[i][j].carryout_cluster = -1;
         subblock_data_ptr->subblock_inf[i][j].carryout_ble = -1;
      }
   }

   //printf("check each blocks carry_in , carry_out pin \n");
	//check each blocks carry_in, carry_out pin
	for(i=0;i<num_blocks;i++) 
	{
		found_exit = -1;
		has_cin = -1;
		has_cout = -1;

		if (block[i].type == EMB)
		{
			continue;
		}

		//check the block has cin or not ?
		if (block[i].nets[block_cin_pin] != OPEN)
		{
			has_cin = 1;
			
			//find from which block cout
			from_block_num = find_carry_from_block(block[i].nets[block_cin_pin] );
			//printf("block %d (%s) has cin , net name : %s, from block : %d\n", i, block[i].name,net[block[i].nets[block_cin_pin]].name, from_block_num);
		}

		if (block[i].nets[block_cout_pin] != OPEN)
		{
			has_cout = 1;
			//find to which block cin
			to_block_num = find_carry_to_block(block[i].nets[block_cout_pin] );
			//printf("block %d has cout , net name : %s, to block : %d \n", i, net[block[i].nets[block_cout_pin]].name, to_block_num);
		}

		//printf("has_cout %d , has_cin %d \n", has_cout, has_cin);
		if( (has_cout == 1) || (has_cin ==1))
		{
			//there is cin or cout to this block 
			//handle the carry in each subblock

			//printf("num_subblocks_per_block[%d] = %d \n", i, subblock_data_ptr->num_subblocks_per_block[i]);
			//printf("from block number %d , to block number %d \n", from_block_num ,to_block_num);
			for(j=0;j<subblock_data_ptr->num_subblocks_per_block[i];j++) 
			{
				//printf("subblock_inf[i][j].outputs[subblock_cout_opin] = %d \n", subblock_data_ptr->subblock_inf[i][j].outputs[subblock_cout_opin]);
				if (subblock_data_ptr->subblock_inf[i][j].outputs[subblock_cout_opin] < pins_per_clb)
				{
					//this subblock connect to the block cout
					if (has_cout == 1)
					{
						//the cout is connect to next block 's subblock 0
						subblock_data_ptr->subblock_inf[i][j].carryout_cluster = to_block_num;				
						subblock_data_ptr->subblock_inf[i][j].carryout_ble = 0;
						found_exit = j;
					}
				}

				//printf("subblock_inf[i][j].outputs[subblock_cin_ipin] = %d \n", subblock_data_ptr->subblock_inf[i][j].outputs[subblock_cin_ipin]);
				//printf("i = %d, j= %d , from block %d\n", i, j, from_block_num);
				if (subblock_data_ptr->subblock_inf[i][j].outputs[subblock_cin_ipin] < pins_per_clb)
				{
					//this subblock connect to the block cin
					if (has_cin == 1)
					{
						//the cin is connect from the previous block's subblock 1
						if (from_block_num != -1)
						{
							subblock_data_ptr->subblock_inf[from_block_num][1].carryin_cluster = i;
							//printf("subblock_data_ptr->subblock_inf[from_block_num][1].carryin_cluster =%d \n", subblock_data_ptr->subblock_inf[from_block_num][1].carryin_cluster);
							subblock_data_ptr->subblock_inf[from_block_num][1].carryin_ble = j;
							//printf("subblock_data_ptr->subblock_inf[from_block_num][1].carryin_ble =%d \n",subblock_data_ptr->subblock_inf[from_block_num][1].carryin_ble);
						}
					}
				}				
			}
		}

		//printf("found_exit = %d \n", found_exit);
		if (found_exit != -1) 
		{
			//the block has cout 
			block[i].carryout_cluster = subblock_data_ptr->subblock_inf[i][found_exit].carryout_cluster;
			block[i].carryout_ble = subblock_data_ptr->subblock_inf[i][found_exit].carryout_ble;
			
			//update the cin of previous block
			block[subblock_data_ptr->subblock_inf[i][found_exit].carryout_cluster].carryin_cluster = i;
			block[subblock_data_ptr->subblock_inf[i][found_exit].carryout_cluster].carryin_ble = found_exit;        
		}
   }

   tmp = (int *)my_malloc(sizeof(int)*num_blocks);

   //set the block as carry body if there is carry out
   for(i=0;i<num_blocks;i++) 
   {
      if (block[i].carryout_cluster != -1) 
	  {
         tmp[i] = 1;
         block[i].carry_body = 1;
      } 
	  else 
	  {
         tmp[i] = 0;
      }
   }

   //set the block as carry out body if there is  carry in
   for(i=0;i<num_blocks;i++) 
   {
      if (block[i].carryout_cluster != -1) 
	  {
         tmp[block[i].carryout_cluster] = 0;
         block[block[i].carryout_cluster].carry_body = 1;     
      }
   }

   //tmp is use to indicate which one is in the carry chain with cout only, no cin
   for(i=0;i<num_blocks;i++) {
      if (tmp[i]) {
         block[i].carry_head = carry_length(i);
      }
   }
   free(tmp);
/*
   printf("carry heads: \n");
   for(i=0;i<num_blocks;i++) 
      if (block[i].carry_head) printf("   %s\n",block[i].name);
   printf("carry bodies: \n");
   for(i=0;i<num_blocks;i++) 
      if (block[i].carry_body) printf("   %s\n",block[i].name);
*/
}

void read_net (char *net_file, char *veb_link_file, t_subblock_data *subblock_data_ptr) {

/* Main routine that parses a netlist file in my (.net) format. */

 char buf[BUFSIZE], *ptr;
 int doall;
 FILE *fp_net;
 int pass;
 int i;
int nindex;

int netpin, blk;
 
/* Make two variables below accessible anywhere in the module because I'm *
 * too lazy to pass them all over the place.                              */

 max_subblocks_per_block = subblock_data_ptr->max_subblocks_per_block;
 subblock_lut_size = subblock_data_ptr->subblock_lut_size;
 subblock_num_pin = subblock_data_ptr->subblock_num_pin;
 subblock_num_ipin = subblock_data_ptr->subblock_num_ipin;
 subblock_num_opin = subblock_data_ptr->subblock_num_opin;


 fp_net = my_fopen (net_file, "rb", 0);

/* First pass builds the symbol table and counts the number of pins  *
 * on each net.  Then I allocate exactly the right amount of storage *
 * for each net.  Finally, the second pass loads the block and net   *
 * arrays.                                                           */

	for (doall=0;doall<=1;doall++) 
	{  /* Pass number. */
		init_parse(doall);

    
		num_emb = 0;
    	num_name_convert = 0;
		name_convert = (struct s_name_convert *) malloc(sizeof(struct s_name_convert));
    
		for (pass = 0 ; pass < 3 ; pass ++)
		{
			linenum = 0;   /* Reset line number. */
    			//ptr = my_fgets (buf, BUFSIZE, fp_net);
			ptr = fgets (buf, BUFSIZE, fp_net);
			
			//pass 0 to get VEB
			while (ptr != NULL) 
			{
				//printf("Line %d  (len: %d) , %c : %s \n",linenum,	 strlen(ptr), ptr[strlen(ptr)-1], ptr);
				ptr = get_tok (buf, doall, fp_net, veb_link_file, pass);
			}
			rewind (fp_net);  /* Start at beginning of file again */
		}

		/*
	printf("Num net %d \n", num_nets);

	for (nindex = 0; nindex < num_nets; nindex ++)
	{
		//printf("Net name : %s, number of pins: %d , num_driver: %d, nindex %d, temp_num_pin %d \n", net[nindex].name, net[nindex].num_pins, num_driver[nindex], nindex, temp_num_pins[nindex]);		
	}
	*/
	
/*
	if (doall)
	{
		for(netpin=0;netpin<net[10869].num_pins;netpin++) 
		{
			blk=net[10869].blocks[netpin];
			printf("Net 10869(%s), netpin %d , blk %d (%s) \n", net[10869].name, netpin, blk, block[blk].name);	
		}
	}
*/
/*		
	for (i=0;i< num_name_convert; i++)
 	{
	 	printf("%s\t %s\t %d\n", name_convert[i].convert_name, name_convert[i].veb_io_name, name_convert[i].io);
	}
*/	
		//free name_convert
 		if (name_convert != NULL)
		{
			for (i=0;i< num_name_convert; i++)
			{
				if (name_convert[i].convert_name != NULL)
					free (name_convert[i].convert_name);
				if (name_convert[i].veb_io_name != NULL)
					free (name_convert[i].veb_io_name);		
			}
			free(name_convert);
		}  
   
	} 

	fclose(fp_net);

	

	printf("Finish reading netlist \n");
	/* Return the three data structures below through subblock_data_ptr.        */

	printf("num_block %d \n", num_blocks);
	/*
	for(i = 0; i< num_blocks; i++)
	{
		printf("num_subblocks_per_block[%d] = %d \n", i, num_subblocks_per_block[i]);
	}
	exit(1);
	*/
	subblock_data_ptr->subblock_inf = subblock_inf;
	subblock_data_ptr->num_subblocks_per_block = num_subblocks_per_block;
	subblock_data_ptr->chunk_head_ptr = ch_subblock_head_ptr;
	printf("check_netlist\n");
	check_netlist (subblock_data_ptr, num_driver);
	update_carryouts(subblock_data_ptr);
	free_parse();
}


static void init_parse(int doall) {

/* Allocates and initializes the data structures needed for the parse. */

 int i, j, k,len, nindex, pin_count, pc;
 int *tmp_ptr;
 struct s_hash_iterator hash_iterator;
 struct s_hash *h_ptr;


 if (doall==2) return;

 if (!doall) {  /* Initialization before first (counting) pass */
    num_nets = 0;  
    hash_table = alloc_hash_table ();

#define INITIAL_BLOCK_STORAGE 12000
    temp_block_storage = INITIAL_BLOCK_STORAGE;
    num_subblocks_per_block = my_malloc (INITIAL_BLOCK_STORAGE *
             sizeof(int));

    ch_subblock_bytes_avail = 0;
    ch_subblock_next_avail_mem = NULL;
    ch_subblock_head_ptr = NULL;
 }

/* Allocate memory for second (load) pass */ 

 else {   
    net = (struct s_net *) my_malloc (num_nets*sizeof(struct s_net));
    block = (struct s_block *) my_malloc (num_blocks*
        sizeof(struct s_block));   
    is_global = (boolean *) my_calloc (num_nets, sizeof(boolean));
    num_driver = (int *) my_malloc (num_nets * sizeof(int));
    temp_num_pins = (int *) my_malloc (num_nets * sizeof(int));

	//printf("num_blocks : %d, max subblock per block :%d \n", num_blocks, max_subblocks_per_block);
    allocate_temporary_carry_name_info(num_blocks,max_subblocks_per_block);
 
    for (i=0;i<num_nets;i++) {
       num_driver[i] = 0;
       net[i].num_pins = 0;
    }

/* Allocate block pin connection storage.  Some is wasted for io blocks. *
 * Method used below "chunks" the malloc of a bunch of small things to   *
 * reduce the memory housekeeping overhead of malloc.                    */
    
    pc = MAX(pins_per_clb,pins_per_memory);
    pc = MAX(pins_per_dsp,pc);

	for (i = 1; i< number_module; i++)
	{
		pc = MAX(constraint[i].num_pins, pc);
	}

    tmp_ptr = (int *) my_malloc (
         pc * num_blocks * sizeof(int));
    for (i=0;i<num_blocks;i++) {
       block[i].nets = tmp_ptr + i * pc;
       block[i].carryout_cluster = -1;       
       block[i].carryout_ble = -1;
       block[i].carryin_cluster = -1;       
       block[i].carryin_ble = -1;
       block[i].carry_head = 0;
       block[i].carry_body = 0;
    }

/* I use my_chunk_malloc for some storage locations below.  my_chunk_malloc  *
 * avoids the 12 byte or so overhead incurred by malloc, but since I call it *
 * with a NULL head_ptr, it will not keep around enough information to ever  *
 * free these data arrays.  If you ever have compatibility problems on a     *
 * non-SPARC architecture, just change all the my_chunk_malloc calls to      *
 * my_malloc calls.                                                          */


    hash_iterator = start_hash_table_iterator ();
    h_ptr = get_next_hash (hash_table, &hash_iterator);
    
    while (h_ptr != NULL) 
	{
       nindex = h_ptr->index;
       pin_count = h_ptr->count;

	   /*
       net[nindex].blocks = (int *) my_chunk_malloc(pin_count *
               sizeof(int), NULL, &chunk_bytes_avail, &chunk_next_avail_mem);

       net[nindex].blk_pin = (int *) my_chunk_malloc (pin_count * 
               sizeof(int), NULL, &chunk_bytes_avail, &chunk_next_avail_mem);
		*/
	   
	   net[nindex].blocks = (int *) my_malloc(pin_count * sizeof(int));

       net[nindex].blk_pin = (int *) my_malloc (pin_count * sizeof(int));
/* For avoiding assigning values beyond end of pins array. */

       temp_num_pins[nindex] = pin_count;

       len = strlen (h_ptr->name);

	   /*
       net[nindex].name = (char *) my_chunk_malloc ((len + 1) *
            sizeof(char), NULL, &chunk_bytes_avail, &chunk_next_avail_mem);
			*/
	   net[nindex].name = (char *) my_malloc ((len + 1) * sizeof(char));
       strcpy (net[nindex].name, h_ptr->name);
       h_ptr = get_next_hash (hash_table, &hash_iterator);
    }

/* Allocate storage for subblock info. (what's in each logic block) */

   num_subblocks_per_block = (int *) my_realloc (num_subblocks_per_block,
                  num_blocks * sizeof (int));
   subblock_inf = (t_subblock **) my_malloc (num_blocks * 
                        sizeof(t_subblock *));

   for (i=0;i<num_blocks;i++) 
   {

      if (num_subblocks_per_block[i] == 0) 
         subblock_inf[i] = NULL;
      else {
		  /*
         subblock_inf[i] = (t_subblock *) my_chunk_malloc (
              num_subblocks_per_block[i] * sizeof (t_subblock), 
              &ch_subblock_head_ptr, &ch_subblock_bytes_avail, 
              &ch_subblock_next_avail_mem);
			  */
		  subblock_inf[i] = (t_subblock *) my_malloc (
              num_subblocks_per_block[i] * sizeof (t_subblock));

         for (j=0;j<num_subblocks_per_block[i];j++)  
		 {

			 /*
            subblock_inf[i][j].function = (char *)my_chunk_malloc
                ( sizeof(char)*((1<<subblock_lut_size)+1), &ch_subblock_head_ptr,
                 &ch_subblock_bytes_avail, &ch_subblock_next_avail_mem);
				 */
			 subblock_inf[i][j].function = (char *)my_malloc
                ( sizeof(char)*((1<<subblock_lut_size)+1));

            if (subblock_inf[i][j].function == NULL) {
               printf("error allocating function memory\n"); exit(0);
	    }
             for(k=0;k<(1<<subblock_lut_size);k++) {
               subblock_inf[i][j].function[k] = '0';
	    }
            subblock_inf[i][j].function[1<<subblock_lut_size] = '\0';
/*
			subblock_inf[i][j].inputs = (int *) my_chunk_malloc
                 (subblock_num_ipin * sizeof(int), &ch_subblock_head_ptr, 
                 &ch_subblock_bytes_avail, &ch_subblock_next_avail_mem);

			subblock_inf[i][j].outputs = (int *) my_chunk_malloc
                 (subblock_num_opin * sizeof(int), &ch_subblock_head_ptr, 
                 &ch_subblock_bytes_avail, &ch_subblock_next_avail_mem);
				 */
			subblock_inf[i][j].inputs = (int *) my_malloc
                 (subblock_num_ipin * sizeof(int));

			subblock_inf[i][j].outputs = (int *) my_malloc
                 (subblock_num_opin * sizeof(int));
	 }
      }
   }
 }

/* Initializations for both passes. */

 linenum = 0;
 num_p_inputs = 0;
 num_p_outputs = 0;
 num_clbs = 0;
 num_dsp = 0;
 num_memory = 0;
 num_blocks = 0;
 num_globals = 0;
}


static char *get_tok (char *buf, int doall, FILE *fp_net, char *veb_link_file, int pass) {

/* Figures out which, if any token is at the start of this line and *
 * takes the appropriate action.  It always returns a pointer to    *
 * the next line (I need to do this so I can do some lookahead).    */

 char *ptr; 
//printf("%s (len = %d) \n", buf, strlen(buf));
 ptr = my_strtok(buf,TOKENS,fp_net,buf);
 

 if (ptr == NULL) {                       /* Empty line.  Skip. */
    ptr = my_fgets(buf, BUFSIZE, fp_net);
    return (ptr);
 }

	//add embedded block first 
	if (pass ==0)
	{
		//handle the veb block
		if (strcmp(ptr,".emb") == 0) 
		{
			//printf("Add emb_clb, doall %d \n", doall);
			
			ptr = add_emb (doall, fp_net, buf, veb_link_file);		
			return (ptr);
		}		
	}

	//add veb block first 
	if (pass ==1)
	{
		//handle the veb block
		if (strcmp(ptr,".veb") == 0) 
		{
			printf("Add veb_clb, doall %d \n", doall);
			ptr = add_veb (doall, fp_net, buf, veb_link_file);		
			return (ptr);
		}		
	}
 
if (pass == 2)
{
	if (strcmp(ptr,".clb") == 0) {
		//printf("Add clb, doall %d \n", doall);
	
	ptr = add_clb (doall, fp_net, buf);
	return (ptr);
	}
	
	if (strcmp(ptr,".dsp") == 0) {
	
	ptr = add_dsp (doall, fp_net, buf);
	ptr = my_fgets(buf, BUFSIZE, fp_net);
	return (ptr);
	}
	
	if (strcmp(ptr,".memory") == 0) {
	ptr = add_memory (doall, fp_net, buf);
	ptr = my_fgets(buf, BUFSIZE, fp_net);
	return (ptr);
	}
	
	if (strcmp(ptr,".input") == 0) {
	add_io (doall, INPAD, fp_net, buf);
	ptr = my_fgets(buf, BUFSIZE, fp_net);
	return (ptr);
	}
	
	if (strcmp(ptr,".output") == 0) {
		//printf("Add output \n");
	
	add_io (doall, OUTPAD, fp_net, buf);
	ptr = my_fgets(buf, BUFSIZE, fp_net);
	//printf("%s \n", ptr);
	return (ptr);
	}
	
	if (strcmp(ptr,".global") == 0) {
		//printf("Add global, doall %d \n", doall);
	add_global (doall, fp_net, buf);
	ptr = my_fgets(buf, BUFSIZE, fp_net);
	return (ptr);
	}
}

ptr = my_fgets(buf, BUFSIZE, fp_net);
return(ptr);
 //printf ("Error in get_tok while parsing netlist file.\n");
 //printf ("Line %d starts with an invalid token (%s).\n", linenum, ptr);

 //exit (1);
}


static int get_pin_number (char *ptr) {

/* Returns the pin number to which a subblock pin connects.  This pin number *
 * can be OPEN, one of the clb pins (from 0 to clb_size-1) or a "hidden"     *
 * pin that refers to the output of one of the subblocks within the clb      *
 * (the output of subblock 0 is pin clb_size, the output of subblock         *
 * max_subblocks_per_block is clb_size + max_subblocks_per_block-1).         */

 int val;
 int next_block;
 int next_pin;
 int *temp_ptr;
 char temp_char[2];

 //open pin
 if (strcmp("open",ptr) == 0) 
    return (OPEN);


 //internal connection
 //ble_nextblock_pinnumber
 if (strncmp ("ble_", ptr, 4) == 0) 
{   /* "Hidden" pin (Subblock output) */
	//printf("get pin number\n");
	temp_char[0] = ptr[4];
	temp_char[1] = '\0';
    //val = my_atoi (ptr + 4);
	next_block = (int)atoi(temp_char);

	temp_char[0] = ptr[6];
	temp_char[1] = '\0';
	next_pin =  (int)atoi (temp_char);
	//printf("ptr %s \n", ptr);
//	printf("next block %d , next_pin %d \n", next_block, next_pin);

	/*
	if (val < 0 || val >= pins_per_clb)
	{
		printf("Error in get_pin_number on line %d of netlist file.\n", 
            linenum);
		printf("Pin ble_%d is out of legal range (ble_%d to ble_%d).\n"
            "Aborting.\n\n", val, 0, pins_per_clb - 1);
		exit (1);		
	}
	val += pins_per_clb;
	//printf("subblock name : %s, value: %d \n", ptr, val);
	*/

	if (next_block >= max_subblocks_per_block)
	{
		printf("Error in get_pin_number on line %d of netlist file.\n", linenum);
		printf("Next block %d is out of max subblocks per block \n", next_block);
		exit(1);
	}

	if (next_pin < 0 || next_pin >= pins_per_clb)
	{
		printf("Error in get_pin_number on line %d of netlist file.\n", 
            linenum);
		printf("Pin ble_%d is out of legal range (ble_%d to ble_%d).\n"
            "Aborting.\n\n", val, 0, pins_per_clb - 1);
		exit (1);		
	}

	val = pins_per_clb + next_block *subblock_num_pin +  next_pin;
	return (val);

}

 /* Clb input pin. */

 val = my_atoi (ptr);

 if (val < 0 || val >= pins_per_clb) {
	
    printf("Error in get_pin_number on line %d of netlist file.\n", 
         linenum);
    printf("Pin %d is out of legal range (%d to %d).\nAborting.\n\n",
         val, 0, pins_per_clb - 1);
    exit (1);
 }

 return (val);
}


static void load_subblock_array (int doall, FILE *fp_net,
           char *temp_buf, int num_subblocks, int bnum, int type, char *veb_name) 
{

/* Parses one subblock line and, if doall is 1, loads the proper   *
 * arrays.  Each subblock line is of the format:                   *
 * subblock: <name> <ipin0> <ipin1> .. <ipin[subblock_lut_size-1]> *
 *          <opin> <clockpin>                                      */
 
 int pin, ipin, opin, len, connect_to;
 char *ptr;
 char *carry_cluster_name = NULL;

 ipin = 0;
 opin = 0;
 pin = 0;
 ptr = my_strtok(NULL,TOKENS,fp_net,temp_buf);

 if (ptr == NULL) {
    printf("Error in load_subblock_array on line %d of netlist file.\n",
             linenum);
    printf("Subblock name is missing.\nAborting.\n\n");
    exit (1);
 }
    
/* Load subblock name if this is the load pass. */
 if (doall == 1) {
    
	if (type == 0) 
	{
		//normal clb
		len = strlen (ptr);
		/*
		subblock_inf[bnum][num_subblocks-1].name = my_chunk_malloc ((len+1) *
             sizeof(char), &ch_subblock_head_ptr, &ch_subblock_bytes_avail,
             &ch_subblock_next_avail_mem);
			 */
		subblock_inf[bnum][num_subblocks-1].name = my_malloc ((len+1) *
             sizeof(char));
		strcpy (subblock_inf[bnum][num_subblocks-1].name, ptr);
		
	}
	else if (type == 1)
	{
		//veb clb
		len = strlen (ptr) + 1 + strlen(veb_name);
		/*
		subblock_inf[bnum][num_subblocks-1].name = my_chunk_malloc ((len+1) *
             sizeof(char), &ch_subblock_head_ptr, &ch_subblock_bytes_avail,
             &ch_subblock_next_avail_mem);
			 */
		subblock_inf[bnum][num_subblocks-1].name = my_malloc ((len+1) *
             sizeof(char));
//		strcpy (subblock_inf[bnum][num_subblocks-1].name, ptr);
		strcpy (subblock_inf[bnum][num_subblocks-1].name, veb_name);
		strcat (subblock_inf[bnum][num_subblocks-1].name, "_");
		strcat (subblock_inf[bnum][num_subblocks-1].name, ptr);

	}
 }

 ptr = my_strtok(NULL,TOKENS,fp_net,temp_buf);
 //printf ("First net name : %s \n", ptr);
 //printf ("------------------------------------------------------------------------------------ \n");
	while (ptr != NULL)
	{
		//For each subblock pin, get the net name
		if (doall ==1)
		{
			//check error
			if (ipin > subblock_num_ipin)
			{
				printf("Error in load_subblock_array at line %d of netlist file.\n",
                   		   linenum); 
	            printf("The number of input pin of the subblock (%d) is greater than expected (%d)\n", ipin, subblock_num_ipin);
				exit(1);
			}

			if (opin > subblock_num_opin)
			{
				printf("Error in load_subblock_array at line %d of netlist file.\n",
                   		   linenum); 
	            printf("he number of output pin of the subblock (%d) is greater than expected (%d)\n", opin, subblock_num_opin);
				exit(1);
			}

			if (pin > subblock_num_pin)
			{
				printf("Error in load_subblock_array at line %d of netlist file.\n",
                   		   linenum); 
	            printf("The number of IO pin of the subblock (%d) is greater than expected (%d)\n", pin, subblock_num_pin);
				exit(1);
			}

			switch (pin)
			{												
				case B_CIN: //CIN					
					
					pin ++;
					connect_to = get_pin_number (ptr);
					
					if (connect_to != OPEN)
					{
						if (connect_to < pins_per_clb)
						{
							if (block[bnum].nets[connect_to] == OPEN)
							{
								subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
							}else{						
								//part of carry chain
								//get the carry chain name 																
								subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;								
							}
						}
						else
						{
							//internal connection
							subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;						
						}
					}
					else
					{
						subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
					}
										
					ipin ++;
					

					break;
				case B_COUT: //COUT					
					
					pin ++;					
					connect_to = get_pin_number (ptr);
					
					if (connect_to != OPEN)
					{
						if (connect_to < pins_per_clb)
						{
							if (block[bnum].nets[connect_to] == OPEN)
							{
								subblock_inf[bnum][num_subblocks-1].outputs[opin] = OPEN;
							}else{						
								//part of carry chain							
								subblock_inf[bnum][num_subblocks-1].outputs[opin] = connect_to;								
								
							}
						}
						else
						{
							//internal connection							
							subblock_inf[bnum][num_subblocks-1].outputs[opin] = connect_to;							
						}
					}
					else
					{
						subblock_inf[bnum][num_subblocks-1].outputs[opin] = OPEN;
					}										
					opin ++;
					break;
				case B_SHIFTIN: //SHIFTIN					
					/*
				case 4:	//F4/G4
				case 5:	//F3/G3
				case 6:	//F2/G2
				case 7:	//F1/G1
				*/
				case B_BY:	//BX/BY								
				case B_CE:	//CE
				case B_SR:	//SR
				case B_FXINA: // FXINA
				case B_FXINB: //FXINB
					//input
					
					pin ++;

					connect_to = get_pin_number (ptr);
					//printf("bnum : %d, num_subblocks : %d, ipin : %d,  Connect to : %d \n", bnum, num_subblocks-1, ipin, connect_to);
					
					if (connect_to != OPEN)
					{
						if (connect_to < pins_per_clb)
						{
							if (block[bnum].nets[connect_to] == OPEN)
							{
								subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
							}else{						
								subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;
							}
						}else
						{
							//internal connection
							subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;							
						}
					}
					else
					{
						subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
					}
					
					//subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;
					ipin ++;

					break;

				case B_DY:	//DX/DY
					//special handle for input DX/DY
					//internal connection to output X/Y 

					pin ++;

					connect_to = get_pin_number (ptr);
					if (connect_to != OPEN)
					{
						//need connect to X/Y
						if (subblock_inf[bnum][num_subblocks-1].outputs[4] == OPEN)
						{
							//X/Y is open, need to create non open connection to create a source node
							subblock_inf[bnum][num_subblocks-1].outputs[4] = pins_per_clb + ipin;	//set as internal connection
						}
						subblock_inf[bnum][num_subblocks-1].inputs[ipin] = pins_per_clb + ipin; //set as internal connection
					}
					else
					{
						//open
						subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
					}
					ipin ++;

					break;
								
				case B_CLK:	//CLK
					
					pin ++;

					connect_to = get_pin_number (ptr);
					subblock_inf[bnum][num_subblocks-1].clock = connect_to;										
					subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;	//assume to be open
					ipin ++;

					break;
				case B_SHIFTOUT: //SHIFTOUT					
				case B_BYOUT:	//BXOUT/BYOUT
				case B_YB:	//XB/YB	
				case B_Y:	//X/Y
				case B_YQ:	//XQ/XY
				case B_FX:  //FX
					//output
										
					pin ++;

					connect_to = get_pin_number (ptr);
															
					if (connect_to != OPEN)
					{
						if (connect_to < pins_per_clb)
						{
							if (block[bnum].nets[connect_to] == OPEN)
							{
								subblock_inf[bnum][num_subblocks-1].outputs[opin] = OPEN;
							}else{						
								subblock_inf[bnum][num_subblocks-1].outputs[opin] = connect_to;
							}
						}
						else
						{
							//internal connection
							subblock_inf[bnum][num_subblocks-1].outputs[opin] = connect_to;
						}
					}	
					else
					{
						subblock_inf[bnum][num_subblocks-1].outputs[opin] = OPEN;
					}				
					//printf("load_subblock_array , bnum %d, subblock num %d, opin %d, connect_to %d \n", bnum, num_subblocks-1, opin, subblock_inf[bnum][num_subblocks-1].outputs[opin]);
					
					//subblock_inf[bnum][num_subblocks-1].outputs[ipin] = connect_to;
					opin ++;

					break;

				default:
					//LUT input 
					if ((pin >=  B_LUT_IN)  &&  (pin < B_LUT_IN + subblock_lut_size))
					{
						//input
						
						pin ++;

						connect_to = get_pin_number (ptr);
						//printf("bnum : %d, num_subblocks : %d, ipin : %d,  Connect to : %d \n", bnum, num_subblocks-1, ipin, connect_to);
						
						if (connect_to != OPEN)
						{
							if (connect_to < pins_per_clb)
							{
								if (block[bnum].nets[connect_to] == OPEN)
								{
									subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
								}else{						
									subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;
								}
							}else
							{
								//internal connection
								subblock_inf[bnum][num_subblocks-1].inputs[ipin] = connect_to;							
							}
						}
						else
						{
							subblock_inf[bnum][num_subblocks-1].inputs[ipin] = OPEN;
						}
											
						ipin ++;
					}
					break;
			}			
		}
		ptr = my_strtok(NULL,TOKENS,fp_net,temp_buf);
	}
}


static void set_subblock_count (int bnum, int num_subblocks) {

/* Sets the temporary subblock count for block bnum to num_subblocks. *
 * Properly allocates whatever temporary storage is needed.           */
 
 if (bnum >= temp_block_storage) {
    temp_block_storage *= 2;
    num_subblocks_per_block = (int *) my_realloc 
            (num_subblocks_per_block, temp_block_storage * sizeof (int));
 }
 
 num_subblocks_per_block[bnum] = num_subblocks;
}


/* DSP blocks and memories contain one subblock structure.  This is where
   the activity information will later be stored.  It would make more
   sense to include this as some sort of top-level structure, but it is 
   done this way to minimize the number of changes to VPR required */

static void add_one_empty_subblock(int doall, int bnum)
{
  int len;

  if (doall == 0)
     set_subblock_count (bnum, 1);
  else if (doall ==1)
    assert (num_subblocks_per_block[bnum]==1);

  if (doall == 1) 
  {
    len = (int)strlen(block[bnum].name)+1;
	/*
    subblock_inf[bnum][0].name = my_chunk_malloc ((len) *
                sizeof(char), &ch_subblock_head_ptr, &ch_subblock_bytes_avail,
                &ch_subblock_next_avail_mem);
				*/
	subblock_inf[bnum][0].name = my_malloc ((len) *
                sizeof(char));
    strcpy (subblock_inf[bnum][0].name,block[bnum].name);
  }
}



static char *parse_subblocks (int doall, FILE *fp_net, char *buf,
               int bnum, int type, char *veb_name) {

/* Loads the subblock arrays with the proper values. */

 char temp_buf[BUFSIZE], *ptr;
 int num_subblocks;

 num_subblocks = 0;

 while (1) 
 {
    ptr = my_fgets (temp_buf, BUFSIZE, fp_net);

    if (ptr == NULL)   
       break;          /* EOF */

  /* Save line in case it's not a sublock */
    strcpy (buf, temp_buf);

    ptr = my_strtok(temp_buf,TOKENS,fp_net,temp_buf);
    if (ptr == NULL) 
       continue;       /* Blank or comment line.  Skip. */

    if (strcmp("subblock:", ptr) == 0) 
	{
       num_subblocks++; 
	   //load the subblock pin
       load_subblock_array (doall, fp_net, temp_buf, num_subblocks,
               bnum, type, veb_name);
    }
    else 
	{
       break;  /* Subblock list has ended.  Buf contains next line. */
    }
 }   /* End infinite while */

 if (num_subblocks < 1 || num_subblocks > max_subblocks_per_block) 
 {
    printf("Error in parse_subblocks on line %d of netlist file.\n",
         linenum);
    printf("Block #%d has %d subblocks.  Out of range.\n",
         bnum, num_subblocks);
    printf("Aborting.\n\n");
    exit (1);
 }

 if (doall == 0)
    set_subblock_count (bnum, num_subblocks);
 else if (doall == 1)
    assert (num_subblocks == num_subblocks_per_block[bnum]);

 return (ptr);
}


static char *add_clb (int doall, FILE *fp_net, char *buf) {

/* Adds the clb (.clb) currently being parsed to the block array.  Adds *
 * its pins to the nets data structure by calling add_net.  If doall is *
 * zero this is a counting pass; if it is 1 this is the final (loading) *
 * pass.                                                                */

 char *ptr;
 int pin_index, iclass, inet;
 enum e_pin_type type;
 char *temp_ptr;

 num_blocks++;
 parse_name_and_pinlist (doall, fp_net, buf,0);
 num_clbs++;
 
 if (doall==1) 
 {
    block[num_blocks - 1].type = CLB;
	block[num_blocks - 1].module_index = 0;	//it is belong to top module
 }

 pin_index = -1;
 ptr = my_strtok(NULL,TOKENS,fp_net,buf);

 //get the pin information of the Slice
 while (ptr != NULL) 
 {
    pin_index++;
    if (pin_index >= pins_per_clb) 
	{
       printf("Error in add_clb on line %d of netlist file.\n",linenum);
       printf("Too many pins on this clb %d.  Expected %d.\n",pin_index, pins_per_clb);
       exit (1);
    }
    
    iclass = clb_pin_class[pin_index];
    type = class_inf[iclass].type;      /* DRIVER or RECEIVER */

    if (strcmp(ptr,"open") != 0) {     /* Pin is connected. */
		if (type == DRIVER)
		{
			//driver follow top net name
    	   	inet = add_net (ptr, type, num_blocks-1, pin_index, doall);		
		}
		else
		{
			//receiver follow black box net name
    			temp_ptr =  veb_convert_veb_net (ptr); //convert veb internal to top modulee net name
			//printf("original name : %s, converted name : %s \n", ptr, temp_ptr);
    	   	inet = add_net (temp_ptr, type, num_blocks-1, pin_index, doall);		
		}
       if (doall==1)                      /* Loading pass only */
          block[num_blocks - 1].nets[pin_index] = inet; 
    }
    else {                             /* Pin is unconnected (open) */
       if (doall==1) 
          block[num_blocks - 1].nets[pin_index] = OPEN;
    }

    ptr = my_strtok(NULL,TOKENS,fp_net,buf);
 }

 if (pin_index != pins_per_clb - 1) {
    printf("Error in add_clb on line %d of netlist file.\n",linenum);
    printf("Expected %d pins on clb, got %d.\n", pins_per_clb, pin_index + 1);
    exit (1);
 }

 //get the pin information of the each subblock
 ptr = parse_subblocks (doall, fp_net, buf, num_blocks-1, 0, NULL);
 
 //printf("num_subblocks_per_block[%d] = %d \n", num_blocks - 1, num_subblocks_per_block[num_blocks - 1]);
 
 return (ptr);
}



/* 
	Added by Yu Chi Wai
*/
/*****************************************************************************************************************************************************************/
static char *add_emb (int doall, FILE *fp_net, char *buf, char *emb_link_file) {

/* Get the embedded block net file from the link file. Get the pin list  *
 * from top module's net file. Also get the pin list from the *
 * emb net file. Compare the number of pin, exit the program  *
 * if there is error										  */

 char buffer[BUFSIZE];
 char *pinlist;
 char *emb_name;

 char *ptr;
 int pin_index, inet;
 enum e_pin_type type;
 char * temp_ptr;
 
 int len;
 int emb_linenum;
 int i;

 int module_index;
 

 
 //get the veb name

 ptr = my_strtok(NULL, TOKENS, fp_net,buf);
 len = strlen(ptr);

 
 emb_name = malloc((len + 1) * sizeof(char));
 strcpy(emb_name, ptr);

 printf("emb name : %s\n", emb_name);
 


 module_index = get_module_index(emb_name);

 printf ("EMB (%s), module index %d, number of module %d\n", emb_name, module_index,number_module);
 
if (module_index == -1)
{
	printf("Error getting emb index\n");
	exit(1);
}


 num_blocks++;
 num_emb ++ ;
 
 ptr = my_fgets(buf, BUFSIZE, fp_net);

 //get the net name
 ptr = my_strtok(buf, TOKENS, fp_net,buf);		//pinlist
	
 ptr = my_strtok(NULL, TOKENS, fp_net,buf); //signal name
 pin_index = 0;
 while (ptr != NULL)
 {	
		if (doall==1) 
		{
			if (pin_index >= constraint[module_index].num_pins) 
			{
				printf("Error in add_emb on line %d of netlist file.\n",linenum);
				printf("Too many pins on this emb, num pin %d, pins per emb %d\n", pin_index, constraint[module_index].num_pins);
				exit (1); 
			}
		}

		if (constraint[module_index].is_outpin[pin_index]) 
		{
			type = DRIVER;
		} else 
		{
			type = RECEIVER;
		}

		if (strcmp(ptr,"open") != 0 && strcmp(ptr,"OPEN") != 0) 
		{     /* Pin is connected. */
	   		
    	   	inet = add_net (ptr, type, num_blocks-1, pin_index, doall);
			
			if (doall==1)                      /* Loading pass only */
				block[num_blocks - 1].nets[pin_index] = inet; 

			//printf("pin_index %d , %s  %d\n", pin_index, ptr, type);
		}
		else 
		{                             /* Pin is unconnected (open) */
			if (doall==1) 
				block[num_blocks - 1].nets[pin_index] = OPEN;
		}
		pin_index++;
		ptr = my_strtok(NULL,TOKENS,fp_net,buf);	 
 }

 if (doall==1) 
 {
    block[num_blocks - 1].type = EMB;
	block[num_blocks - 1].module_index = module_index;
	len = strlen (emb_name);
	/*
    block[num_blocks-1].name = my_chunk_malloc ((len + 1) * sizeof(char),
                  NULL, &chunk_bytes_avail, &chunk_next_avail_mem);
				  */
	block[num_blocks-1].name = my_malloc ((len + 1) * sizeof(char));

	block[num_blocks-1].x = constraint[module_index].lx;
	block[num_blocks-1].y = constraint[module_index].ly;

    strcpy (block[num_blocks-1].name, emb_name);

 }
 
 ptr = my_fgets(buf, BUFSIZE, fp_net);

 //close the emb net file
 //add_one_empty_subblock(doall, num_blocks-1);

 return (ptr);
}



static char *add_veb (int doall, FILE *fp_net, char *buf, char *veb_link_file) {

/* Get the veb net file from the link file. Get the pin list  *
 * from top module's net file. Also get the pin list from the *
 * veb net file. Compare the number of pin, exit the program  *
 * if there is error										  */

 char *ptr;
 int pin_index, iclass, inet;
 enum e_pin_type type;

 //veb varialbe 
 char veb_link_buf[BUFSIZE], *veb_link_ptr;
 char veb_buf[BUFSIZE], *veb_ptr;
 char buffer[BUFSIZE];
 char *netname;
 char *pinlist;
 char *veb_name;
 char *temp_ptr;
 char *veb_net_name;
 
 FILE *fp_veb;
 FILE *fp_veb_link;
 FILE *fp_veb_io;
 char veb_io_name[BUFSIZE];  
 char *io_ptr;
 
 int len;
 int veb_linenum;
 int i;

 int module_index;
int temp_num_name_convert; 
int temp_num;
 

 
 //get the veb name
 ptr = my_strtok(NULL, TOKENS, fp_net,buf);
 len = strlen(ptr);

 
 veb_name = malloc((len + 1) * sizeof(char));
 strcpy(veb_name, ptr);

 printf("Veb name : %s\n", veb_name);
 

 //check the veb link file empty or not
 if (strlen(veb_link_file) == 0) 
 {
	 printf("Error in add_veb on line %d of netlist file.\n", linenum);
	 printf("Missing link file for veb\n");
	 exit(1);
 }

 //open the link file to get the net file for veb
 fp_veb_link = my_fopen(veb_link_file, "r", 0);
 // get a line in the link file
 veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);
 
 veb_net_name = NULL;

 //getting the net file name of the veb from the link file
 while (veb_link_ptr != NULL)
 {
	temp_ptr = my_strtok(veb_link_buf, TOKENS, fp_veb_link, veb_link_buf);

	if ((temp_ptr != NULL) && (strcmp(temp_ptr,veb_name) == 0))
	{
		//match the veb name
		temp_ptr = my_strtok(NULL, TOKENS, fp_veb_link, veb_link_buf);
		printf ("VEB file name of veb %s : %s\n", veb_name, temp_ptr);
		veb_net_name = malloc (strlen(temp_ptr) + 1 * sizeof(char));
		strcpy(veb_net_name, temp_ptr);
		break;
	}
	veb_link_ptr = my_fgets(veb_link_buf, BUFSIZE, fp_veb_link);
 }

 if (veb_net_name == NULL)
 {
	//can't find any match veb name in the link file
	 printf("Error in reading link file %s\n", veb_link_file);
	 printf("Can't find the veb net file for the VEB %s \n", veb_name);
	 fclose(fp_veb_link);
	 exit(1);
 }

 //close the link file
 fclose(fp_veb_link);
 

//open the net file for veb
fp_veb = my_fopen(veb_net_name, "r", 0);

strcpy(veb_io_name, veb_name);
strcat(veb_io_name, ".io");

fp_veb_io = my_fopen(veb_io_name, "r", 0);
printf("Open file %s \n", veb_io_name);
if (fp_veb_io == NULL)
{
	printf("Error openning file : %s \n", veb_io_name);
	exit(1);
}
 


 veb_ptr = my_fgets(veb_buf, BUFSIZE, fp_veb); 
 
 
 pinlist = my_fgets(buffer, BUFSIZE, fp_net);
 netname = strtok(pinlist, " \t\n");
 netname = strtok(NULL, " \t\n");

temp_num_name_convert= 0;

 while(netname != NULL)
 {    	 	
 	temp_num_name_convert++;
	temp_num = temp_num_name_convert + num_name_convert;
	name_convert = (struct s_name_convert *)realloc(name_convert, temp_num * (sizeof(struct s_name_convert))); 
	name_convert[temp_num - 1].convert_name = malloc(strlen(netname) + 1);
	strcpy (name_convert[temp_num - 1].convert_name, netname);
	netname = strtok(NULL, " \t\n");				
 }
 
 
 for (i=0;i< temp_num_name_convert; i++)
 {
 	if (my_fgets(veb_io_name, BUFSIZE, fp_veb_io) == NULL)
	{
		printf("Error in IO convertion reading, not enough veb IO \n"); 
	}
	io_ptr = strtok(veb_io_name, " \t\n");
	temp_num = i + num_name_convert;
	name_convert[temp_num].veb_io_name = malloc(strlen(veb_name) + 1+ strlen(io_ptr) + 1);
	strcpy (name_convert[temp_num].veb_io_name, veb_name);
	strcat (name_convert[temp_num].veb_io_name, "_");
	strcat (name_convert[temp_num].veb_io_name, io_ptr);
	temp_ptr = strtok(NULL, " \t\n");
	name_convert[temp_num].io = atoi(temp_ptr);
 }
 
 num_name_convert = num_name_convert + temp_num_name_convert;
 preprocess_veb_io(); 
// printf("Number of convert : %d \n", num_name_convert);
 
 
 
	
 module_index = get_module_index(veb_name);
if (module_index == -1)
{
	printf("Error getting veb index\n");
	exit(1);
}
//get clb in VEB net file
 while(veb_ptr != NULL)
 {
 //	printf("%s \n", veb_buf);
	veb_ptr = my_strtok(veb_buf, TOKENS, fp_veb, veb_buf);	
	if (veb_ptr != NULL)
	{
		/* Handling clb of VEB */
		if (strcmp(veb_ptr,".clb") == 0) 
		{
			veb_ptr = add_veb_clb (doall, fp_veb, veb_buf, veb_name, module_index, name_convert, num_name_convert);
			
		}
		else
		{
			veb_ptr = my_fgets(veb_buf, BUFSIZE, fp_veb); 
		}
	}
	else
	{
		veb_ptr = my_fgets(veb_buf, BUFSIZE, fp_veb); 
	}
	
 }

 
 ptr = my_fgets(buf, BUFSIZE, fp_net);

 //close the veb net file
 fclose(fp_veb_io);
 fclose(fp_veb);
 


 return (ptr);
}


static char *add_veb_clb (int doall, FILE *fp_veb, char *buf, char *veb_name, int module_number, struct s_name_convert *name_convert, int num_name_convert) {

/* Adds the clb (.clb) in VEB net file currently being parsed to the block array.  Adds *
 * its pins to the nets data structure by calling add_veb.  If doall is					*
 * zero this is a counting pass; if it is 1 this is the final (loading)					*
 * pass. Return 1 if fail to add the clb												*/

 char *ptr;
 int pin_index, iclass, inet;
 enum e_pin_type type;
 char net_name[BUFSIZE];
 int i;
 
 char *temp_ptr;
 


 //printf("add_veb_clb\n");

 num_blocks++;
 if (veb_parse_name_and_pinlist (doall, fp_veb, buf,veb_name) == 1)
 {
	 exit (1);
 }
 num_clbs++;

 if (doall==1) 
 {
    block[num_blocks - 1].type = CLB;
	block[num_blocks - 1].module_index = module_number;
 }

 pin_index = -1;
 ptr = my_strtok(NULL,TOKENS,fp_veb,buf);

 while (ptr != NULL) {
    pin_index++;
    if (pin_index >= pins_per_clb) 
    {
       printf("Error in add_veb_clb of veb netlist file.\n");
       printf("Too many pins on this clb.  Expected %d.\n",pins_per_clb);
       exit (1);
    }
    
    iclass = clb_pin_class[pin_index];
    type = class_inf[iclass].type;      /* DRIVER or RECEIVER */

    
    
    if (strcmp(ptr,"open") != 0) {     /* Pin is connected. */
		if (type == DRIVER)
		{
			//driver follow own name
	    	strcpy(net_name, veb_name);
			strcat(net_name, "_");
    		strcat(net_name, ptr);
    		inet = add_net (net_name, type, num_blocks-1, pin_index, doall);
			//printf("Driver input name :%s , output name: %s \n", net_name, net_name);

			if (doall==1)                      // Loading pass only 
				block[num_blocks - 1].nets[pin_index] = inet;     	

		}
		else
		{
			//driver follow top name
			//receiver follow black box net name
	    	strcpy(net_name, veb_name);
			strcat(net_name, "_");
    		strcat(net_name, ptr);
    		temp_ptr =  veb_convert_top_net(net_name); //convert veb internal to top modulee net name

//			printf("Receiver input name :%s , output name: %s \n", net_name , temp_ptr);
			if (strcmp(temp_ptr, "OPEN") != 0)
			{
    		   	inet = add_net (temp_ptr, type, num_blocks-1, pin_index, doall);	
				if (doall==1)                      // Loading pass only 
					block[num_blocks - 1].nets[pin_index] = inet; 
			}	
			else
			{
				if (doall==1) 
        	  		block[num_blocks - 1].nets[pin_index] = OPEN;

			}
		}
    }
    else 
    {                             /* Pin is unconnected (open) */
       if (doall==1) 
          block[num_blocks - 1].nets[pin_index] = OPEN;
    }

    ptr = my_strtok(NULL,TOKENS,fp_veb,buf);
 }

 if (pin_index != pins_per_clb - 1) {
    printf("Error in add_veb_clb of veb netlist file.\n");
    printf("Expected %d pins on clb, got %d.\n", pins_per_clb, pin_index + 1);
    exit (1);
 }

 ptr = parse_subblocks (doall, fp_veb, buf, num_blocks-1, 1, veb_name);
 return (ptr);
}

static char * veb_convert_veb_net (char * in_string)
{
	//convert the  top module net to veb net name to ensure netname consistent
	int i;
	int find_name;
	
	find_name = -1;
	for (i = 0; i< num_name_convert; i++)
	{
		if (strcmp(in_string, name_convert[i].convert_name) == 0)
		{
			if (strcpy (name_convert[i].convert_name, "OPEN") != 0)
			{
				if (name_convert[i].io == 0)		//change name if output of veb 
					return (name_convert[i].veb_io_name);
			}
		}
	/*
		if (strcmp(in_string, name_convert[i].veb_io_name) == 0)
		{
			if (strcmp(name_convert[i].convert_name, "OPEN") != 0)
			{
				return (name_convert[i].convert_name);
			}
			else
			{
				find_name = i;
			}
		}	
		*/
	}
	
	//can't find the net name , return the origonal string
	return (in_string);
	
	/*
	if(find_name != -1)
	{
		//the net name in top module is OPEN
		return (name_convert[find_name].convert_name);
	}
	else
	{
		//can't find the net name , return the origonal string
		return (in_string);
	}
	*/
}

static char * veb_convert_top_net (char * in_string)
{
	//convert the veb net name top module net to ensure netname consistent
	int i;
	int find_name;
	
	find_name = -1;
	for (i = 0; i< num_name_convert; i++)
	{
		if (strcmp(in_string, name_convert[i].veb_io_name) == 0)
		{
			if (strcmp(name_convert[i].convert_name, "OPEN") != 0)
			{
				if (name_convert[i].io == 1) //change name when input to VEB
					return (name_convert[i].convert_name);
			}
/*
			else
			{
				find_name = i;
			}
*/
		}	
	}
	
	//can't find the net name , return the origonal string
	return (in_string);

/*	
	if(find_name != -1)
	{
		//the net name in top module is OPEN
		return (name_convert[find_name].convert_name);
	}
	else
	{
		//can't find the net name , return the origonal string
		return (in_string);
	}
*/
}

static void preprocess_veb_io ()
{
	//rename the net from VEB to VEB itself
	int i;
	int net_count;
	int j;
	
	for (i = 0; i< num_name_convert; i++)
	{
		if (name_convert[i].io ==1 && strcmp(name_convert[i].convert_name, "OPEN")!= 0)
		{
			net_count = 0;
			//for receiver in veb only
			for (j = 0 ; j < num_name_convert; j++)
			{
				if (name_convert[j].io ==0)
				{
					//check against output of VEB
					if (strcmp(name_convert[j].convert_name, name_convert[i].convert_name) == 0)
					{
						net_count ++;
						name_convert[i].convert_name = realloc(name_convert[i].convert_name, strlen(name_convert[j].veb_io_name) + 1);
						strcpy (name_convert[i].convert_name, name_convert[j].veb_io_name);
						break;
					}
				}
			}
		}
	}
}



static int veb_parse_name_and_pinlist (int doall, FILE *fp_veb, char *buf, char *veb_name) {

/* This routine does the first part of the parsing of a block.  It is *
 * called whenever any type of block (.clb, .input or .output) is to  *
 * be parsed.  It increments the block count (num_blocks), and checks *
 * that the block has a name.  If doall is 1, this is the loading     *
 * pass and it copies the name to the block data structure.  Finally  *
 * it checks that the pinlist: keyword exists.  On return, my_strtok  *
 * is set so that the next call will get the first net connected to   *
 * this block.                                                        *
 * return 1 if fail to parse name and pinlist						  */

 char *ptr;
 int len;
 
/* Get block name. */

 ptr = my_strtok(NULL,TOKENS,fp_veb,buf);
 if (ptr == NULL) {
    printf("Error in veb_parse_name_and_pinlist on line %d of netlist file.\n",
       linenum);
    printf(".clb, .dsp, .memory, .input or .output line has no associated name.\n");
    return (1);
 }
 
 if (doall == 1) {    /* Second (loading) pass, store block name and adding veb name before the block name*/
    len = strlen (ptr) + strlen(veb_name) + 1;
	/*
    block[num_blocks-1].name = my_chunk_malloc ((len + 1) * sizeof(char),
                  NULL, &chunk_bytes_avail, &chunk_next_avail_mem);
				  */
	block[num_blocks-1].name = my_malloc ((len + 1) * sizeof(char));

	strcpy (block[num_blocks-1].name, veb_name);
	strcat (block[num_blocks-1].name, "_");
	strcat (block[num_blocks-1].name, ptr);
//    strcpy (block[num_blocks-1].name, ptr);
 }
 
 ptr = my_strtok (NULL,TOKENS,fp_veb,buf);

/* Now get pinlist from the next line.  Note that a NULL return value *
 * from my_gets means EOF, while a NULL return from my_strtok just    *
 * means we had a blank or comment line.                              */
 
 do {
    ptr = my_fgets (buf, BUFSIZE, fp_veb);
    if (ptr == NULL) {
       printf("Error in veb_parse_name_and_pinlist on line %d of netlist file.\n",
          linenum);
       printf("Missing pinlist: keyword.\n");
       return (1);
    }

    ptr = my_strtok(buf,TOKENS,fp_veb,buf);
 } while (ptr == NULL);
  block[num_blocks - 1].type = MEMORY;
 if (strcmp (ptr, "pinlist:") != 0) {
    printf("Error in veb_parse_name_and_pinlist on line %d of netlist file.\n",
       linenum);
    printf("Expected pinlist: keyword, got %s.\n",ptr);
    return (1);
 }

 return(0);
}



/*****************************************************************************************************************************************************************/


static char *add_dsp (int doall, FILE *fp_net, char *buf) {

/* Adds the dsp (.dsp) currently being parsed to the block array.  Adds *
 * its pins to the nets data structure by calling add_net.  If doall is *
 * zero this is a counting pass; if it is 1 this is the final (loading) *
 * pass.                                                                */

 char *ptr;
 int pin_index, inet;
 enum e_pin_type type;
 char * temp_ptr;

 if (dsp_inf->rep == 0) {
    printf("DSP block in netlist, but no DSP block specified in architecture\n");
    exit(0);
 }


 num_blocks++;
 parse_name_and_pinlist (doall, fp_net, buf,1);
 num_dsp++;


 if (doall==1) 
    block[num_blocks - 1].type = DSP;

 pin_index = -1;
 ptr = my_strtok(NULL,TOKENS,fp_net,buf);

	while (ptr != NULL) 
	{
		pin_index++;
		if (doall==1) 
		{
			if (pin_index >= pins_per_dsp) 
			{
				printf("Error in add_dsp on line %d of netlist file.\n",linenum);
				printf("Too many pins on this dsp\n");
				exit (1); 
			}
		}

		if (dsp_is_output_pin[pin_index]) 
		{
			type = DRIVER;
		} else 
		{
			type = RECEIVER;
		}

		if (strcmp(ptr,"OPEN") != 0) 
		{     /* Pin is connected. */
	   		if (type == DRIVER)
			{
				//follow top module name
    	   		inet = add_net (ptr, type, num_blocks-1, pin_index, doall);

			}
			else
			{
				//follow veb name
				temp_ptr =  veb_convert_veb_net (ptr); //convert veb internal to top modulee net name
				inet = add_net (temp_ptr, type, num_blocks-1, pin_index, doall);
			}

			if (doall==1)                      /* Loading pass only */
				block[num_blocks - 1].nets[pin_index] = inet; 
		}
		else 
		{                             /* Pin is unconnected (open) */
			if (doall==1) 
				block[num_blocks - 1].nets[pin_index] = OPEN;
		}
		ptr = my_strtok(NULL,TOKENS,fp_net,buf);
	}

	if (pin_index+1 != pins_per_dsp) {
		printf("Error in add_dsp on line %d of netlist file.  Wrong number of pins.\n",linenum);
		printf("Found %d, but architecture file told me to expect %d\n",
			    pin_index+1, pins_per_dsp);
		exit (1);
	}

	add_one_empty_subblock(doall, num_blocks-1);

	return (ptr);
}



static char *add_memory (int doall, FILE *fp_net, char *buf) {

/* Adds the memory (.memory) currently being parsed to the block array.  Adds *
 * its pins to the nets data structure by calling add_net.  If doall is *
 * zero this is a counting pass; if it is 1 this is the final (loading) *
 * pass.                                                                */

 char *ptr;
 int pin_index, inet;
 enum e_pin_type type;
 char *temp_ptr;

 if (memory_inf->rep == 0) {
    printf("MEMORY block in netlist, but no MEMORY block specified in architecture\n");
    exit(0);
 }


 num_blocks++;
 parse_name_and_pinlist (doall, fp_net, buf,1);
 num_memory++;


 if (doall==1) 
    block[num_blocks - 1].type = MEMORY;

 pin_index = -1;
 ptr = my_strtok(NULL,TOKENS,fp_net,buf);

 while (ptr != NULL) {
    pin_index++;
    if (doall==1) {
      if (pin_index >= pins_per_memory) {
       printf("Error in add_memory on line %d of netlist file.\n",linenum);
       printf("Too many pins on this memory\n");
       exit (1); 
      }
    }
    if (memory_is_output_pin[pin_index]) {
       type = DRIVER;
    } else {
       type = RECEIVER;
    }

    if (strcmp(ptr,"open") != 0) {     /* Pin is connected. */
    	if (type == DRIVER)
		{
			//follow top module name
    	   inet = add_net (ptr, type, num_blocks-1, pin_index, doall);

		}
		else
		{
			//follow veb name
	       temp_ptr =  veb_convert_veb_net (ptr); //convert veb internal to top modulee net name
    	   inet = add_net (temp_ptr, type, num_blocks-1, pin_index, doall);

		}
	
       if (doall==1)                      /* Loading pass only */
          block[num_blocks - 1].nets[pin_index] = inet; 
    }
    else {                             /* Pin is unconnected (open) */
       if (doall==1) 
          block[num_blocks - 1].nets[pin_index] = OPEN;
    }
    ptr = my_strtok(NULL,TOKENS,fp_net,buf);
 }

 if (pin_index+1 != pins_per_memory) {
    printf("Error in add_memory on line %d of netlist file.  Wrong number of pins.\n",linenum);
    printf("Found %d, but architecture file told me to expect %d\n",
            pin_index+1, pins_per_memory);

    exit (1);
 }
 add_one_empty_subblock(doall, num_blocks-1);

 return (ptr);
}

/* Modified by Yu Chi Wai*/
static void add_io (int doall, int block_type, FILE *fp_net, char *buf) {

/* Adds the INPAD or OUTPAD (specified by block_type)  currently being  *
 * parsed to the block array.  Adds its pin to the nets data structure  *
 * by calling add_net.  If doall is zero this is a counting pass; if it *
 * is 1 this is the final (loading) pass.                               */

 char *ptr;
 int inet, pin_index, i;
 enum e_pin_type type;
 char *temp_ptr;
 
 num_blocks++;
 if (doall == 0)
 	set_subblock_count (num_blocks-1, 0);    /* No subblocks for IO */
	
parse_name_and_pinlist (doall, fp_net, buf, 0); 

 if (block_type == INPAD) {
    num_p_inputs++;
    type = DRIVER;
 }
 else {
    num_p_outputs++;
    type = RECEIVER;
 }
 
 if (doall==1)
    block[num_blocks - 1].type = block_type;
 
 pin_index = -1;
 ptr = my_strtok(NULL,TOKENS,fp_net,buf);

	/* The OUTPAD of virtex2 has many input pins, O1, O2, OCE OCLK1, SR, REV , OCLK2 , need special handle*/ 
 if (block_type == INPAD)
 {
	 while (ptr != NULL) {
    	pin_index++;
		    if (pin_index >= 1) {
		       printf("Error in add_io on line %d of netlist file.\n",linenum);
		       printf("Too many pins on this io.  Expected 1.\n");
		       exit (1);
	    }
 
    	if (strcmp(ptr,"open") == 0) {     /* Pin unconnected. */
	       printf("Error in add_io, line %d of netlist file.\n",linenum);
    	   printf("Inputs and Outputs cannot have open pins.\n");
	       exit (1);
	    }
   
	/* Note the dummy pin number for IO pins.  Change this if necessary. I set *
	 * them to OPEN because I want the code to crash if I try to look up the   *
	 * class of an I/O pin (since I/O pins don't have classes).                */
    	inet = add_net (ptr, type, num_blocks-1, OPEN, doall);
 
	    if (doall==1)                      /* Loading pass only */
    	   block[num_blocks - 1].nets[pin_index] = inet;

	    ptr = my_strtok(NULL,TOKENS,fp_net,buf);
	 }
 
	 if (pin_index != 0) {
    	printf("Error in add_io on line %d of netlist file.\n",linenum);
	    printf("Expected 1 pin on pad, got %d.\n", pin_index + 1);
    	exit (1);
	 }
 
	 if (doall==1) {
    	for (i=1;i<pins_per_clb;i++) 
	       block[num_blocks-1].nets[i] = OPEN;
	 }
 }
 else
 {
	//OUTPAD
	 while (ptr != NULL) 
	 {
    	pin_index++;
 
	/* Note the dummy pin number for IO pins.  Change this if necessary. I set *
	 * them to OPEN because I want the code to crash if I try to look up the   *
	 * class of an I/O pin (since I/O pins don't have classes).                */
	 	//printf("%s\n", ptr);
	   	if (strcmp(ptr,"open") != 0) {     /* Pin unconnected. */

			temp_ptr =  veb_convert_veb_net(ptr); //convert top modulee net name to veb net name
//			printf("OUTPAD add net %s \n", temp_ptr);
    		inet = add_net (temp_ptr, type, num_blocks-1, pin_index, doall);
		    if (doall==1)                      /* Loading pass only */
			{		
    		   block[num_blocks - 1].nets[pin_index] = inet;
	//		printf("block name : %s , pin_index %d\n", block[num_blocks -1].name, pin_index);
			}
	    }
		else
		{
		    if (doall==1)                      /* Loading pass only */
			{	
			    block[num_blocks - 1].nets[pin_index] = OPEN;
			}
		}
	    ptr = my_strtok(NULL,TOKENS,fp_net,buf);
	 }
 
 	if (doall==1) 
	{
    		for (i=pin_index ;i<pins_per_clb;i++) 
		       block[num_blocks-1].nets[i] = OPEN;
//			 printf("Output pad, number of net %d \n",pin_index);

	}
 }
}

static void parse_name_and_pinlist (int doall, FILE *fp_net, char *buf, int dsp_or_mem) {

/* This routine does the first part of the parsing of a block.  It is *
 * called whenever any type of block (.clb, .input or .output) is to  *
 * be parsed.  It increments the block count (num_blocks), and checks *
 * that the block has a name.  If doall is 1, this is the loading     *
 * pass and it copies the name to the block data structure.  Finally  *
 * it checks that the pinlist: keyword exists.  On return, my_strtok  *
 * is set so that the next call will get the first net connected to   *
 * this block.                                                        */

 char *ptr;
 int len;
 
/* Get block name. */
 
//printf("buf : %s \n", buf);
 ptr = my_strtok(NULL,TOKENS,fp_net,buf);


 if (ptr == NULL) {
    printf("Error in parse_name_and_pinlist on line %d of netlist file.\n",
       linenum);
    printf(".clb, .dsp, .memory, .input or .output line has no associated name.\n");
    exit (1);
 }
 
 if (doall == 1) {    /* Second (loading) pass, store block name */
    len = strlen (ptr);
	/*
    block[num_blocks-1].name = my_chunk_malloc ((len + 1) * sizeof(char),
                  NULL, &chunk_bytes_avail, &chunk_next_avail_mem);
				  */
	block[num_blocks-1].name = my_malloc ((len + 1) * sizeof(char));

    strcpy (block[num_blocks-1].name, ptr);

 }
 //printf("block name : %s \n", ptr);
 
 ptr = my_strtok (NULL,TOKENS,fp_net,buf);
/* Now get pinlist from the next line.  Note that a NULL return value *
 * from my_gets means EOF, while a NULL return from my_strtok just    *
 * means we had a blank or comment line.                              */
 
 do 
 {
    ptr = my_fgets (buf, BUFSIZE, fp_net);
	//printf ("%s \n" ,ptr);
    if (ptr == NULL) {
       printf("Error in parse_name_and_pinlist on line %d of netlist file.\n",
          linenum);
       printf("Missing pinlist: keyword.\n");
       exit (1);
    }

    ptr = my_strtok(buf,TOKENS,fp_net,buf);
 } while (ptr == NULL);
 
 if (strcmp (ptr, "pinlist:") != 0) 
 {
    printf("Error in parse_name_and_pinlist on line %d of netlist file.\n",
       linenum);
    printf("Expected pinlist: keyword, got %s.\n",ptr);
    exit (1);
 }
 //printf("%s \n" , ptr);
}

static void add_global (int doall, FILE *fp_net, char *buf) {

/* Doall is 0 for the first (counting) pass and 1 for the second           *
 * (loading) pass.  fp_net is a pointer to the netlist file.  This         *
 * routine sets the proper entry(ies) in is_global to TRUE during the      *
 * loading pass.  The routine does nothing during the counting pass. If    *
 * is_global = TRUE for a net, it will not be considered in the placement  *
 * cost function, nor will it be routed.  This is useful for global        *
 * signals like clocks that generally have dedicated routing in FPGAs.     */

 char *ptr;
 struct s_hash *h_ptr;
 int nindex;

/* Do nothing if this is the counting pass. */

 if (doall == 0) 
    return; 

 ptr = my_strtok(NULL,TOKENS,fp_net,buf);

 while (ptr != NULL) {     /* For each .global signal */
    num_globals++;

    h_ptr = get_hash_entry (hash_table, ptr);

    if (h_ptr == NULL) {        /* Net was not found in list! */
       printf("Error in add_global on netlist file line %d.\n",linenum);
       printf("Global signal %s does not exist.\n",ptr);
       exit(1);
    }
   
    nindex = h_ptr->index;
    is_global[nindex] = TRUE;    /* Flagged as global net */

    ptr = my_strtok(NULL,TOKENS,fp_net,buf);
 }
}


static int add_net (char *ptr, enum e_pin_type type, int bnum, int blk_pnum, 
          int doall) {   

/* This routine is given a net name in *ptr, either DRIVER or RECEIVER *
 * specifying whether the block number given by bnum is driving this   *
 * net or in the fan-out and doall, which is 0 for the counting pass   *
 * and 1 for the loading pass.  It updates the net data structure and  *
 * returns the net number so the calling routine can update the block  *
 * data structure.                                                     */

 struct s_hash *h_ptr;
 int j, nindex;

 if (doall == 0) {             /* Counting pass only */
    h_ptr = insert_in_hash_table (hash_table, ptr, num_nets);
    nindex = h_ptr->index;

    if (nindex == num_nets)    /* Net was not in the hash table */
       num_nets++;

    return (nindex);
 }

 else if (doall == 1) {                        /* Load pass */

    h_ptr = get_hash_entry (hash_table, ptr);
    nindex = h_ptr->index;

    if (h_ptr == NULL) {
       printf("Error in add_net:  the second (load) pass found could not\n"); 
       printf("find net %s in the symbol table.\n", ptr);
       exit(1);
    }

    net[nindex].num_pins++;


    if (type == DRIVER) {
       num_driver[nindex]++;
       j=0;           /* Driver always in position 0 of pinlist */
    }    

    else {
       j = net[nindex].num_pins - num_driver[nindex];

   /* num_driver is the number of signal drivers of this net. *
    * should always be zero or 1 unless the netlist is bad.   */
//		printf("Net name : %s, number of pins: %d , num_driver: %d, nindex %d, temp_num_pin %d \n", net[nindex].name, net[nindex].num_pins, num_driver[nindex], nindex, temp_num_pins[nindex]);

       if (j >= temp_num_pins[nindex]) {
          printf("Error:  Net #%d (%s) has no driver and will cause\n",
             nindex, ptr);
          printf("memory corruption.\n");
          exit(1);
       } 
    }   
    net[nindex].blocks[j] = bnum;
    net[nindex].blk_pin[j] = blk_pnum;

	/*
	if (nindex == 10869)
	{
		printf("Add net 18069 (%s), bnum %d, blo_pnum %d, num_driver %d , num_pin %d \n", ptr, bnum, blk_pnum, num_driver[nindex], net[nindex].num_pins);
	}
	*/
	
    return (nindex);
 }
}


static void free_parse (void) {  

/* Release memory needed only during circuit netlist parsing. */

 int i;

 free (num_driver);
 free_hash_table (hash_table);
 free (temp_num_pins);

}


static void parse_carry_cluster_name(char * netname, char* cluster_name)
{
	int i;
	int check_count=0;
	//get the cluster name until "_cry"
	for (i=0;i < strlen(netname) -1; i++)
	{
		
		if (check_count <4)
		{
			cluster_name[i] = netname[i];
			if (check_count == 0)
			{
				if (netname[i] == '_')
				{
					check_count ++;
				}
			}
			else if (check_count == 1)
			{
				if (netname[i] == 'c')
				{
					check_count ++;
				}
				else
				{
					check_count =0;
				}
			}
			else if (check_count == 2)
			{
				if (netname[i] == 'r')
				{
					check_count ++;
				}
				else
				{
					check_count =0;
				}
			}
			else if (check_count == 3)
			{
				if (netname[i] == 'y')
				{
					check_count ++;
				}
				else
				{
					check_count =0;
				}
			}
		}
		else
		{
			//get until "_cry"
			cluster_name[i] = '\0';
			return;
		}
	}	
	return;
}



